/***************************************************************************
 *            ApplicationSettings.cc
 *
 *  Mon Nov 29 19:32:34 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/** \file
 * Application settings management class.
 */

#include "debug.h"

#include "ApplicationSettings.h"
#include "XMLSettingsManager.h"
#include "PopupTreeView.h"
#include "SettingsAdaptors.h"

#include "langsupport.h"
#include "support.h"

#include <gtkmm/fontbutton.h>
#include <gtkmm/checkbutton.h>
#include <gtkmm/radiobutton.h>
#include <gtkmm/spinbutton.h>
#include <gtkmm/range.h>
#include <gtkmm/liststore.h>
#include <gtkmm/dialog.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/stock.h>

// The real class

/** Initialize the class. The widgets are referenced, setup and loaded with 
 * the database values if available. Afterwards all signals are connected. 
 * An XMLSettingsManager is used.
 */

ApplicationSettings::ApplicationSettings() :
  Settings (SettingsManagerPtr (new XMLSettingsManager (".vocabtrainrc"))),
  scorefade (*this, "score", "fade", 
    "Fade the colored score display (0..1)",
    "Use a value of 0 for full color to 1 for invisible.",
    0, RangeAdaptor<double> ("scorefade")),
  scorevisible (*this, "score", "visible", 
    "Display the score using a colored background (true/false)",
    "If enabled the background of the main vocabulary list is displayed "
    "in shades from green to red to represent the current score.",
    true, CheckButtonAdaptor<bool> ("scorevisible"), VisibleAction ("scorefade")),
  scorecolumns (*this, "score", "columns", 
    "Show additional columns with score details (true/false)",
    "If enabled two more columns displaying the date of expiration and "
    "the current score are shown.",    
    false, CheckButtonAdaptor<bool> ("scorecolumns")),
  scorerepeats (*this, "score", "repeats", 
    "Repeat a vocabulary at least n times (>1)",
    "The score value is reduced in n steps from 1 to 0.",
    10, SpinButtonAdaptor<unsigned> ("scorerepeats")),
  scoreexpires (*this, "score", "expires", 
    "Maximum time for the expiration of a vocabulary (>1, in days)",
    "This time corresponds to a score value of 0, higher score values "
    "cause a shorter time between queries.",
    28, SpinButtonAdaptor<unsigned> ("scoreexpires")),
  scoreomit (*this, "training", "omit", 
    "Don't ask not-expired vocabularies (true/false)",
    "If enabled, only vocabularies that are expired (i.e. that were not "
    "asked during a certain time), are included in the training.",
    true, CheckButtonAdaptor<bool> ("scoreomit")),
  trainingmode (*this, "training", "mode", 
    "Ask in linear or random order (0=linear, 1=random)",
    "This determines whether the vocabularies should be shuffled before "
    "being presented to the user.",
    0, RadioButtonAdaptor (ParamList<std::string> ("selectionlinear")("selectionrandom"))),
  timerannoying (*this, "timer", "annoying",
    "Be a little bit more aggressive while asking in timer mode (X only)",
    "This causes keyboard lock and fullscreen.",
   false, CheckButtonAdaptor<bool> ("timerannoying")),
  sizewidth (*this, "geometry", "width", 
    "Initial window width (pixels)",
    "This value is autosaved on exit.",
    0),
  sizeheight (*this, "geometry", "height", 
    "Initial window height (pixels)",
    "This value is autosaved on exit.",
    0),
  sizestate (*this, "geometry", "maximized", 
    "Whether the initial window state is maximized (true/false)",
    "This value is autosaved on exit.",
    false),
  voicedetails (*this, "voice", "details",
    "Whether details for the voice analysis are expanded (true/false)",
    "This value is autosaved on exit.",
    false, ExpanderAdaptor ("voicedetails")),
  userfit (*this, "voice", "userfit",
    "Whether the user spectrogram display is set to scale-to-fit (true/false)",
    "This value is autosaved on exit.",
    false, ToggleButtonAdaptor<bool> ("userfit")),
  reffit (*this, "voice", "reffit",
    "Whether the reference spectrogram display is set to scale-to-fit (true/false)",
    "This value is autosaved on exit.",
    false, ToggleButtonAdaptor<bool> ("reffit")),
  askmode (*this, "default", "buttons", 
    "Let the user pick an alternative or enter the answer directly (0=input, 1=buttons)",
    "This determines the default setting from the preference panel. "
    "The user can change it temporarily for the current file.",
    0, RadioButtonAdaptor (ParamList<std::string> ("askwritten_default")("askpick_default"))),
  presentsolution (*this, "default", "acousticsolution",
    "Whether speech synthesis should be used to present the solution (true/false)",
    "This enables speech synthesis for solution presentation if "
    "synthesis support for this language is available.",
    false, ToggleButtonAdaptor<bool> ("presentsolution_default")),
  lastfiles (*this, "recently_used", 5, 
    MenuMRUAdaptor ("menu_mru")),
  stress_store (Gtk::ListStore::create (stress_columns)),
  synth_store (Gtk::ListStore::create (synth_columns)),
  word_store (Gtk::ListStore::create (word_columns))
{ 
    debug_log ("Loading up application settings"); 
    
    // This is for the automatic part
    init();
    
    // This mode is not supported on Win32
#ifdef WIN32
    Gtk::Widget *annoying;
    xml->get_widget ("timerannoying", annoying);
    annoying->hide();
#endif
    
    // and all this for the manual one
    xml->get_widget ("presenable", presenable);
    presenable->signal_toggled().connect (sigc::mem_fun (*this, &ApplicationSettings::on_presfont_enable_changed));
    xml->get_widget ("presfont", presfont);
    presfont->signal_font_set().connect (sigc::mem_fun (*this, &ApplicationSettings::on_presfont_name_changed));
    
    xml->get_widget ("deflang", deflang);
    deflang->signal_changed().connect (sigc::mem_fun (*this, &ApplicationSettings::on_default_languages_changed));

    xml->get_widget ("timesspin", timesspin);
    timesspin->signal_changed().connect (sigc::mem_fun (*this, &ApplicationSettings::on_repeat_num_changed));
    xml->get_widget ("timeswrongspin", timeswrongspin);
    timeswrongspin->signal_changed().connect (sigc::mem_fun (*this, &ApplicationSettings::on_repeat_wrong_num_changed));

    xml->get_widget ("buttonssorted", buttonssorted);
    buttonssorted->signal_toggled().connect (sigc::mem_fun (*this, &ApplicationSettings::on_buttons_sorted_changed));
    xml->get_widget ("prefersame", prefersame);
    prefersame->signal_toggled().connect (sigc::mem_fun (*this, &ApplicationSettings::on_buttons_same_type_changed));
    xml->get_widget ("buttonsnum", buttonsnum);
    buttonsnum->signal_changed().connect (sigc::mem_fun (*this, &ApplicationSettings::on_buttons_num_changed));

    xml->get_widget("firstlanguage_default", firstlanguage_default);
    firstlanguage_default->signal_changed().connect (sigc::bind<int> (sigc::mem_fun (*this, &ApplicationSettings::on_language_default_changed), 0));
    
    xml->get_widget ("secondlanguage_default", secondlanguage_default);
    secondlanguage_default->signal_changed().connect (sigc::bind<int> (sigc::mem_fun (*this, &ApplicationSettings::on_language_default_changed), 1));
    
    xml->get_widget ("presentwritten_default", presentwritten_default);
    presentwritten_default->signal_toggled().connect (sigc::mem_fun (*this, &ApplicationSettings::on_presentwritten_default_changed));
    xml->get_widget ("presentaudio_default", presentaudio_default);
    presentaudio_default->signal_toggled().connect (sigc::mem_fun (*this, &ApplicationSettings::on_presentaudio_default_changed));
    xml->get_widget ("usetimer_default", usetimer_default);
    usetimer_default->signal_toggled().connect (sigc::mem_fun (*this, &ApplicationSettings::on_use_timer_default_changed));
    
    xml->get_widget ("timerusevlimit", timerusevlimit);
    timerusevlimit->signal_toggled().connect (sigc::mem_fun (*this, &ApplicationSettings::on_use_timer_vocab_limit_changed));
    xml->get_widget ("timerlimitanswers", timerlimitanswers);
    timerlimitanswers->signal_toggled().connect (sigc::mem_fun (*this, &ApplicationSettings::on_use_timer_answer_limit_changed));
    xml->get_widget ("timerinterval", timerinterval);
    timerinterval->signal_changed().connect (sigc::mem_fun (*this, &ApplicationSettings::on_timer_interval_changed));
    xml->get_widget ("timervocabnum", timervocabnum);
    timervocabnum->signal_changed().connect (sigc::mem_fun (*this, &ApplicationSettings::on_timer_vocab_num_changed));

    xml->get_widget_derived ("stresslist", stresslist);
    stresslist->signal_delete_row().connect (sigc::mem_fun (*this, &ApplicationSettings::on_stressbase_delete_base));
    stresslist->set_model (stress_store);
    stresslist->append_column_editable (_("Language"), stress_columns.language);
    stresslist->append_column_editable (_("Base"), stress_columns.base);
    xml->get_widget ("stressadd", stressadd);
    stressadd->signal_clicked().connect (sigc::mem_fun (*this, &ApplicationSettings::on_stressbase_add_base));
    xml->get_widget ("stressdelete", stressdelete);
    stressdelete->signal_clicked().connect (sigc::mem_fun (*this, &ApplicationSettings::on_stressbase_delete_base));
    xml->get_widget ("stressselect", stressselect);
    stressselect->signal_clicked().connect (sigc::mem_fun (*this, &ApplicationSettings::on_stressbase_select_base));

    xml->get_widget_derived ("synthlist", synthlist);
    synthlist->signal_delete_row().connect (sigc::mem_fun (*this, &ApplicationSettings::on_synthesis_delete_program));
    synthlist->set_model (synth_store);
    synthlist->append_column_editable (_("Language"), synth_columns.language);
    synthlist->append_column_editable (_("Program"), synth_columns.program);
    synthlist->append_column_editable (_("Parameters"), synth_columns.params);
    xml->get_widget ("synthadd", synthadd);
    synthadd->signal_clicked().connect (sigc::mem_fun (*this, &ApplicationSettings::on_synthesis_add_program));
    xml->get_widget ("synthdelete", synthdelete);
    synthdelete->signal_clicked().connect (sigc::mem_fun (*this, &ApplicationSettings::on_synthesis_delete_program));
    xml->get_widget ("synthselect", synthselect);
    synthselect->signal_clicked().connect (sigc::mem_fun (*this, &ApplicationSettings::on_synthesis_select_program));

    xml->get_widget ("synthlimit", synthlimit);
    synthlimit->signal_toggled().connect (sigc::mem_fun (*this, &ApplicationSettings::on_synthesis_timeout_enable_changed));
    xml->get_widget ("synthtimeout", synthtimeout);
    synthtimeout->signal_changed().connect (sigc::mem_fun (*this, &ApplicationSettings::on_synthesis_timeout_changed));
    
    xml->get_widget_derived ("wordlist", wordlist);
    wordlist->signal_delete_row().connect (sigc::mem_fun (*this, &ApplicationSettings::on_localized_delete));
    wordlist->set_model (word_store);
    wordlist->append_column_editable (_("Language"), word_columns.language);
    wordlist->append_column_editable (_("Or"), word_columns.word_or);
    xml->get_widget ("wordadd", wordadd);
    wordadd->signal_clicked().connect (sigc::mem_fun (*this, &ApplicationSettings::on_localized_add));
    xml->get_widget ("worddelete", worddelete);
    worddelete->signal_clicked().connect (sigc::mem_fun (*this, &ApplicationSettings::on_localized_delete));

    update_default_languages();
    update_presfont_enable();
    update_presfont_name();
    update_repeat_num();
    update_buttons_sorted();
    update_buttons_same_type();
    update_buttons_num();
    update_first_default();
    update_second_default();
    update_audio_default();
    update_written_default();
    update_use_timer_default();
    update_synthesis_timeout_enable();
    update_synthesis_timeout();
    
    update_use_timer_vocab_limit();
    update_use_timer_answer_limit();
    update_timer_interval();
    update_timer_vocab_num();

    update_stressbase();
    update_synthesis();
    update_localized();

    // These two have to come after the update process
    stress_store->signal_row_changed().connect (sigc::mem_fun (*this, &ApplicationSettings::on_stressbase_changed));
    synth_store->signal_row_changed().connect (sigc::mem_fun (*this, &ApplicationSettings::on_synthesis_changed));
    word_store->signal_row_changed().connect (sigc::mem_fun (*this, &ApplicationSettings::on_localized_changed));

    // Signals
    manager->get_signal ("fonts", "pres_font").connect (signal_presfont_v.make_slot());
    manager->get_signal ("fonts", "pres_enable").connect (signal_presfont_v.make_slot());
    
    manager->get_signal ("training", "repeat_initially").connect (signal_repeat_num_v.make_slot());
    manager->get_signal ("training", "repeat_wrong").connect (signal_repeat_num_v.make_slot());
    manager->get_signal ("training", "buttons_sort").connect (signal_buttons_v.make_slot());
    manager->get_signal ("training", "prefer_same").connect (signal_buttons_v.make_slot());
    manager->get_signal ("training", "buttons_num").connect (signal_buttons_v.make_slot());
    
    manager->get_signal ("default", "first").connect (signal_languages_default_v.make_slot());
    manager->get_signal ("default", "second").connect (signal_languages_default_v.make_slot());
    manager->get_signal ("default", "audio").connect (signal_mode_default_v.make_slot());
    manager->get_signal ("default", "written").connect (signal_mode_default_v.make_slot());
    manager->get_signal ("default", "buttons").connect (signal_mode_default_v.make_slot());
    
    manager->get_signal ("timer", "vocablimit_enable").connect (signal_timer_v.make_slot());
    manager->get_signal ("timer", "vocablimit_answers").connect (signal_timer_v.make_slot());
    manager->get_signal ("timer", "vocablimit").connect (signal_timer_v.make_slot());
    manager->get_signal ("timer", "interval").connect (signal_timer_v.make_slot());

    manager->get_signal ("new", "languages").connect (signal_default_languages_v.make_slot());

    manager->get_signal ("synthesis_options", "timeout_enable").connect (signal_synthesis_timeout_v.make_slot());
    manager->get_signal ("synthesis_options", "timeout").connect (signal_synthesis_timeout_v.make_slot());
}

/** Shut down the class. Currently only a debug message is printed out.
 */

ApplicationSettings::~ApplicationSettings()
{
    debug_log ("Shutting down application settings"); 
}

/** Returns a copy of the signal for changes to the initial language selection for new
 * files.
 * \return The notification signal
 */

sigc::signal<void> ApplicationSettings::signal_default_languages()
{
    return signal_default_languages_v;
}

void ApplicationSettings::on_default_languages_changed()
{
    debug_log ("Default languages were changed to %1", deflang->get_text()); 
    set_default_languages (deflang->get_text());
}

void ApplicationSettings::set_default_languages (const std::string& value)
{
    // Set the new value
    debug_log ("Changing default languages in config"); 
    manager->set_string ("new", "languages", value);
}

void ApplicationSettings::update_default_languages()
{
    debug_log ("Updating default languages"); 
    std::string value;
    if (manager->get_checked_string ("new", "languages", value))
      deflang->set_text (value);
    on_default_languages_changed();
}

// Font selection

/** Returns a copy of the signal for changes of the presentation font or size.
 * \return The notification signal
 */

sigc::signal<void> ApplicationSettings::signal_presfont()
{
    return signal_presfont_v;
}

void ApplicationSettings::on_presfont_enable_changed()
{
    debug_log ("Presfont status was changed to %1", presenable->get_active()); 
    set_presfont_enable (presenable->get_active());
}

void ApplicationSettings::on_presfont_name_changed()
{
    debug_log ("Presfont name was changed to %1", presfont->get_font_name());
    set_presfont_name (presfont->get_font_name());
}

void ApplicationSettings::set_presfont_enable (const bool value)
{
    // Set the new value
    debug_log ("Changing presfont status in config"); 
    manager->set_bool ("fonts", "pres_enable", value);
    // Update sensitivity of font selector
    presfont->set_sensitive (value);
}

void ApplicationSettings::update_presfont_enable()
{
    debug_log ("Updating presfont status"); 
    bool value;
    if (manager->get_checked_bool ("fonts", "pres_enable", value))
      presenable->set_active (value);
    on_presfont_enable_changed();
}

/** Returns whether the presentation font is user defined.
 * \return State of the presentation font
 */

bool ApplicationSettings::is_presfont() const
{
    return manager->get_bool ("fonts", "pres_enable");
}

void ApplicationSettings::set_presfont_name (const std::string& value)
{
    // Set the new value
    debug_log ("Setting new presfont name in config"); 
    manager->set_string ("fonts", "pres_font", value);
}

void ApplicationSettings::update_presfont_name()
{
    debug_log ("Updating presfont name"); 
    std::string value;
    if (manager->get_checked_string ("fonts", "pres_font", value))
      presfont->set_font_name (value);
    on_presfont_name_changed();
}

/** Returns the specified presentation font that should be used instead of the system
 * default one. Use is_presfont() to see whether this font should really be used.
 * \return Name and size of the presentation font
 */

std::string ApplicationSettings::get_presfont() const
{
    return manager->get_string ("fonts", "pres_font");
}

// Repeat count

/** Returns a copy of the signal for changes to the numbers of repeats for right or wrong answers.
 * \return The notification signal
 */

sigc::signal<void> ApplicationSettings::signal_repeat_num()
{
    return signal_repeat_num_v;
}

void ApplicationSettings::on_repeat_num_changed()
{
    set_repeat_num (timesspin->get_value_as_int());
}

void ApplicationSettings::set_repeat_num (const unsigned value)
{
    // Set the new value
    manager->set_unsigned ("training", "repeat_initially", value);
}

void ApplicationSettings::update_repeat_num()
{
    unsigned value;
    if (manager->get_checked_unsigned ("training", "repeat_initially", value))
      timesspin->set_value (value);
    on_repeat_num_changed();
}

/** Returns the repeat count for right answers.
 * \return Repeat count
 */

unsigned ApplicationSettings::get_repeat_num() const
{
    return manager->get_unsigned ("training", "repeat_initially");
}

void ApplicationSettings::on_repeat_wrong_num_changed()
{
    set_repeat_wrong_num (timeswrongspin->get_value_as_int());
}

void ApplicationSettings::set_repeat_wrong_num (const unsigned value)
{
    // Set the new value
    manager->set_unsigned ("training", "repeat_wrong", value);
}

void ApplicationSettings::update_repeat_wrong_num()
{
    unsigned value;
    if (manager->get_checked_unsigned ("training", "repeat_wrong", value))
      timeswrongspin->set_value (value);
    on_repeat_wrong_num_changed();
}

/** Returns the repeat count for wrong answers.
 * \return Repeat count
 */

unsigned ApplicationSettings::get_repeat_wrong_num() const
{
    return manager->get_unsigned ("training", "repeat_wrong");
}

// Alternative presentation - number of buttons

/** Returns a copy of the signal for changes to the number of alternative buttons.
 * \return The notification signal
 */

sigc::signal<void> ApplicationSettings::signal_buttons()
{
    return signal_buttons_v;
}

void ApplicationSettings::on_buttons_num_changed()
{
    set_buttons_num (buttonsnum->get_value_as_int());
}

void ApplicationSettings::set_buttons_num (const unsigned value)
{
    // Set the new value
    manager->set_unsigned ("training", "buttons_num", value);
}

void ApplicationSettings::update_buttons_num()
{
    unsigned value;
    if (manager->get_checked_unsigned ("training", "buttons_num", value))
      buttonsnum->set_value (value);
    on_buttons_num_changed();
}

/** Returns the number of alternative buttons.
 * \return Number of buttons
 */

unsigned ApplicationSettings::get_buttons_num() const
{
    return manager->get_unsigned ("training", "buttons_num");
}

// Alternative presentation - sort buttons

void ApplicationSettings::on_buttons_sorted_changed()
{
    set_buttons_sorted (buttonssorted->get_active());
}

void ApplicationSettings::set_buttons_sorted (const bool value)
{
    // Set the new value
    manager->set_bool ("training", "buttons_sorted", value);
}

void ApplicationSettings::update_buttons_sorted()
{
    bool value;
    if (manager->get_checked_bool ("training", "buttons_sorted", value))
      buttonssorted->set_active (value);
    on_buttons_sorted_changed();
}

/** Returns whether the alternative buttons are sorted or in random order.
 * \return Whether the buttons are sorted
 */

bool ApplicationSettings::is_buttons_sorted() const
{
    return manager->get_bool ("training", "buttons_sorted");
}

// Alternative presentation - prefer same type

void ApplicationSettings::on_buttons_same_type_changed()
{
    set_buttons_same_type (prefersame->get_active());
}

void ApplicationSettings::set_buttons_same_type (const bool value)
{
    // Set the new value
    manager->set_bool ("training", "prefer_same", value);
}

void ApplicationSettings::update_buttons_same_type()
{
    bool value;
    if (manager->get_checked_bool ("training", "prefer_same", value))
      prefersame->set_active (value);
    on_buttons_same_type_changed();
}

/** Returns whether alternatives from the same type as the asked vocabulary are preferred
 * over other alternatives.
 * \return Whether the same type is preferred
 */

bool ApplicationSettings::is_buttons_same_type() const
{
    return manager->get_bool ("training", "prefer_same");
}

// Default page - Timer enabled

/** Returns a copy of the signal for changes to the use of the timer enabled training
 * after file loading.
 * \return The notification signal
 */

sigc::signal<void> ApplicationSettings::signal_use_timer_default()
{
    return signal_use_timer_default_v;
}

void ApplicationSettings::on_use_timer_default_changed()
{
    set_use_timer_default (usetimer_default->get_active());
}

void ApplicationSettings::set_use_timer_default (const bool value)
{
    // Set the new value
    manager->set_bool ("default", "use_timer", value);
}

void ApplicationSettings::update_use_timer_default()
{
    bool value;
    if (manager->get_checked_bool ("default", "use_timer", value))
      usetimer_default->set_active (value);
    on_use_timer_default_changed();
}

/** Returns whether the timer is enabled after a new file is loaded.
 * \return Timer state
 */

bool ApplicationSettings::is_use_timer_default() const
{
    return manager->get_bool ("default", "use_timer");
}

// Default page - Mode selection

/** Returns a copy of the signal for changes to the training mode selection after file
 * loading.
 * \return The notification signal
 */

sigc::signal<void> ApplicationSettings::signal_mode_default()
{
    return signal_mode_default_v;
}

void ApplicationSettings::on_presentwritten_default_changed()
{
    set_written_default (presentwritten_default->get_active());
}

void ApplicationSettings::on_presentaudio_default_changed()
{
    set_audio_default (presentaudio_default->get_active());
}

void ApplicationSettings::set_audio_default (const bool value)
{
    // Set the new value
    manager->set_bool ("default", "audio", value);
    // Try to set other checkbox if its necessary
    if (! value && ! is_written_default()) {
        set_written_default (true);
	update_written_default();
    }
}

void ApplicationSettings::set_written_default (const bool value)
{
    // Set the new value
    manager->set_bool ("default", "written", value);
    // Try to set other checkbox if its necessary
    if (! value && ! is_audio_default()) {
        set_audio_default (true);
	update_audio_default();
    }
}

void ApplicationSettings::update_written_default()
{
    bool value;
    if (manager->get_checked_bool ("default", "written", value))
      presentwritten_default->set_active (value);
    on_presentwritten_default_changed();
}

void ApplicationSettings::update_audio_default()
{
    bool value;
    if (manager->get_checked_bool ("default", "audio", value))
      presentaudio_default->set_active (value);
    on_presentaudio_default_changed();
}

/** Returns whether audio output is enabled after file loading.
 * \return Audio state
 */

bool ApplicationSettings::is_audio_default() const
{
    return manager->get_bool ("default", "audio");
}

/** Returns whether written output is enabled after file loading.
 * \return Written state
 */

bool ApplicationSettings::is_written_default() const
{
    return manager->get_bool ("default", "written");
}

// Default page - Languages

void ApplicationSettings::on_language_default_changed (const int which)
{
    if (! which)
      set_first_default (firstlanguage_default->get_text());
    else
      set_second_default (secondlanguage_default->get_text());
}

/** Returns a copy of the signal for changes to the language selection after file loading.
 * \return The notification signal
 */

sigc::signal<void> ApplicationSettings::signal_languages_default()
{
    return signal_languages_default_v;
}

/** Returns the selected primary language after file loading.
 * \return Primary language
 */

std::string ApplicationSettings::get_first_default() const
{
    return manager->get_string ("default", "first");
}

/** Returns the selected secondary language after file loading.
 * \return Secondary language
 */

std::string ApplicationSettings::get_second_default() const
{
    return manager->get_string ("default", "second");
}

void ApplicationSettings::set_first_default (const std::string& value)
{
    // Set the new value
    manager->set_string ("default", "first", value);
}

void ApplicationSettings::update_first_default()
{
    std::string value;
    if (manager->get_checked_string ("default", "first", value))
      firstlanguage_default->set_text (value);
    on_language_default_changed (0);
}

void ApplicationSettings::set_second_default (const std::string& value)
{
    // Set the new value
    manager->set_string ("default", "second", value);
}

void ApplicationSettings::update_second_default()
{
    std::string value;
    if (manager->get_checked_string ("default", "second", value))
      secondlanguage_default->set_text (value);
    on_language_default_changed (1);
}

// Timer page - Vocab limit enable

/** Returns whether the timer uses a vocabulary limit.
 * \return State of vocabulary limit
 */

bool ApplicationSettings::is_use_timer_vocab_limit() const
{
    return manager->get_bool ("timer", "vocablimit_enable");
}

void ApplicationSettings::on_use_timer_vocab_limit_changed()
{
    set_use_timer_vocab_limit (timerusevlimit->get_active());
}
    
void ApplicationSettings::set_use_timer_vocab_limit (const bool value)
{
    manager->set_bool ("timer", "vocablimit_enable", value);
}

void ApplicationSettings::update_use_timer_vocab_limit()
{
    bool value;
    if (manager->get_checked_bool ("timer", "vocablimit_enable", value))
      timerusevlimit->set_active (value);
    on_use_timer_vocab_limit_changed();
}
    
// Timer page - Limit answers instead of vocab number

/** Returns whether the timer uses a limit on right answers instead of on the number of vocabularies.
 * \return Whether the number of answers is limited
 */

bool ApplicationSettings::is_use_timer_answer_limit() const
{
    return manager->get_bool ("timer", "vocablimit_answers");
}

void ApplicationSettings::on_use_timer_answer_limit_changed()
{
    set_use_timer_answer_limit (timerlimitanswers->get_active());
}
    
void ApplicationSettings::set_use_timer_answer_limit (const bool value)
{
    manager->set_bool ("timer", "vocablimit_answers", value);
}

void ApplicationSettings::update_use_timer_answer_limit()
{
    bool value;
    if (manager->get_checked_bool ("timer", "vocablimit_answers", value))
      timerlimitanswers->set_active (value);
    on_use_timer_answer_limit_changed();
}

// Timer page - Vocab limit

/** Returns the vocabulary or answer limit for the timer.
 * \return Number of necessary right vocabularies or answers
 */

unsigned ApplicationSettings::get_timer_vocab_num() const
{
    return manager->get_unsigned ("timer", "vocablimit");
}

void ApplicationSettings::on_timer_vocab_num_changed()
{
    set_timer_vocab_num (timervocabnum->get_value_as_int());
}
    
void ApplicationSettings::set_timer_vocab_num (const unsigned value)
{
    manager->set_unsigned ("timer", "vocablimit", value);
}

void ApplicationSettings::update_timer_vocab_num()
{
    unsigned value;
    if (manager->get_checked_unsigned ("timer", "vocablimit", value))
      timervocabnum->set_value (value);
    on_timer_vocab_num_changed();
}

// Timer page - Timer interval

/** Returns the length of the pause the timer mode makes between test sessions.
 * \return Pause length in minutes
 */

unsigned ApplicationSettings::get_timer_interval() const
{
    return manager->get_unsigned ("timer", "interval");
}

void ApplicationSettings::on_timer_interval_changed()
{
    set_timer_interval (timerinterval->get_value_as_int());
}
    
void ApplicationSettings::set_timer_interval (const unsigned value)
{
    manager->set_unsigned ("timer", "interval", value);
}

void ApplicationSettings::update_timer_interval()
{
    unsigned value;
    if (manager->get_checked_unsigned ("timer", "interval", value))
      timerinterval->set_value (value);
    on_timer_interval_changed();
}

/** Returns a copy of the signal for changes to timer related settings.
 * \return The notification signal
 */

sigc::signal<void> ApplicationSettings::signal_timer()
{
    return signal_timer_v;
}

// Stressbase page

void ApplicationSettings::on_stressbase_add_base()
{
    stressbase_add ("New language", "");
    rebuild_stressbase();
}

void ApplicationSettings::stressbase_add (const std::string& language, const std::string& path)
{
    Gtk::TreeModel::Row row = *(stress_store->append());
    row[stress_columns.language] = language;
    row[stress_columns.base] = path;
}

void ApplicationSettings::on_stressbase_delete_base ()
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    if (! cursor_visible (stresslist)) return;
    stresslist->get_cursor (p, c);
    stress_store->erase (stress_store->get_iter (p));
    rebuild_stressbase();
}

void ApplicationSettings::on_stressbase_select_base ()
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    if (cursor_visible (stresslist)) {
	stresslist->get_cursor (p, c);
        Gtk::TreeModel::Row row = *(stress_store->get_iter (p));
	
	Gtk::Dialog *settings;    
	xml->get_widget ("SettingsDlg", settings);
	
	Gtk::FileChooserDialog select (_("Select Stress Database"), Gtk::FILE_CHOOSER_ACTION_OPEN);
	select.set_filename (row[stress_columns.base]);
	select.set_transient_for (*settings);

	//Add response buttons the the dialog:
	select.add_button (Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	select.add_button (Gtk::Stock::OPEN, Gtk::RESPONSE_OK)->has_default();

	//Add filters, so that only certain file types can be selected
	Gtk::FileFilter filter_txt;
	filter_txt.set_name (_("Database files"));
	filter_txt.add_pattern ("*.stb"); 
	select.add_filter (filter_txt);

	Gtk::FileFilter filter_any;
	filter_any.set_name ("Any files");
	filter_any.add_pattern ("*");
	select.add_filter (filter_any); 

	if (select.run() == Gtk::RESPONSE_OK) {
	    select.hide();
	    std::string filename = select.get_filename();
	    row[stress_columns.base] = filename;
	}
    }
    rebuild_stressbase();
}

void ApplicationSettings::on_stressbase_changed (const Gtk::TreePath& path, const Gtk::TreeIter& iter)
{
    debug_log ("Stressbase change at %1", path.to_string());
    rebuild_stressbase();
}

void ApplicationSettings::rebuild_stressbase()
{
    std::string language;
    std::list<std::string> available;
    while ((language = manager->get_name ("stressbase", 0)) != "")
      manager->delete_entry ("stressbase", language);
    Gtk::TreeNodeChildren children = stress_store->children();
    for (Gtk::TreeIter i = children.begin(); i != children.end(); ++i)
      manager->set_string ("stressbase", 
      Glib::ustring ((*i)[stress_columns.language]), 
      Glib::ustring ((*i)[stress_columns.base]));
}

void ApplicationSettings::update_stressbase()
{
    std::string language;
    for (unsigned i = 0; (language = manager->get_name ("stressbase", i)) != ""; i++)
    {
	std::string path = manager->get_string ("stressbase", language);
	stressbase_add (language, path);
    }
}

/** Returns the stressbase filename for a particular language.
 * \return Filename or an empty string if not available
 */

std::string ApplicationSettings::get_stressbase_base (const std::string& language) const
{
    std::string base;
    if (manager->get_checked_string ("stressbase", language, base))
      return base;
    else
      return "";
}

// Synthesis page

void ApplicationSettings::on_synthesis_add_program()
{
    synthesis_add ("New language", "", "");
    rebuild_synthesis();
}

void ApplicationSettings::synthesis_add (const std::string& language, const std::string& path,
  const std::string& params)
{
    Gtk::TreeModel::Row row = *(synth_store->append());
    row[synth_columns.language] = language;
    row[synth_columns.program] = path;
    row[synth_columns.params] = params;
}

void ApplicationSettings::on_synthesis_delete_program ()
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    if (! cursor_visible (synthlist)) return;
    synthlist->get_cursor (p, c);
    synth_store->erase (synth_store->get_iter (p));
    rebuild_synthesis();
}

void ApplicationSettings::on_synthesis_select_program ()
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    if (cursor_visible (synthlist)) {
	synthlist->get_cursor (p, c);
        Gtk::TreeModel::Row row = *(synth_store->get_iter (p));
	
	Gtk::Dialog *settings;    
	xml->get_widget ("SettingsDlg", settings);
	
	Gtk::FileChooserDialog select (_("Select Synthesis Program"), Gtk::FILE_CHOOSER_ACTION_OPEN);
	select.set_filename (row[synth_columns.program]);
	select.set_transient_for (*settings);

	//Add response buttons the the dialog:
	select.add_button (Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	select.add_button (Gtk::Stock::OPEN, Gtk::RESPONSE_OK)->has_default();

	//Add filters, so that only certain file types can be selected
	Gtk::FileFilter filter_exe;
	filter_exe.set_name (_("Executables"));
	filter_exe.add_mime_type ("application/x-executable"); 
	filter_exe.add_mime_type ("application/x-shellscript");
#ifdef WIN32
	filter_any.add_pattern ("*.exe");
#endif
	select.add_filter (filter_exe);

	Gtk::FileFilter filter_any;
	filter_any.set_name (_("Any files"));
	filter_any.add_pattern ("*");
	select.add_filter (filter_any); 

	if (select.run() == Gtk::RESPONSE_OK) {
	    select.hide();
	    std::string filename = select.get_filename();
	    row[synth_columns.program] = filename;
	}
    }
}

void ApplicationSettings::on_synthesis_changed (const Gtk::TreePath& path, const Gtk::TreeIter& iter)
{
    debug_log ("Synthesis program change at %1", path.to_string());
    rebuild_synthesis();
}

void ApplicationSettings::rebuild_synthesis()
{
    std::string language;
    std::list<std::string> available;
    while ((language = manager->get_name ("synthesis", 0)) != "")
      manager->delete_entry ("synthesis", language);
    while ((language = manager->get_name ("synthesis_params", 0)) != "")
      manager->delete_entry ("synthesis_params", language);
    Gtk::TreeNodeChildren children = synth_store->children();
    for (Gtk::TreeIter i = children.begin(); i != children.end(); ++i) {
        manager->set_string ("synthesis", 
          Glib::ustring ((*i)[synth_columns.language]), 
          Glib::ustring ((*i)[synth_columns.program]));
        manager->set_string ("synthesis_params", 
          Glib::ustring ((*i)[synth_columns.language]), 
          Glib::ustring ((*i)[synth_columns.params]));
    }
}

void ApplicationSettings::update_synthesis()
{
    std::string language;
    for (unsigned i = 0; (language = manager->get_name ("synthesis", i)) != ""; i++)
    {
	std::string path = manager->get_string ("synthesis", language);
	std::string params = manager->get_string ("synthesis_params", language);
	synthesis_add (language, path, params);
    }
}

/** Returns the synthesis program to use for a particular language.
 * \return Path and filename or an empty string if not available
 */

std::string ApplicationSettings::get_synthesis_program (const std::string& language) const
{
    return manager->get_string ("synthesis", language);
}

/** Returns the parameter for the synthesis program to use for a particular language.
 * \return Parameters or an empty string if not available
 */

std::string ApplicationSettings::get_synthesis_params (const std::string& language) const
{
    return manager->get_string ("synthesis_params", language);
}

// Synthesis Timeout

/** Returns a copy of the signal for changes to the timeout of synthesis programs.
 * \return The notification signal
 */

sigc::signal<void> ApplicationSettings::signal_synthesis_timeout()
{
    return signal_synthesis_timeout_v;
}

void ApplicationSettings::on_synthesis_timeout_enable_changed()
{
    debug_log ("Synthesis timeout enable was changed to %1", synthlimit->get_active()); 
    set_synthesis_timeout_enable (synthlimit->get_active());
}

void ApplicationSettings::on_synthesis_timeout_changed()
{
    debug_log ("Synthesis timeout was changed to %1", synthtimeout->get_value_as_int());
    set_synthesis_timeout (synthtimeout->get_value_as_int());
}

void ApplicationSettings::set_synthesis_timeout_enable (const bool value)
{
    // Set the new value
    debug_log ("Changing synthesis timeout status in config"); 
    manager->set_bool ("synthesis_options", "timeout_enable", value);
    // Update sensitivity of timeout value
    synthtimeout->set_sensitive (value);
}

void ApplicationSettings::update_synthesis_timeout_enable()
{
    debug_log ("Updating synthesis timeout enable status"); 
    bool value;
    if (manager->get_checked_bool ("synthesis_options", "timeout_enable", value))
      synthlimit->set_active (value);
    on_synthesis_timeout_enable_changed();
}

/** Returns whether the synthesis programs are killed after a certain time or remain started infinitely.
 * \return Whether the programs are time limited
 */

bool ApplicationSettings::is_synthesis_timeout() const
{
    return manager->get_bool ("synthesis_options", "timeout_enable");
}

void ApplicationSettings::set_synthesis_timeout (const unsigned value)
{
    // Set the new value
    debug_log ("Setting new synthesis timeout in config"); 
    manager->set_unsigned ("synthesis_options", "timeout", value);
}

void ApplicationSettings::update_synthesis_timeout()
{
    debug_log ("Updating synthesis timeout"); 
    unsigned value;
    if (manager->get_checked_unsigned ("synthesis_options", "timeout", value))
      synthtimeout->set_value (value);
    on_synthesis_timeout_changed();
}

/** Returns the timeout for synthesis programs. Refer to is_synthesis_timeout() to see
 * whether it should be used.
 * \return Interval in seconds
 */

unsigned ApplicationSettings::get_synthesis_timeout() const
{
    return manager->get_unsigned ("synthesis_options", "timeout");
}

// Localized page

void ApplicationSettings::on_localized_add()
{
    localized_add ("New language", "");
    rebuild_localized();
}

void ApplicationSettings::localized_add (const std::string& language, const std::string& word_or)
{
    Gtk::TreeModel::Row row = *(word_store->append());
    row[word_columns.language] = language;
    row[word_columns.word_or] = word_or;
}

void ApplicationSettings::on_localized_delete()
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    if (! cursor_visible (wordlist)) return;
    wordlist->get_cursor (p, c);
    word_store->erase (word_store->get_iter (p));
    rebuild_localized();
}

void ApplicationSettings::on_localized_changed (const Gtk::TreePath& path, const Gtk::TreeIter& iter)
{
    debug_log ("Localized change at %1", path.to_string());
    rebuild_localized();
}

void ApplicationSettings::rebuild_localized()
{
    std::string language;
    std::list<std::string> available;
    while ((language = manager->get_name ("localized_or", 0)) != "")
      manager->delete_entry ("localized_or", language);
    Gtk::TreeNodeChildren children = word_store->children();
    for (Gtk::TreeIter i = children.begin(); i != children.end(); ++i)
      manager->set_string ("localized_or", 
      Glib::ustring ((*i)[word_columns.language]), 
      Glib::ustring ((*i)[word_columns.word_or]));
}

void ApplicationSettings::update_localized()
{
    std::string language;
    for (unsigned i = 0; (language = manager->get_name ("localized_or", i)) != ""; i++)
    {
	std::string word_or = manager->get_string ("localized_or", language);
	localized_add (language, word_or);
    }
}

/** Returns a localized version of "or" for a particular language.
 * \return Localized version or a gettextized version of "or" to the users system
 * language as fallback
 */

std::string ApplicationSettings::get_localized_or (const std::string& language) const
{
    std::string word_or;
    if (manager->get_checked_string ("localized_or", language, word_or))
      return word_or;
    else
      return _("or");
}
