/***************************************************************************
 *            ApplicationSettings.h
 *
 *  Mon Nov 29 19:28:36 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _APPLICATIONSETTINGS_H
#define _APPLICATIONSETTINGS_H

/** \file
 * Application settings management class.
 */

#include "SettingsHelper.h"
#include "RegisteredSingleton.h"

#include <gtkmm/treemodel.h>

#include <boost/utility.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/tokenizer.hpp>

#ifndef DOXYGEN_SHOULD_SKIP_THIS

namespace Gtk {
    class Button;
    class TreeView;
    class Entry;
    class FontButton;
    class CheckButton;
    class RadioButton;
    class SpinButton;
    class ListStore;
    class TreeViewColumn;
    class TreePath;
    class TreeIter;
}

class PopupTreeView;
    
class StressListModelColumns : public Gtk::TreeModel::ColumnRecord
{
  public:
    StressListModelColumns() { add (language); add (base); };

    Gtk::TreeModelColumn<Glib::ustring> language;
    Gtk::TreeModelColumn<Glib::ustring> base;
};

class SynthListModelColumns : public Gtk::TreeModel::ColumnRecord
{
  public:
    SynthListModelColumns() { add (language); add (program); add (params); };

    Gtk::TreeModelColumn<Glib::ustring> language;
    Gtk::TreeModelColumn<Glib::ustring> program;
    Gtk::TreeModelColumn<Glib::ustring> params;
};

class WordListModelColumns : public Gtk::TreeModel::ColumnRecord
{
  public:
    WordListModelColumns() { add (language); add (word_or); };

    Gtk::TreeModelColumn<Glib::ustring> language;
    Gtk::TreeModelColumn<Glib::ustring> word_or;
};

#endif

/** Provides global settings throughout the application. It is implemented as a
 * singleton and uses an SettingsManager to store the actual configuration on
 * disk. Settings changes from the configuration dialog are converted instantly
 * to SettingsManager key and value pairs. Read only access to the
 * SettingsManager is given by get_* and is_* methods, several signals are
 * provided for notification of changes.
 */

class ApplicationSettings : 
  public RegisteredSingleton<ApplicationSettings, 90>, 
  public Settings 
{
  public:
    GenericSetting<double> scorefade;
    GenericSetting<bool> scorevisible;
    GenericSetting<bool> scorecolumns;
    GenericSetting<unsigned> scorerepeats;
    GenericSetting<unsigned> scoreexpires;
    GenericSetting<bool> scoreomit;
    GenericSetting<unsigned> trainingmode;
    GenericSetting<bool> timerannoying;
    GenericSetting<unsigned> sizewidth;
    GenericSetting<unsigned> sizeheight;
    GenericSetting<bool> sizestate;
    GenericSetting<bool> voicedetails;
    GenericSetting<bool> userfit;
    GenericSetting<bool> reffit;
    GenericSetting<unsigned> askmode;
    GenericSetting<bool> presentsolution;
  
    MRUSetting lastfiles;
    
  public:
    ~ApplicationSettings();
    ApplicationSettings();

    //*** Training page
  
    // Repeat count
    unsigned get_repeat_num() const;
    unsigned get_repeat_wrong_num() const;
    sigc::signal<void> signal_repeat_num();

    // Buttons 
    unsigned get_buttons_num() const;
    bool is_buttons_sorted() const;
    bool is_buttons_same_type() const;
    sigc::signal<void> signal_buttons();

    //*** Defaults page
    
    // Languages
    std::string get_first_default() const;
    std::string get_second_default() const;
    sigc::signal<void> signal_languages_default();

    // Training mode
    bool is_audio_default() const;
    bool is_written_default() const;
    sigc::signal<void> signal_mode_default();

    bool is_use_timer_default() const;
    sigc::signal<void> signal_use_timer_default();
    
    //*** Stressbase page
    std::string get_stressbase_base (const std::string& language) const;

    //*** Synthesis page

    // Timeout
    unsigned get_synthesis_timeout() const;
    bool is_synthesis_timeout() const;
    sigc::signal<void> signal_synthesis_timeout();

    // Programs
    std::string get_synthesis_program (const std::string& language) const;
    std::string get_synthesis_params (const std::string& language) const;

    //*** Localized page
    std::string get_localized_or (const std::string& language) const;

    //*** Timer page
    
    // Timer stuff
    bool is_use_timer_vocab_limit() const;
    bool is_use_timer_answer_limit() const;
    unsigned get_timer_vocab_num() const;
    unsigned get_timer_interval() const;
    sigc::signal<void> signal_timer();

    //*** Misc page

    // Font selection
    std::string get_presfont() const;
    bool is_presfont() const;
    sigc::signal<void> signal_presfont();

    // Default languages
    template <class iter> void get_default_languages (iter insert) const;
    sigc::signal<void> signal_default_languages();

  private:
    // here we have:
    //   the UI handler
    //   the set and update functions 
    //   the GTK variables
    //   the notify signal
  
    void on_default_languages_changed();
    void set_default_languages (const std::string& value);
    void update_default_languages();
    Gtk::Entry *deflang;
    sigc::signal<void> signal_default_languages_v;

    void on_presfont_name_changed();
    void set_presfont_name (const std::string& value);
    void update_presfont_name();
    Gtk::FontButton *presfont;

    void on_presfont_enable_changed();
    void set_presfont_enable (const bool value);
    void update_presfont_enable();
    Gtk::CheckButton *presenable;

    sigc::signal<void> signal_presfont_v;

    void on_repeat_num_changed();
    void set_repeat_num (const unsigned value);
    void update_repeat_num();
    Gtk::SpinButton *timesspin;

    void on_repeat_wrong_num_changed();
    void set_repeat_wrong_num (const unsigned value);
    void update_repeat_wrong_num();
    Gtk::SpinButton *timeswrongspin;

    sigc::signal<void> signal_repeat_num_v;

    void on_buttons_sorted_changed();
    void set_buttons_sorted (const bool value);
    void update_buttons_sorted();
    Gtk::CheckButton *buttonssorted;

    void on_buttons_same_type_changed();
    void set_buttons_same_type (const bool value);
    void update_buttons_same_type();
    Gtk::CheckButton *prefersame;

    void on_buttons_num_changed();
    void set_buttons_num (const unsigned value);
    void update_buttons_num();
    Gtk::SpinButton *buttonsnum;
    
    sigc::signal<void> signal_buttons_v;

    void on_language_default_changed (const int which);
    void set_first_default (const std::string& value);
    void set_second_default (const std::string& value);
    void update_first_default();
    void update_second_default();
    Gtk::Entry *firstlanguage_default;
    Gtk::Entry *secondlanguage_default;
    sigc::signal<void> signal_languages_default_v;

    void on_presentwritten_default_changed();
    void set_written_default (const bool value);
    void update_written_default();
    Gtk::CheckButton *presentwritten_default;
    
    void on_presentaudio_default_changed();
    void set_audio_default (const bool value);
    void update_audio_default();
    Gtk::CheckButton *presentaudio_default;

    void on_use_timer_default_changed();
    void set_use_timer_default (const bool value);
    void update_use_timer_default();
    Gtk::CheckButton *usetimer_default;
    sigc::signal<void> signal_use_timer_default_v;

    sigc::signal<void> signal_mode_default_v;

    void on_use_timer_vocab_limit_changed();
    void set_use_timer_vocab_limit (const bool value);
    void update_use_timer_vocab_limit();
    Gtk::CheckButton *timerusevlimit;

    void on_use_timer_answer_limit_changed();
    void set_use_timer_answer_limit (const bool value);
    void update_use_timer_answer_limit();
    Gtk::CheckButton *timerlimitanswers;

    void on_timer_interval_changed();
    void set_timer_interval (const unsigned value);
    void update_timer_interval();
    Gtk::SpinButton *timerinterval;

    void on_timer_vocab_num_changed();
    void set_timer_vocab_num (const unsigned value);
    void update_timer_vocab_num();
    Gtk::SpinButton *timervocabnum;

    sigc::signal<void> signal_timer_v;

    void stressbase_add (const std::string& language, const std::string& path);
    void on_stressbase_add_base();
    void on_stressbase_delete_base();
    void on_stressbase_select_base();
    void on_stressbase_changed (const Gtk::TreePath& path, const Gtk::TreeIter& iter); 
    void update_stressbase();
    void rebuild_stressbase();
    Gtk::Button *stressadd;
    Gtk::Button *stressdelete;
    Gtk::Button *stressselect;
    PopupTreeView *stresslist;
    StressListModelColumns stress_columns;
    Glib::RefPtr<Gtk::ListStore> stress_store;

    void synthesis_add (const std::string& language, const std::string& path, const std::string& params);
    void on_synthesis_add_program();
    void on_synthesis_delete_program();
    void on_synthesis_select_program();
    void on_synthesis_changed (const Gtk::TreePath& path, const Gtk::TreeIter& iter); 
    void update_synthesis();
    void rebuild_synthesis();
    Gtk::Button *synthadd;
    Gtk::Button *synthdelete;
    Gtk::Button *synthselect;
    PopupTreeView *synthlist;
    SynthListModelColumns synth_columns;
    Glib::RefPtr<Gtk::ListStore> synth_store;

    void on_synthesis_timeout_changed();
    void set_synthesis_timeout (const unsigned value);
    void update_synthesis_timeout();
    Gtk::SpinButton *synthtimeout;

    void on_synthesis_timeout_enable_changed();
    void set_synthesis_timeout_enable (const bool value);
    void update_synthesis_timeout_enable();
    Gtk::CheckButton *synthlimit;

    sigc::signal<void> signal_synthesis_timeout_v;

    void localized_add (const std::string& language, const std::string& word_or);
    void on_localized_add();
    void on_localized_delete();
    void on_localized_changed (const Gtk::TreePath& path, const Gtk::TreeIter& iter); 
    void update_localized();
    void rebuild_localized();
    Gtk::Button *wordadd;
    Gtk::Button *worddelete;
    PopupTreeView *wordlist;
    WordListModelColumns word_columns;
    Glib::RefPtr<Gtk::ListStore> word_store;
};

/** Return the default languages. The data is inserted at the iterator given at \c insert. 
 * You have to make sure that enough space is available in the container, or provide e.g. an 
 * std::insert_iterator.
 * \param insert An iterator to a std::string or Glib::ustring
 */

template <class iter> 
void ApplicationSettings::get_default_languages (iter insert) const
{
    std::string value = manager->get_string ("new", "languages");
    typedef boost::char_separator<char> separator;
    typedef boost::tokenizer<separator, Glib::ustring::const_iterator, Glib::ustring> tokenizer;
    const separator sep (" ,;");
    const tokenizer tokens (value, sep);
    std::copy (tokens.begin(), tokens.end(), insert);
}

#endif /* _APPLICATIONSETTINGS_H */
