/***************************************************************************
 *            BinaryStressBase.cc
 *
 *  Mon Dec  6 19:43:17 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/** \file
 * Binary StressBase class.
 */

#include "BinaryStressBase.h"

#include "debug.h"

#include <iostream>
#include <algorithm>

/** Initialize without a stressbase. It shouldn't be used in this state.
 */

BinaryStressBase::BinaryStressBase() : 
  size (0)
{
}

/** Initialize and load a stressbase.
 * \param file Filename and path of the binary stressbase
 */

BinaryStressBase::BinaryStressBase (const std::string& file) :
  size (0)
{
    load (file);
}
     
int BinaryStressBase::get_idx (const std::string& name) const {
    long long unsigned md5 = get_md5 (name);
    long long unsigned *idx = std::lower_bound (keys.get(), keys.get() + size, md5);
    if (idx != keys.get() + size && *idx == md5) {
        return idx - keys.get();
    } else {
	debug_log ("Word '%1' was not found in stressbase", name);
        return -1;
    }
}
	
/** Get the accent position for a particular word.
 * \param name Word in UTF-8 encoding
 * \return position of accent (starting with 1) or -1 if not found
 */

int BinaryStressBase::get_accent (const std::string& name) const {
    int idx = get_idx (name);
    if (idx != -1)
      return accents[idx];
    else
      return -1;
}

/** Load a new stressbase. If one is already loaded, the contents is replaced.
 * \param file Filename and path of the binary stressbase
 */

void BinaryStressBase::load (const std::string& file) 
{
    FILE *f;
    if ((f = fopen (file.c_str(), "rb")) != NULL) {
	unsigned size;
	fread (&size, sizeof (size), 1, f);
	allocate (size);
	fread (keys.get(), sizeof (keys[0]), size, f);
	fread (accents.get(), sizeof (accents[0]), size, f);
	fclose (f);
    }
}
	
void BinaryStressBase::allocate (const int new_size) 
{
    keys.reset (new long long unsigned [new_size]);
    accents.reset (new unsigned char [new_size]);
    size = new_size;
}
