/***************************************************************************
 *            CopyPaste.cc
 *
 *  Sat Jul  9 15:53:33 2005
 *  Copyright  2005  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "CopyPaste.h"
#include "VocabTest.h"

#include <gtkmm/clipboard.h>

const std::string CopyPaste::synth ("#(synthetic)# ");

/** Initialize a new Copy and Paste helper. You can keep one around all time or 
 * instantiate only as needed.
 */
CopyPaste::CopyPaste (const boost::shared_ptr<VocabTest> vocab_test) :
  vocab_test (vocab_test),
  clipboard (Gtk::Clipboard::get())
{
}

void CopyPaste::copy_structure (std::vector<std::pair<int, int> >& data)
{
    init_copy();
    VocabTest::normalize_indizes (data);
    for (std::vector<std::pair<int, int> >::const_iterator i = data.begin(); i != data.end(); ++i) {
	if (i->second == -1)
	  append_chapter (i->first);
	else
	  append_section (i->first, i->second);

    }
    transfer_to_clipboard();
    done();
}

void CopyPaste::copy_vocabs (const int chapter, const int section, 
  const std::vector<int>& data)
{
    init_copy();
    for (std::vector<int>::const_iterator i = data.begin(); i != data.end(); ++i) {
        append_vocab (chapter, section, *i);
    }
    transfer_to_clipboard();
    done();
}

void CopyPaste::paste_structure (const int chapter, const int section, const int vocab) 
{
    paste (chapter, section, vocab);
}

void CopyPaste::paste_vocabs (const int chapter, const int section, const int vocab) 
{
    paste (chapter, section, vocab);
}

/** Handles both structure and vocablist insertions with no difference.
 */
void CopyPaste::paste (const int chapter, const int section, const int vocab)
{
    init_paste();
    if (transfer_from_clipboard()) {
	// Correct impossible insertion points
	if (chapter == -1)
	  type = CHAPTERS;
	if (section == -1 && type > SECTIONS)
	  type = SECTIONS;
	switch (type) {
	  case CHAPTERS:
	    last_target_chapter = chapter;
	    for (int i = 0; i < source->get_chapter_num(); ++i)
	      append_chapter (i);
	    break;
	  case SECTIONS:
	    last_target_chapter = chapter;
	    last_target_section = section;
	    for (int i = 0; i < source->get_chapter_num(); ++i)
	      for (int j = 0; j < source->get_section_num (i); ++j)
	      append_section_internal (i, j);
	    break;
	  case VOCABS:
	    last_target_chapter = chapter;
	    last_target_section = section;
	    last_target_vocab = vocab;
	    for (int i = 0; i < source->get_chapter_num(); ++i)
	      for (int j = 0; j < source->get_section_num (i); ++j)
	      for (int k = 0; k < source->get_vocab_num (i, j); ++k)
	      append_vocab_internal (i, j, k);
	    break;
	  case NOTHING:
	    break;
	}
    };
    done();
}
    
void CopyPaste::init_copy() 
{
    copy.reset (new VocabTest());
    languages = vocab_test->get_languages();
    for (std::vector<std::string>::const_iterator i = languages.begin(); 
      i != languages.end(); ++i)
      copy->add_language (*i);	
    source = vocab_test;
    target = copy;
    init();
}

void CopyPaste::init_paste() 
{
    copy.reset (new VocabTest());
    source = copy;
    target = vocab_test;
    init();
}

void CopyPaste::init() 
{
    last_source_chapter = -1;
    last_source_section = -1;
    
    last_target_chapter = -1;
    last_target_section = -1;
    last_target_vocab = -1;
    
    type = NOTHING;
}

void CopyPaste::transfer_to_clipboard() 
{
    std::ostringstream stream;
    target->write_xml (stream);
    clipboard->set_text (stream.str());
}

bool CopyPaste::transfer_from_clipboard() 
{
    std::string data = clipboard->wait_for_text();
    try {
	source->read_xml_data (data);
	scan_for_type();
	languages = source->get_languages();
	return true;
    } catch (...) {
	return false;
    }
}

/** Get the highest non-synthetic member.
 */
void CopyPaste::scan_for_type() 
{
    for (int chapter = 0; chapter < source->get_chapter_num(); ++chapter) {
	std::string name = source->get_chapter_name (chapter);
	if (! is_synthetic (name)) {
	    type = CHAPTERS;
	    return;
	}
	if (type > SECTIONS)
          for (int section = 0; section < source->get_section_num (chapter); ++section) {
	    name = source->get_section_name (chapter, section);
	    if (! is_synthetic (name)) {
		type = SECTIONS;
		break;
	    }
	    if (type > VOCABS && source->get_vocab_num (chapter, section) > 0)
	      type = VOCABS;
	}
    }
}

bool CopyPaste::is_synthetic (const std::string& name) const
{
    return name.compare (0, synth.size(), synth) == 0;
}

std::string CopyPaste::strip_synthetic (const std::string& name) const
{
    if (is_synthetic (name))
      return name.substr (synth.size());
    else
      return name;
}

void CopyPaste::done() 
{
    copy.reset();
}

/** Append a new chapter and all sections and vocabularies in it. 
 */
void CopyPaste::append_chapter (const int chapter) 
{
    create_chapter (chapter);
    for (int i = 0; i < source->get_section_num (chapter); ++i)
      append_section_internal (chapter, i);
}

/** Create a new chapter. Does not copy sections and vocabs.
 */
void CopyPaste::create_chapter (const int chapter, const bool synthetic) 
{
    last_source_chapter = chapter;
    last_source_section = -1;
    last_target_chapter++;
    last_target_section = -1;
    last_target_vocab = -1;
    target->insert_chapter (last_target_chapter);
    target->set_chapter_name (last_target_chapter, 
      (synthetic ? synth : "") + strip_synthetic (source->get_chapter_name (chapter)));
}

/** Append a new section and all vocabularies in it. 
 */
void CopyPaste::append_section_internal (const int chapter, const int section) 
{
    create_section (chapter, section);
    for (int i = 0; i < source->get_vocab_num (chapter, section); ++i)
      append_vocab_internal (chapter, section, i);
}

/** Append a new vocabulary, create chapter and section as necessary. 
 */
void CopyPaste::append_section (const int chapter, const int section) 
{
    if (chapter != last_source_chapter)
      create_chapter (chapter, true);
    append_section_internal (chapter, section);
}

/** Create a new section. Does not copy vocabs.
 */
void CopyPaste::create_section (const int chapter, const int section, const bool synthetic) 
{
    last_source_chapter = chapter;
    last_source_section = section;
    last_target_section++;
    last_target_vocab = -1;
    target->insert_section (last_target_chapter, last_target_section);
    target->set_section_name (last_target_chapter, last_target_section, 
      (synthetic ? synth : "") + strip_synthetic (source->get_section_name (chapter, section)));
}

/** Append a new vocabulary. 
 */
void CopyPaste::append_vocab_internal (const int chapter, const int section, 
  const int vocab) 
{
    last_target_vocab++;
    target->insert_vocab (last_target_chapter, last_target_section, last_target_vocab);
    target->set_type (last_target_chapter, last_target_section, last_target_vocab, 
      source->get_type (chapter, section, vocab));
    for (std::vector<std::string>::const_iterator i = languages.begin(); 
      i != languages.end(); ++i) {
	std::string translation = source->get_vocab (chapter, section, vocab, *i);
	if (translation != "")
          target->set_vocab (last_target_chapter, last_target_section, last_target_vocab, 
	    *i, translation);
    }
    target->set_id (last_target_chapter, last_target_section, last_target_vocab, 
      source->get_id (chapter, section, vocab));
}

/** Append a new vocabulary, create chapter and section as necessary. 
 */
void CopyPaste::append_vocab (const int chapter, const int section, const int vocab) 
{
    if (chapter != last_source_chapter)
      create_chapter (chapter, true);
    if (section != last_source_section)
      create_section (chapter, section, true);
    append_vocab_internal (chapter, section, vocab);
}
