/***************************************************************************
 *            EditTreeView.cc
 *
 *  Mon Oct 25 17:50:10 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/** \file
 * An editable Gtk::TreeView.
 */

#include "support.h"
#include "EditTreeView.h"

#include "debug.h"

#include <gdk/gdkkeysyms.h>
#include <gtkmm/treepath.h>
#include <gtkmm/main.h>
#include <gtkmm/window.h>

#include <boost/utility.hpp>

#include <iostream>
#include <algorithm>

/** Initialize with libglade.
 */

EditTreeView::EditTreeView (BaseObjectType* cobject, const Glib::RefPtr<Gnome::Glade::Xml>& refGlade)
  : Glib::ObjectBase (typeid (EditTreeView)), 
    PopupTreeView (cobject, refGlade),
    xml (refGlade)
{
    old_idx = IDX_INIT;
    newly_created = false;
}

void EditTreeView::on_editing_canceled()
{
    debug_log ("Editing was canceled");
    disconnect_signals();
    if (cursor_visible (this) && newly_created && old_idx == get_next_editable_column (-1)) {
	Gtk::TreePath p; Gtk::TreeViewColumn* c;
	get_cursor (p, c);
	signal_delete_row_path_v.emit (p, c);
	p.prev();
	set_cursor (p);
    }
    old_idx = IDX_INIT;
    newly_created = false;
}
    
// This uses pretty ugly code to jump to the next column for editing. 
// The direct way doesn't work. 
void EditTreeView::on_editing_done (const Glib::ustring& path, const Glib::ustring& new_text)
{
    debug_log ("Editing was finished");
    disconnect_signals();

    if (! wasenter) {
	old_idx = IDX_INIT;
	newly_created = false;
	return;
    }
    const int new_idx = get_next_editable_column (old_idx);
    if (new_idx > -1)
      simulate_key (new_idx);
    else
      simulate_key (IDX_TRY_NEW_ROW);
}

int EditTreeView::get_next_editable_column (const int idx) const
{
    typedef Glib::ListHandle<const Gtk::TreeViewColumn*> ColumnList;
    ColumnList cs = get_columns();
    int new_idx = idx + 1;
    ColumnList::const_iterator i = boost::next (cs.begin(), new_idx);
    for (; i != cs.end(); ++i, ++new_idx) {
	if (is_editable_column (*i))
	  return new_idx;
    }
    return -1;
}

bool EditTreeView::is_editable_column (const Gtk::TreeViewColumn* c) 
{
    if (! c) 
      return false;
    const Gtk::CellRendererText* renderer = dynamic_cast<const Gtk::CellRendererText*> 
      (c->get_first_cell_renderer());
    return (renderer && renderer->property_editable());
}

bool EditTreeView::on_key_press_event (GdkEventKey* event)
{
    switch (event->keyval) {
      case GDK_F2:
	if (old_idx == IDX_INIT) {
	    // Start F2 insert mode
	    old_idx = IDX_TRY_EDIT;
	    old_event = *event;
	} else if (simulated_key) {
	    // Simulate an F2 key press
	    simulated_key = false;
	} else {
	    // Stop F2 insert mode
	    disconnect_signals();
	    old_idx = IDX_INIT;
	    break;
	}
	connect_signals();
	break;
	
      default:
	return PopupTreeView::on_key_press_event (event);
    }

    return true;
}

// And this is even more ugly using a key_snooper to get whether the done signal was because 
// of enter or a lost focus; we can't get the edit widget for signal_event
int EditTreeView::on_key_snoop (Gtk::Widget * where, GdkEventKey * event) 
{
    if (event->type == GDK_KEY_PRESS) switch (event->keyval) {
      case GDK_Return:
      case GDK_KP_Enter:
      case GDK_ISO_Enter:
      case GDK_3270_Enter:
	debug_log ("Widget of enter press was %1 (%2)", where->get_name(), where);
        // If it was no context menu but a window, then we suspect it was a history popup
        // Testing: check for switching to the next field on focus-out after 
        // e.g. Enter in the context menu, in the history window after 
        // selecting/not selecting an entry ...
        if (dynamic_cast<Gtk::Window*> (where) != NULL)
	  wasenter = true;
        break;
    }
    return false;
}

    
void EditTreeView::simulate_key (const int new_idx, const bool created)
{
    old_idx = new_idx;
    newly_created = created;
    simulated_key = true;
    Gdk::Event e ((GdkEvent*) &old_event);
    e.put();
}

// This needs some state machine style cleanup
void EditTreeView::connect_signals()
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    if (! cursor_visible (this)) {
	// seems like a bug: it is still possible that we have a selection
	std::vector<Gtk::TreePath> sel = get_selection()->get_selected_rows();
	if (sel.size())
          set_cursor (*sel.begin());
    }
    int first_column = get_next_editable_column (-1);
    if (first_column == -1) 
      first_column = 0;
    get_cursor (p, c);
    if (! cursor_visible (this) && old_idx == IDX_TRY_EDIT || old_idx == IDX_TRY_NEW_ROW) {
	if (cursor_visible (this) && old_idx == IDX_TRY_NEW_ROW)
          p.next();
	else
	  p = Gtk::TreePath ("0");
	signal_create_row_path_v.emit (p, c);
	set_cursor (p, *get_column (first_column), false);
	simulate_key (IDX_TRY, true);
    } else if (!p.size()) {
	old_idx = IDX_INIT;
    } else {
	if (old_idx == IDX_TRY || old_idx == IDX_TRY_EDIT) {
	    if (c) {
		Glib::ListHandle<Gtk::TreeViewColumn*> cs = get_columns();
	        old_idx = std::distance (cs.begin(), std::find (cs.begin(), cs.end(), c));
		if (! is_editable_column (c)) {
		    int new_idx = get_next_editable_column (old_idx);
		    if (new_idx == -1)
		    {/*TODO*/}// if the non-editable fields are at the end
		    else
		      old_idx = new_idx;
		    c = get_column (old_idx);
		}
	    } else {
		old_idx = first_column;
	        c = get_column (old_idx);
	    }
	} else {
	    c = get_column (old_idx);
	}
	set_cursor (p, *c, true);
  
	wasenter = false;
	
	old_editing_done = dynamic_cast<Gtk::CellRendererText*> 
	  (c->get_first_cell_renderer())->signal_edited().connect 
	  (sigc::mem_fun (*this, &EditTreeView::on_editing_done));
	old_editing_canceled = dynamic_cast<Gtk::CellRenderer*> 
	  (c->get_first_cell_renderer())->signal_editing_canceled().connect 
	  (sigc::mem_fun (*this, &EditTreeView::on_editing_canceled));
	old_key_snoop = Gtk::Main::instance()->signal_key_snooper().connect 
	  (sigc::mem_fun (*this, &EditTreeView::on_key_snoop));
	old_key_snoop = Gtk::Main::instance()->signal_key_snooper().connect 
	  (sigc::mem_fun (*this, &EditTreeView::on_key_snoop));
    }
}

void EditTreeView::disconnect_signals()
{
    if (old_editing_canceled) 
      old_editing_canceled.disconnect();
    if (old_editing_done)
      old_editing_done.disconnect();
    if (old_key_snoop)
      old_key_snoop.disconnect();
}

/** Connect to this signal to provide the insert functionality.
 * It is called when a new row is needed.
 */

EditTreeView::type_signal_create_row_path EditTreeView::signal_create_row_path()
{
    return signal_create_row_path_v;
}

/** Connect to this signal to provide the delete functionality.
 * It is called when a row needs to be deleted.
 */

EditTreeView::type_signal_delete_row_path EditTreeView::signal_delete_row_path()
{
    return signal_delete_row_path_v;
}
