/***************************************************************************
 *            MainSession.cc
 *
 *  Sat Dec  4 14:05:37 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "debug.h"

#include "MainSession.h"
#include "SessionFeeder.h"
#include "GlobalXML.h"
#include "MainWindow.h"
#include "VocabCache.h"

#include "langsupport.h"

#include <gdk/gdkkeysyms.h>

MainSession::MainSession (MainWindow* mw, const boost::shared_ptr<VocabTest> vocab_test) :
  mw (mw),
  xml (GlobalXML::get().get_glade()), 
  vocab_test (vocab_test),
  settings (ApplicationSettings::get()),
  tsettings (TransientSettings::get())
{
    debug_log ("Loading up main session");
 
    iconify_connection = mw->signal_window_state_event().connect (sigc::mem_fun (this, &MainSession::on_window_state_changed));
}

MainSession::~MainSession()
{
    iconify_connection.disconnect();
    clear_timer();
    be_nice();
}

void MainSession::start() 
{
    // Set up the vocabulary feeder
    SessionFeederPtr feeder = SessionFeeder::create (vocab_test);
    feeder->set_languages (tsettings.get_first(), tsettings.get_second());
    feeder->set_ors (settings.get_localized_or (tsettings.get_first()), 
      settings.get_localized_or (tsettings.get_second()));
    feeder->set_mode (
      settings.trainingmode.get_value() == 1,
      settings.scoreomit.get_value(),
      settings.get_repeat_num(), 
      settings.get_repeat_wrong_num());
    feeder->set_alternatives (settings.get_buttons_num(), 
      settings.is_buttons_sorted(), settings.is_buttons_same_type());

    std::vector<std::pair<int, int> > data;
    mw->get_selected_structure (data);

    for (std::vector<std::pair<int, int> >::const_iterator i = data.begin();
      i != data.end(); ++i) {
	if (i->second != -1) 
	  feeder->add_section (i->first, i->second);
	else
	  feeder->add_chapter (i->first);
    }

    // Set up the training session
    session = VocabSessionPtr (new VocabSession (feeder));
    session->set_mode (settings.trainingmode.get_value() == 1);
    session->set_interval (settings.get_timer_vocab_num());
    update_timer_mode();
    
    // Set up the vocabulary cache
    VocabCache& cache = VocabCache::get();
    cache.set_expires (settings.scoreexpires.get_value());
    cache.set_repeats (settings.scorerepeats.get_value());
    
    session->start();
    
    right_answers_left = settings.get_timer_vocab_num();

    if (tsettings.is_buttons())
      mw->create_pick_table (settings.get_buttons_num());
    
    if (session->get_vocabs_left()) {
	tsettings.set_resultsshown (false);
	tsettings.set_testmode (true);
	present_vocabulary();
	be_annoying();
    }
    mw->select_input();    
};

void MainSession::pause()
{
    mw->iconify();
    be_nice();
}

/** Enable annoying mode.
 * Makes the window fullscreen, sticky, above all others and grabs the
 * keyboard. 
 */

void MainSession::be_annoying()
{
    if (! settings.timerannoying.get_value() || ! tsettings.usetimer.get_value()) return;
	
    mw->fullscreen();
    mw->stick();
    mw->set_keep_above(); 
    
    // We wait 100ms
    grabkeyb_connection = Glib::signal_timeout().connect (sigc::mem_fun (this, &MainSession::lock_keyb), 100);
}

/** Lock the keyboard.
 * \return Whether we have to try again because it didn't work
 */

bool MainSession::lock_keyb() const
{
    if (mw->get_window()->keyboard_grab (true, GDK_CURRENT_TIME) != Gdk::GRAB_SUCCESS) {
	debug_log ("Grabbing keyboard failed");
	return true;
    } else {
	debug_log ("Grabbing keyboard succeeded");
	return false;
    }	
}

/** Disable annoying mode.
 * Reverts the changes from be_annoying().
 */

void MainSession::be_nice()
{
    if (! settings.timerannoying.get_value()) return;
	
    mw->unfullscreen();
    mw->unstick();
    mw->set_keep_above (false);
    
    debug_log ("Ungrabbing keyboard");
    grabkeyb_connection.disconnect();
    mw->get_window()->keyboard_ungrab (GDK_CURRENT_TIME);
}
    

void MainSession::awake()
{
    // Reset right answer counter to initial value
    right_answers_left = settings.get_timer_vocab_num();
    session->feed();
    
    mw->present();
    mw->raise();
    be_annoying();
}

bool MainSession::on_window_state_changed (GdkEventWindowState* event)
{
    // We don't care without being an active session
    if (! tsettings.is_testmode()) return false;
    
    if (event->changed_mask & Gdk::WINDOW_STATE_ICONIFIED) {
	if (! (event->new_window_state & Gdk::WINDOW_STATE_ICONIFIED)) {
	    debug_log ("Coming back from sleep.");
	    clear_timer();
	} else {
	    debug_log ("Going to sleep.");
	    if (tsettings.usetimer.get_value()) {
		set_timer (settings.get_timer_interval() * 60);
	    }
	}
    }
    return false;
}

bool MainSession::on_timeout()
{
    awake();
    return false;
}

void MainSession::finish() 
{
    be_nice();
    clear_timer();
    
    tsettings.set_testmode (false);
    tsettings.set_resultsshown (true);
    present_results();
}

void MainSession::reset() 
{
    be_nice();
    clear_timer();
    
    tsettings.set_testmode (false);
    tsettings.set_resultsshown (false);
}

void MainSession::set_timer (const unsigned seconds)
{
    clear_timer();
    debug_log ("Enabling timer in %1 seconds.", seconds);
    timeout_connection = Glib::signal_timeout().connect (sigc::mem_fun (this, &MainSession::on_timeout), seconds * 1000);
}

void MainSession::clear_timer()
{
    if (timeout_connection) {
	debug_log ("Disconnecting timer.");
	timeout_connection.disconnect();
    }
}

void MainSession::update_timer_mode()
{
    session->set_feed_continuous (
      ! tsettings.usetimer.get_value() || 
      ! settings.is_use_timer_vocab_limit() || 
      settings.is_use_timer_answer_limit());
}

void MainSession::present_results() 
{
    mw->clear_result_list();
    session->prepare_stats();
    mw->add_result_list_line ("Total vocabularies", strmanip::ucompose ("%1", session->get_stats_vocabs()));
    unsigned sum = session->get_stats_right() + session->get_stats_wrong();
    if (sum) {
	mw->add_result_list_line ("Guesses", strmanip::ucompose ("%1", sum));
	mw->add_result_list_line ("Right guesses", strmanip::ucompose ("%1 = %2%%", 
	  session->get_stats_right(), (100 * session->get_stats_right()) / sum));
	mw->add_result_list_line ("Wrong guesses", strmanip::ucompose ("%1 = %2%%",
	  session->get_stats_wrong(), (100 * session->get_stats_wrong()) / sum));
    }
}

void MainSession::present_vocabulary (const bool silent) 
{
    present_vocabulary_written();
    
    if (tsettings.is_audio() && ! silent) present_vocabulary_audio (1);
    
    if (tsettings.is_buttons()) present_vocabulary_buttons();
}

void MainSession::present_vocabulary_written() 
{
    if (tsettings.is_written()) 
      mw->present_vocabulary_written (session->get()->get_question());
    else
      mw->present_vocabulary_written (_("Listen to the music ;-)"));
}   

void MainSession::present_vocabulary_buttons() 
{
    for (unsigned i = 0; i < settings.get_buttons_num(); i++)
      mw->set_question_button (i, session->get()->get_alternative (i));
}

void MainSession::present_vocabulary_audio (const unsigned wait) 
{
    mw->present_vocabulary_audio (wait, session->get()->get_question_audio(), tsettings.get_first());
}

void MainSession::present_solution_audio (const unsigned wait) 
{
    mw->present_solution_audio (wait, session->get()->get_solution_audio(), tsettings.get_second());
}

void MainSession::present_vocabulary_audio_solution (const unsigned wait) 
{
    const SessionVocabPtr old = session->get_old();
    if (old)
      mw->present_vocabulary_audio (wait, old->get_solution_audio(), tsettings.get_second());
}

void MainSession::present_solution (const bool solution, const bool right) 
{
    std::string s;
    bool audio = true;
    if (solution && ! right) {
	s = strmanip::ucompose (_("The right translation of \"%1\" would have been: %2"),
	  session->get()->get_question_short(), session->get()->get_solution());
    } else if (solution && right) {
	s = strmanip::ucompose (_("Right. The full translation of \"%1\" was: %2"),
	  session->get()->get_question_short(), session->get()->get_solution());
    } else if (! solution && ! right) {
	s = _("Wrong. Try again!");
	audio = false;
    }
    
    mw->present_solution_written (s, audio);
    if (tsettings.presentsolution.get_value() && audio) 
      present_solution_audio (0);
	
    mw->select_input();
}

void MainSession::handle_guess (const bool right) 
{
    if (right) {
	if (
	  tsettings.usetimer.get_value() && 
	  settings.is_use_timer_vocab_limit() &&
	  settings.is_use_timer_answer_limit())
	  --right_answers_left;
	present_solution (true, true);
	session->get()->answered (true);
	session->defer (false);
	advance();
    } else {
	present_solution (false, false);
	session->get()->answered (false);
    }
}

void MainSession::advance_and_show() 
{
    present_solution (true, false);
    session->get()->unanswered();
    session->defer (true);
    advance();
};

bool MainSession::on_guess_key_press_event (GdkEvent* event)
{
    if (event->type == GDK_KEY_PRESS) {
	if (event->key.keyval == GDK_Escape) {
	    advance_and_show();
	    return true;
	};
    }
    return false;
}

void MainSession::advance() 
{
    // update the feeding behaviour if somebody has changed the timer setting
    update_timer_mode();
    if (! session->advance()) {
	// Try it again, maybe its just an interval sign
	if (session->advance()) {
	    present_vocabulary (true);
	    pause();
	} else finish();
    } else {
	if (
	  tsettings.usetimer.get_value() && 
	  settings.is_use_timer_vocab_limit() && 
	  settings.is_use_timer_answer_limit() && 
	  ! right_answers_left) {
	    present_vocabulary (true);
	    pause();
	} else present_vocabulary();
    }
};

void MainSession::handle_input_guess()
{
    handle_guess (session->get()->check (mw->get_guess_text()));
}

void MainSession::handle_button_guess (const unsigned idx)
{
    handle_guess (session->get()->check (idx));
}

void MainSession::hide_results()
{
    tsettings.set_resultsshown (false);
};
