/***************************************************************************
 *            MainWindow.cc
 *
 *  Mon Oct 18 12:31:58 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "MainWindow.h"
#include "VocabTreeModel.h"
#include "VocabTypesModel.h"
#include "VocabFormatter.h"
#include "VocabCache.h"
#include "CellRendererCompletion.h"
#include "langsupport.h"
#include "support.h"
#include "exception.h"
#include "IconStore.h"
#include "SynthesisInterface.h"
#include "SynthesisAnalysis.h"
#include "MarkerSonogram.h"
#include "CopyPaste.h"
#include "CmdLine.h"
#include "VocabScoreDiagram.h"
#include "TrayIcon.h"

#include "debug.h"
#include "ucompose.hpp"
#include "functions.h"

#include "config.h"

#include <gtkmm/notebook.h>
#include <gtkmm/toggletoolbutton.h>
#include <gtkmm/table.h>
#include <gtkmm/aboutdialog.h>
#include <gtkmm/textview.h>
#include <gtkmm/window.h>
#include <gtkmm/toolbar.h>
#include <gtkmm/scrollbar.h>
#include <gtkmm/menubar.h>
#include <gtkmm/liststore.h>
#include <gtkmm/dialog.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/entry.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/stock.h>
#include <gtkmm/filefilter.h>
#include <gtkmm/treeviewcolumn.h>
#include <gtkmm/main.h>
#include <gtkmm/sizegroup.h>
#include <gtkmm/curve.h>

#include <gtk/gtk.h>

#include <gdk/gdkkeysyms.h>

#include <boost/array.hpp>

#include <iostream>

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

boost::array<unsigned, 4> pos_red_array =   {{0,   0, 255, 255}};
boost::array<unsigned, 4> pos_green_array = {{0,   0,   0, 255}};
boost::array<unsigned, 4> pos_blue_array =  {{0, 255,   0,   0}};

boost::array<func::PieceWiseLinear<unsigned, double, 4>, 3> pos_colors = {{
  func::PieceWiseLinear<unsigned, double, 4> (pos_red_array),
  func::PieceWiseLinear<unsigned, double, 4> (pos_green_array),
  func::PieceWiseLinear<unsigned, double, 4> (pos_blue_array)}};

/** Initialize main window.
 * Retrieves information about most of the various UI widgets 
 * used in most of the windows and dialogs.
 */
 
MainWindow::MainWindow (BaseObjectType* cobject, const Glib::RefPtr<Gnome::Glade::Xml>& refGlade)
  : Gtk::Window (cobject), 
    result_store (Gtk::ListStore::create (result_columns)),
    xml (refGlade), 
    vocab_test (new VocabTest()),
    settings (ApplicationSettings::get()),
    tsettings (TransientSettings::get()),
    main_session (this, vocab_test)
{
    debug_log ("Loading up MainWindow");

    IconStore::get(); // make sure icons are loaded
    
    debug_log ("Resizing");

   // Manually resize to the old size
    if (settings.sizewidth.get_value() > 0 && settings.sizeheight.get_value() > 0)
      resize (settings.sizewidth.get_value(), settings.sizeheight.get_value());
    if (settings.sizestate.get_value())
      maximize();
    
    debug_log ("Connecting settings handler");

    tsettings.signal_languages().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));
    tsettings.signal_mode().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));
    tsettings.usetimer.get_signal().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));

    settings.signal_presfont().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));
    settings.scorevisible.get_signal().connect (sigc::bind<bool> (sigc::mem_fun (*this, &MainWindow::create_edit_structure), true));
    settings.scorecolumns.get_signal().connect (sigc::bind<bool> (sigc::mem_fun (*this, &MainWindow::create_edit_structure), true));
    settings.scorefade.get_signal().connect (sigc::bind<bool> (sigc::mem_fun (*this, &MainWindow::create_edit_structure), true));
    
    settings.userfit.get_signal().connect (sigc::mem_fun (this, &MainWindow::on_fit_changed));
    settings.reffit.get_signal().connect (sigc::mem_fun (this, &MainWindow::on_fit_changed));
    
    settings.lastfiles.get_signal().connect (sigc::mem_fun (*this, &MainWindow::open));
    
    vocab_test->signal_on_modified().connect (sigc::mem_fun (*this, &MainWindow::on_modified));

    debug_log ("Getting Widgets");
    
    GET_GLADE (xml, menu_main);
    GET_GLADE (xml, tool_main);
    
    GET_GLADE (xml, menu_open);
    menu_open->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_open_activate));
    menu_open->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));

    GET_GLADE (xml, menu_save);
    menu_save->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_save_activate));

    GET_GLADE (xml, menu_saveas);
    menu_saveas->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_saveas_activate));

    GET_GLADE (xml, menu_close);
    menu_close->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_close_activate));
    menu_close->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));

    GET_GLADE (xml, menu_new); // Does the same like close
    menu_new->signal_activate().connect (sigc::mem_fun (*menu_close, &Gtk::MenuItem::activate));
    
    GET_GLADE (xml, menu_quit);
    menu_quit->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_quit_activate));

    GET_GLADE (xml, menu_cut);
    menu_cut->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_cut_activate));

    GET_GLADE (xml, menu_copy);
    menu_copy->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_copy_activate));

    GET_GLADE (xml, menu_paste);
    menu_paste->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_paste_activate));

    GET_GLADE (xml, menu_delete);
    menu_delete->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_delete_activate));

    GET_GLADE (xml, menu_settings);
    menu_settings->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_settings_activate));

    GET_GLADE (xml, menu_find);
    menu_find->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_find));

    GET_GLADE (xml, menu_fileproperties);
    menu_fileproperties->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_fileproperties_activate));

    GET_GLADE (xml, menu_addlang);
    menu_addlang->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_addlang_activate));

    GET_GLADE (xml, menu_about);
    menu_about->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_about_activate));

    GET_GLADE (xml, showprev);
    showprev->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::on_showprev_clicked));
    showprev->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));

    GET_GLADE (xml, shownext);
    shownext->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::on_shownext_clicked));
    shownext->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));

    GET_GLADE (xml, addvocab);
    addvocab->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::on_vocab_add));

    GET_GLADE (xml, delvocab);
    delvocab->signal_clicked().connect (sigc::bind (sigc::mem_fun (*this, &MainWindow::on_vocab_delete), true));
    delvocab->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));

    GET_GLADE (xml, structuremenu);
    GET_GLADE (xml, structuremenu_cut);
    structuremenu_cut->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_structure_cut));
    GET_GLADE (xml, structuremenu_copy);
    structuremenu_copy->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_structure_copy));
    GET_GLADE (xml, structuremenu_paste);
    structuremenu_paste->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_structure_paste));
    GET_GLADE (xml, structuremenu_delete);
    structuremenu_delete->signal_activate().connect (sigc::bind (sigc::mem_fun (*this, &MainWindow::on_structure_delete), true));
    GET_GLADE (xml, structuremenu_addchapter);
    structuremenu_addchapter->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_structure_addchapter));
    GET_GLADE (xml, structuremenu_addsection);
    structuremenu_addsection->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_structure_addsection));
    
    xml->get_widget_derived ("structuretree", structuretree);
    structuretree->signal_delete_row().connect (sigc::bind (sigc::mem_fun (*this, &MainWindow::on_structure_delete), true));
    structuretree->get_selection()->signal_changed().connect (sigc::bind<bool> (sigc::mem_fun (*this, &MainWindow::create_edit_structure), false));
    structuretree->get_selection()->signal_changed().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));
    structuretree->set_search_equal_func (sigc::mem_fun (*this, &MainWindow::on_search_equal));
    structuretree->set_popup (structuremenu);
    structuretree->get_selection()->set_mode (Gtk::SELECTION_MULTIPLE);
    structuretree->enable_model_drag_source();
    structuretree->enable_model_drag_dest();

    GET_GLADE (xml, vocabmenu);
    GET_GLADE (xml, vocabmenu_cut);
    vocabmenu_cut->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_vocab_cut));
    GET_GLADE (xml, vocabmenu_copy);
    vocabmenu_copy->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_vocab_copy));
    GET_GLADE (xml, vocabmenu_paste);
    vocabmenu_paste->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_vocab_paste));
    GET_GLADE (xml, vocabmenu_delete);
    vocabmenu_delete->signal_activate().connect (sigc::bind (sigc::mem_fun (*this, &MainWindow::on_vocab_delete), true));
    GET_GLADE (xml, vocabmenu_add);
    vocabmenu_add->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_vocab_add));

    xml->get_widget_derived ("vocabedit", vocabedit);
    vocabedit->signal_cursor_changed().connect (sigc::mem_fun (*this, &MainWindow::recheck_ui));
    vocabedit->signal_create_row_path().connect (sigc::mem_fun (*this, &MainWindow::on_vocab_add_path));
    vocabedit->signal_delete_row_path().connect (sigc::mem_fun (*this, &MainWindow::on_vocab_delete_path));
    vocabedit->signal_delete_row().connect (sigc::bind (sigc::mem_fun (*this, &MainWindow::on_vocab_delete), true));
    vocabedit->set_search_equal_func (sigc::mem_fun (*this, &MainWindow::on_search_equal));
    vocabedit->set_popup (vocabmenu);
    vocabedit->get_selection()->set_mode (Gtk::SELECTION_MULTIPLE);
    vocabedit->enable_model_drag_source();
    vocabedit->enable_model_drag_dest();
    
    GET_GLADE (xml, mainbook);
    mainbook->signal_switch_page().connect (sigc::hide (sigc::hide (
      sigc::mem_fun (*this, &MainWindow::recheck_ui))));

    GET_GLADE (xml, findvalue);
    findvalue->signal_changed().connect (sigc::mem_fun (*this, &MainWindow::on_find_change));
    findvalue->signal_activate().connect (sigc::mem_fun (*this, &MainWindow::on_find));
    findvalue->signal_event().connect (sigc::mem_fun (*this, &MainWindow::on_find_key_press_event));
    GET_GLADE (xml, find);
    find->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::on_find));
    GET_GLADE (xml, findbox);
    GET_GLADE (xml, findstate);
    
    GET_GLADE (xml, questionbox);
    GET_GLADE (xml, questionlabel);
    
    GET_GLADE (xml, buttonquestionlabel);
    GET_GLADE (xml, playagain);
    playagain->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::on_play_audio));
    
    // Set some pictures
    Gtk::Image *playimage;
    GET_GLADE (xml, playimage);
    playimage->set (Gtk::StockID (VT_STOCK_PLAY), Gtk::ICON_SIZE_LARGE_TOOLBAR);
    xml->get_widget ("playimage3", playimage); debug_assert (playimage != NULL);
    playimage->set (Gtk::StockID (VT_STOCK_PLAY), Gtk::ICON_SIZE_LARGE_TOOLBAR);
    
    GET_GLADE (xml, solutionbox);
    GET_GLADE (xml, solutionlabel);

    GET_GLADE (xml, buttonsolutionlabel);
    GET_GLADE (xml, playsolution);
    playsolution->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::on_play_audio_solution));

    GET_GLADE (xml, tool_usetimer);
    Gtk::Image* timer_image = new Gtk::Image (Gtk::StockID (VT_STOCK_TIMER), Gtk::ICON_SIZE_LARGE_TOOLBAR);
    timer_image->show();
    tool_usetimer->set_icon_widget (*Gtk::manage (timer_image));

    // File properties vocab types list
    Glib::RefPtr<VocabTypesModel> types = VocabTypesModel::create (vocab_test->get_types());
    Gtk::TreeView *vocabtypeslist;
    GET_GLADE (xml, vocabtypeslist);
    vocabtypeslist->set_model (types);
    vocabtypeslist->append_column ("Type", types->get_column_text());
    vocabtypeslist->append_column ("Count", types->get_column_num());

    // Settings score clear and expire buttons
    xml->get_widget ("SettingsDlg", settingsdlg); debug_assert (settingsdlg != NULL);
    GET_GLADE (xml, scoreerase); 
    scoreerase->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::on_scoreerase_clicked));
    GET_GLADE (xml, scoreexpire);
    scoreexpire->signal_clicked().connect (sigc::mem_fun (*this, &MainWindow::on_scoreexpire_clicked));

    // Session part
    tsettings.signal_testmode().connect (sigc::mem_fun (*this, &MainWindow::on_testmode_changed));
    tsettings.signal_resultsshown().connect (sigc::mem_fun (*this, &MainWindow::on_resultsshown_changed));
    
    GET_GLADE (xml, starttraining);
    starttraining->signal_clicked().connect (sigc::mem_fun (main_session, &MainSession::start));
    GET_GLADE (xml, pausetraining);
    pausetraining->signal_clicked().connect (sigc::mem_fun (main_session, &MainSession::pause));
    GET_GLADE (xml, endtraining);
    endtraining->signal_clicked().connect (sigc::mem_fun (main_session, &MainSession::finish));
    GET_GLADE (xml, showvocab);
    showvocab->signal_clicked().connect (sigc::mem_fun (main_session, &MainSession::advance_and_show));
    GET_GLADE (xml, showvocab2);
    showvocab2->signal_clicked().connect (sigc::mem_fun (main_session, &MainSession::advance_and_show));
    GET_GLADE (xml, guessready);
    guessready->signal_clicked().connect (sigc::mem_fun (main_session, &MainSession::handle_input_guess));
    GET_GLADE (xml, confirmresult);
    confirmresult->signal_clicked().connect (sigc::mem_fun (main_session, &MainSession::hide_results));
    GET_GLADE (xml, guess);
    guess->signal_event().connect (sigc::mem_fun (main_session, &MainSession::on_guess_key_press_event));
   
    GET_GLADE (xml, resultlist);
    resultlist->set_model (result_store);
    resultlist->append_column ("what", result_columns.what);
    resultlist->append_column ("content", result_columns.content);

    GET_GLADE (xml, pickbox);
    GET_GLADE (xml, picktable);

    GET_GLADE (xml, enterbox);
    GET_GLADE (xml, showbuttons);
    GET_GLADE (xml, resultbuttons);

    // File properties dialog
    xml->get_widget ("FilePropertiesDlg", fileproperties); debug_assert (fileproperties != NULL);
    GET_GLADE (xml, fileauthor);
    GET_GLADE (xml, filenotes);
    GET_GLADE (xml, filelicense);
    xml->get_widget_derived ("scorecurve", scorecurve); debug_assert (scorecurve != NULL);

#ifdef EVALUATION
    // Voice analysis dialog
    xml->get_widget ("AnalysisDlg", analysisdlg); debug_assert (analysisdlg != NULL);
    xml->get_widget_derived ("voiceref", voiceref); debug_assert (voiceref != NULL);
    xml->get_widget_derived ("voiceuser", voiceuser); debug_assert (voiceuser != NULL);
    GET_GLADE (xml, voicedata);
    GET_GLADE (xml, userh);
    GET_GLADE (xml, refh);
    GET_GLADE (xml, userv);
    GET_GLADE (xml, refv);
    voiceref->set_hadjustment (*refh->get_adjustment());
    voiceref->set_vadjustment (*refv->get_adjustment());
    voiceuser->set_hadjustment (*userh->get_adjustment());
    voiceuser->set_vadjustment (*userv->get_adjustment());
    voiceref->set_lock (0, 150);
    voiceuser->set_lock (0, 150);
    on_fit_changed();
    GET_GLADE (xml, playuser);
    playuser->signal_clicked().connect (sigc::bind<bool> (sigc::mem_fun (this, &MainWindow::on_play_voice), true));
    GET_GLADE (xml, playref);
    playref->signal_clicked().connect (sigc::bind<bool> (sigc::mem_fun (this, &MainWindow::on_play_voice), false));
    xml->get_widget ("playimage1", playimage); debug_assert (playimage != NULL);
    playimage->set (Gtk::StockID (VT_STOCK_PLAY), Gtk::ICON_SIZE_BUTTON);
    xml->get_widget ("playimage2", playimage); debug_assert (playimage != NULL);
    playimage->set (Gtk::StockID (VT_STOCK_PLAY), Gtk::ICON_SIZE_BUTTON);
    
    // This interface does not work with the keyboard TODO
    GET_GLADE (xml, comparevoice);
    comparevoice->signal_pressed().connect (sigc::mem_fun (*this, &MainWindow::on_record_audio));
    comparevoice->signal_released().connect (sigc::mem_fun (*this, &MainWindow::on_analyze_audio));
    GET_GLADE (xml, comparevoice1);
    comparevoice1->signal_pressed().connect (sigc::mem_fun (*this, &MainWindow::on_record_audio));
    comparevoice1->signal_released().connect (sigc::mem_fun (*this, &MainWindow::on_analyze_audio));
#else
    GET_GLADE (xml, comparevoice);
    comparevoice->hide();
#endif

    // this will activate a lot of initialization
    menu_close->activate();
    structuretree->grab_focus();

    // Event chaining for the buttons
    GET_GLADE (xml, tool_open);
    tool_open->signal_clicked().connect (sigc::hide_return (sigc::mem_fun (*menu_open, &Gtk::MenuItem::activate)));
    
    GET_GLADE (xml, tool_save);
    tool_save->signal_clicked().connect (sigc::hide_return (sigc::mem_fun (*menu_save, &Gtk::MenuItem::activate)));

    GET_GLADE (xml, tool_settings);
    tool_settings->signal_clicked().connect (sigc::hide_return (sigc::mem_fun (*menu_settings, &Gtk::MenuItem::activate)));

    GET_GLADE (xml, tool_fileproperties);
    tool_fileproperties->signal_clicked().connect (sigc::hide_return (sigc::mem_fun (*menu_fileproperties, &Gtk::MenuItem::activate)));
    
    // Disabled until we figure out how to stop crashing when gnome-panel gets killed
/*
#ifndef WIN32
    debug_log ("Loading tray icon");
    
    tray_icon.reset (new TrayIcon());
    tray_icon->set();
#endif    
*/
    
    debug_log ("MainWindow loaded");
        
    const std::vector<std::string>& params = CmdLine::get();
    if (params.size() > 1)
      open (params[params.size() - 1]);
};

void MainWindow::quiet_timer_mode()
{
    if (vocab_test->get_filename() != "") {
	tsettings.usetimer.set_value (true);
	tsettings.usetimer.status_to_ui();
	structuretree->get_selection()->select_all();
	main_session.start();
	main_session.pause();
    }
}

MainWindow::ImageDataPtr MainWindow::get_displayed (const TwoDeData::Data& sonogram) const
{
    // Get dimensions
    unsigned maxi = sonogram.shape()[0];
    unsigned maxj = sonogram.shape()[1];
    // Build a flat alias to the sonogram data
    TwoDeData::ConstFlatRef sono = TwoDeData::flatten (sonogram);

    // Allocate data
    ImageDataPtr A (new ImageData (boost::extents[maxi][maxj][3]));
    // Build a flat alias to it
    ImageDataRef B (*A);
    boost::array<ImageDataRef::index,3> dims = {{1, maxi * maxj, 3}};
    B.reshape (dims);

    // Convert to color
    for (unsigned i = 0; i < 3; ++i) {
        // Getting the right color plane
        ImageDataRef::array_view<1>::type c = B[boost::indices[0][ImageDataRef::index_range()][i]];
        // Converting
        std::transform (sono.begin(), sono.end(), c.begin(), pos_colors[i]);
    }
    return A;
}

Glib::RefPtr<Gdk::Pixbuf> MainWindow::get_pixbuf (const ImageDataPtr A) const
{
    unsigned maxi = A->shape()[0];
    unsigned maxj = A->shape()[1];
    if (! (maxi * maxj))
      return Glib::RefPtr<Gdk::Pixbuf>();
    else
      // we use scale_simple to get a new pixbuf which is not dependent on data
      // of A (which is freed soon) anymore
      return Gdk::Pixbuf::create_from_data (A->data(), Gdk::COLORSPACE_RGB,
        false, 8, maxj, maxi, maxj * 3)->scale_simple (maxj, maxi, Gdk::INTERP_NEAREST);
}

void MainWindow::set_sonogram (TwoDeData::DataPtr raw_data, const bool user)
{
    MarkerSonogram* data = user ? voiceuser : voiceref;
    if (! raw_data)
      data->clear_sonogram();
    else
      data->set_sonogram (get_pixbuf (get_displayed (*raw_data)));
    data->clear_labels();
    data->clear_paths();
}

void MainWindow::add_sonogram_label (const SynthesisDisplayLabel& label, const bool user)
{
    MarkerSonogram* data = user ? voiceuser : voiceref;
    data->add_label (label.posx, label.posy, label.label);
}

void MainWindow::on_fit_changed()
{
    voiceref->set_fit (settings.reffit.get_value());
    voiceuser->set_fit (settings.userfit.get_value());
}

void MainWindow::on_size_allocate (Gtk::Allocation& allocation)
{
    Gtk::Window::on_size_allocate (allocation);
    settings.sizewidth.set_value (allocation.get_width());
    settings.sizeheight.set_value (allocation.get_height());
}

bool MainWindow::on_window_state_event (GdkEventWindowState* event)
{
    if (event->changed_mask & Gdk::WINDOW_STATE_MAXIMIZED)
      settings.sizestate.set_value ((event->new_window_state & Gdk::WINDOW_STATE_MAXIMIZED) > 0);
    return false;
}

void MainWindow::on_play_audio() 
{
    int chapter, section, vocab;
    if (tsettings.is_testmode()) {
        main_session.present_vocabulary_audio();
    } else if (get_current_vocab (chapter, section, vocab) == 3) {
	present_vocabulary_audio (0, VocabFormatter::format (
	  vocab_test->get_vocab (chapter, section, vocab, tsettings.get_first()), 3),
	  tsettings.get_first());
    }
}

void MainWindow::on_play_audio_solution() 
{
    int chapter, section, vocab;
    if (tsettings.is_testmode() || tsettings.is_resultsshown()) {
        main_session.present_vocabulary_audio_solution();
    } else if (get_current_vocab (chapter, section, vocab) == 3) {
	present_vocabulary_audio (0, VocabFormatter::format (
	  vocab_test->get_vocab (chapter, section, vocab, tsettings.get_second()), 3),
	  tsettings.get_second());
    }
}

void MainWindow::on_record_audio() 
{
    debug_log ("Recording...");
    analysis.reset (new SynthesisAnalysis());
}

void MainWindow::on_play_voice (const bool user) 
{
    if (! analysis) return;
    if (user)
      analysis->play_user();
    else
      analysis->play_ref();
}

void MainWindow::on_analyze_audio() 
{
    debug_log ("Analyzing...");
    analysis->analyze();
    
    set_sonogram (analysis->get_user(), true);
    const SynthesisDisplayLabelData& user_labels = analysis->get_user_labels();
    std::for_each (user_labels.begin(), user_labels.end(), 
      boost::bind (&MainWindow::add_sonogram_label, this, _1, true));
    set_sonogram (analysis->get_ref(), false);
    const SynthesisDisplayLabelData& ref_labels = analysis->get_ref_labels();
    std::for_each (ref_labels.begin(), ref_labels.end(), 
      boost::bind (&MainWindow::add_sonogram_label, this, _1, false));
    
    voicedata->set_label (analysis->get_text());
    if (! analysisdlg->is_visible()) {
	analysisdlg->set_transient_for (*this);    
	analysisdlg->run();
	analysisdlg->hide();
	analysis.reset();
    }
}

void MainWindow::on_open_activate() 
{
    open ("");
};

void MainWindow::open (const std::string& filename) 
{
    if (! handle_modified()) return;
    
    std::string selected;
    if (filename != "") {
	selected = filename;
    } else {
	Gtk::FileChooserDialog open (_("Open File"), Gtk::FILE_CHOOSER_ACTION_OPEN);
	open.set_transient_for (*this);
    
	//Add response buttons the the dialog:
	open.add_button (Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	open.add_button (Gtk::Stock::OPEN, Gtk::RESPONSE_OK)->grab_default();
    
	//Add filters, so that only certain file types can be selected:
	Gtk::FileFilter filter_xml;
	filter_xml.set_name ("XML files");
	filter_xml.add_mime_type ("application/xml");
	filter_xml.add_mime_type ("text/xml");
	filter_xml.add_pattern ("*.xml"); // This is a windows workaround
	open.add_filter (filter_xml);
    
	Gtk::FileFilter filter_any;
	filter_any.set_name ("Any files");
	filter_any.add_pattern ("*");
	open.add_filter (filter_any); 

	if (open.run() == Gtk::RESPONSE_OK) {
	    open.hide();
	    selected = open.get_filename();
	} else return;
    }
    try {
	vocab_test->read_xml (selected);
	chdir (Glib::path_get_dirname (selected).c_str());
	set_defaults();
	create_structure();
	settings.lastfiles.add_mru (selected);
    } catch (const ApplicationError &ex) {
	Gtk::MessageDialog error (*this, strmanip::ucompose (
	    _("Could not read file %1: %2"), selected, ex.what()), false, Gtk::MESSAGE_ERROR);
	error.run();
    }
};

void MainWindow::on_save_activate() {
    if (! vocab_test->get_filename().length()) {
	on_saveas_activate();
	return;
    };
    try {
	vocab_test->write_xml();
    } catch (const ApplicationError &ex) {
	Gtk::MessageDialog error (*this, strmanip::ucompose (
	    _("Could not write file %1: %2"), vocab_test->get_filename(), ex.what()), false, Gtk::MESSAGE_ERROR);
	error.run();
    }
}
    
void MainWindow::on_saveas_activate() 
{
    Gtk::FileChooserDialog save(_("Save File as"), Gtk::FILE_CHOOSER_ACTION_SAVE);
    save.set_transient_for(*this);

    save.set_filename (vocab_test->get_filename());
    
    //Add response buttons the the dialog:
    save.add_button (Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
    save.add_button (Gtk::Stock::SAVE_AS, Gtk::RESPONSE_OK)->grab_default();

    //Add filters, so that only certain file types can be selected:
    Gtk::FileFilter filter_xml;
    filter_xml.set_name("XML files");
    filter_xml.add_mime_type("application/xml");
    filter_xml.add_mime_type("text/xml");
    save.add_filter(filter_xml);

    Gtk::FileFilter filter_any;
    filter_any.set_name("Any files");
    filter_any.add_pattern("*");
    save.add_filter(filter_any); 

    if (save.run() == Gtk::RESPONSE_OK) {
	save.hide();
	std::string filename = save.get_filename();
    	try {
	    vocab_test->write_xml (filename);
	    settings.lastfiles.add_mru (filename);
	} catch (const ApplicationError &ex) {
	    Gtk::MessageDialog error (*this, strmanip::ucompose (
	      _("Could not write file %1: %2"), filename, ex.what()), false, Gtk::MESSAGE_ERROR);
	    error.run();
	}
    }
};

void MainWindow::on_close_activate() {
    if (! handle_modified()) return;
	
    vocab_test->simple();
    std::vector<std::string> languages;
    settings.get_default_languages (std::back_inserter (languages));
    std::for_each (languages.begin(), languages.end(), 
      bind (&VocabTest::add_language, vocab_test, _1));
    set_defaults();
    create_structure();
};

void MainWindow::on_quit_activate() {
    if (! handle_modified()) return;
    Gtk::Main::quit();
};

bool MainWindow::on_delete_event (GdkEventAny* event) {
    on_quit_activate();
    return true;
}

bool MainWindow::handle_modified() {
    if (vocab_test->get_modified()) {
	Gtk::MessageDialog whattodo (*this, _("File was modified. Do you want to loose these changes?"), 
	  false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO);
	switch (whattodo.run()) {
	  case Gtk::RESPONSE_NO:
	    on_save_activate();
	    if (vocab_test->get_modified()) return false;
	    break;
	  case Gtk::RESPONSE_YES:
	    break;
	  default:
	    return false;
	}
    }
    return true;
}

void MainWindow::on_about_activate() {
    Gtk::AboutDialog *about;  
    xml->get_widget ("AboutDlg", about);
    about->set_version (VERSION);
    about->set_transient_for (*this);
    
    about->run();
    about->hide();
};

void MainWindow::on_addlang_activate() {
    Gtk::Dialog *addlang;  
    xml->get_widget ("AddLanguageDlg", addlang);
    addlang->set_transient_for (*this);
    
    addlang->run();
    addlang->hide();
    
    Gtk::Entry *addlanguage;  
    xml->get_widget ("addlanguage", addlanguage);
    vocab_test->add_language (addlanguage->get_text());
    const std::vector<std::string> languages = vocab_test->get_languages();
    tsettings.set_languages (languages.begin(), languages.end(), 
    tsettings.get_first(), tsettings.get_second());
    create_edit_structure();
};

/** Delete the currently selected vocabularies.
 * Calls the apropriate method depending on the current focus.
 */
void MainWindow::on_delete_activate() {
    const Gtk::Widget* focused = get_focus();
    if (focused == vocabedit)
      on_vocab_delete();
    else if (focused == structuretree)
      on_structure_delete();
};

/** Copy the currently selected vocabularies.
 * Calls the apropriate method depending on the current focus.
 */
void MainWindow::on_copy_activate() {
    const Gtk::Widget* focused = get_focus();
    if (focused == vocabedit)
      on_vocab_copy();
    else if (focused == structuretree)
      on_structure_copy();
};

/** Cut the currently selected vocabularies.
 * Calls the apropriate method depending on the current focus.
 */
void MainWindow::on_cut_activate() {
    const Gtk::Widget* focused = get_focus();
    if (focused == vocabedit)
      on_vocab_cut();
    else if (focused == structuretree)
      on_structure_cut();
};

/** Paste the vocabularies from the clipboard.
 * Calls the apropriate method depending on the current focus.
 */
void MainWindow::on_paste_activate() {
    const Gtk::Widget* focused = get_focus();
    if (focused == vocabedit)
      on_vocab_paste();
    else if (focused == structuretree)
      on_structure_paste();
};

void MainWindow::on_settings_activate() {
    settingsdlg->set_transient_for (*this);
    settingsdlg->run();
    settingsdlg->hide();
};

/** Display file settings dialog.
 * Settings are read out of vocab_test and, if the dialog is confirmed, 
 * written back.
 */
void MainWindow::on_fileproperties_activate() {
    fileproperties->set_transient_for (*this);
    fileauthor->set_text (vocab_test->get_property ("author"));
    filenotes->get_buffer()->set_text (vocab_test->get_property ("notes"));
    filelicense->get_buffer()->set_text (vocab_test->get_property ("license"));
    scorecurve->set_vocabs (*vocab_test);
    switch (fileproperties->run()) {
      case Gtk::RESPONSE_OK:
	vocab_test->set_property ("author", fileauthor->get_text());
	vocab_test->set_property ("notes", filenotes->get_buffer()->get_text());
	vocab_test->set_property ("license", filelicense->get_buffer()->get_text());
        break;
      default:;
    }
    fileproperties->hide();
};

void MainWindow::on_modified (const bool modified) 
{
    const std::string filename = vocab_test->get_filename();
    Glib::ustring title;
    
    if (filename.length()) 
      title = Glib::filename_display_basename (filename);
    else
      title = _("Unnamed document");
    
    set_title (strmanip::ucompose (_("%1%2 - Vocabulary Trainer"), modified ? "*" : "", title));
};

void MainWindow::on_testmode_changed() 
{
    if (tsettings.is_testmode()) {
	hide_find_box();
	solutionbox->hide();
	resultlist->hide();
	questionbox->show();
	if (tsettings.is_buttons()) {
	    pickbox->show();
	    enterbox->hide();
	} else {
	    pickbox->hide();
	    enterbox->show();
	}
	resultbuttons->hide();
	showbuttons->hide();
	mainbook->set_current_page (PAGE_TRAINING);
	recheck_ui(); // Disable Buttons
    }
};

void MainWindow::on_resultsshown_changed() 
{
    if (tsettings.is_resultsshown()) {
	questionbox->hide();
	pickbox->hide();
	enterbox->hide();
	resultlist->show();
	resultbuttons->show();
	recheck_ui(); // Enable Buttons
    } else {
	resultlist->hide();
	resultbuttons->hide();
	showbuttons->show();
	questionbox->show();
	solutionbox->show();
	recheck_ui(); // Now we have to manually display vocabs
    }
};

void MainWindow::on_vocab_add() 
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c = NULL;
    if (cursor_visible (vocabedit))
      vocabedit->get_cursor (p, c);
    else
      p = Gtk::TreePath ("0");
    on_vocab_add_path (p, NULL);
};

void MainWindow::on_vocab_delete (const bool ask) 
{
    if (! vocabedit->get_selection()->count_selected_rows()) return;
    Gtk::TreePath p = *vocabedit->get_selection()->get_selected_rows().begin();

    if (ask) {
	Gtk::MessageDialog whattodo (*this, _("Do you really want to delete the selected vocabularies?"), 
	  false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_OK_CANCEL);
	if (whattodo.run() != Gtk::RESPONSE_OK) return;
    }

    int chapter, section, vocab;
    get_current_vocab (chapter, section, vocab);
    std::vector<int> vocabs;
    get_selected_vocabs (vocabs);
    vocab_test->delete_vocabs (chapter, section, vocabs);
    vocabedit->set_cursor (p);    
    if (! cursor_visible (vocabedit)) {
	p.prev();
	vocabedit->set_cursor (p);    
    }
};

void MainWindow::on_vocab_copy() 
{
    if (! vocabedit->get_selection()->count_selected_rows()) return;
    Gtk::TreePath p = *vocabedit->get_selection()->get_selected_rows().begin();

    int chapter, section, vocab;
    get_current_vocab (chapter, section, vocab);
    std::vector<int> vocabs;
    get_selected_vocabs (vocabs);
    CopyPaste copy_paste (vocab_test);
    copy_paste.copy_vocabs (chapter, section, vocabs);
};

void MainWindow::on_vocab_paste() 
{
    int chapter, section, vocab;
    get_current_vocab (chapter, section, vocab);
    CopyPaste copy_paste (vocab_test);
    copy_paste.paste_vocabs (chapter, section, vocab);
};

void MainWindow::on_vocab_cut() 
{
    on_vocab_copy();
    on_vocab_delete (false);
};

void MainWindow::on_showprev_clicked()
{
    int chapter, section, vocab;
    int depth = get_current_vocab (chapter, section, vocab);
    if (depth == 2) {
	set_current_vocab (chapter, section, vocab_test->get_vocab_num (chapter, section) - 1);
    } else if (depth == 3) {
	if (vocab) --vocab;
	set_current_vocab (chapter, section, vocab);
    }
};

void MainWindow::on_shownext_clicked() 
{
    int chapter, section, vocab;
    int depth = get_current_vocab (chapter, section, vocab);
    if (depth == 2) {
	set_current_vocab (chapter, section, 0);
    } else if (depth == 3) {
	set_current_vocab (chapter, section, vocab + 1);
    }
};

void MainWindow::on_vocabcell_edited (const Glib::ustring& path, const Glib::ustring& new_text, const std::string language)
{
    int chapter, section, vocab;
    Glib::RefPtr<VocabTreeModel>::cast_dynamic(vocabedit->get_model())->get_vocab_addr (Gtk::TreePath (path), chapter, section, vocab);
    vocab_test->set_vocab (chapter, section, vocab, language, new_text);
}

void MainWindow::on_vocabtype_edited (const Glib::ustring& path, const Glib::ustring& new_text)
{
    int chapter, section, vocab;
    Glib::RefPtr<VocabTreeModel>::cast_dynamic(vocabedit->get_model())->get_vocab_addr (Gtk::TreePath (path), chapter, section, vocab);
    vocab_test->set_type (chapter, section, vocab, new_text);
}

void MainWindow::on_structure_edited (const Glib::ustring& path, const Glib::ustring& new_text)
{
    int chapter, section, vocab;
    Glib::RefPtr<VocabTreeModel>::cast_dynamic(structuretree->get_model())->get_vocab_addr (Gtk::TreePath (path), chapter, section, vocab);
    
    if (section == -1)
      vocab_test->set_chapter_name (chapter, new_text);
    else
      vocab_test->set_section_name (chapter, section, new_text);
    
}

void MainWindow::on_structure_addsection()
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    int chapter, section, vocab;
    if (cursor_visible (structuretree)) {
	structuretree->get_cursor (p, c);
	Glib::RefPtr<VocabTreeModel>::cast_dynamic(structuretree->get_model())->get_vocab_addr 
	  (p, chapter, section, vocab);
	add_section (chapter, section);
    }
}

void MainWindow::on_structure_addchapter() 
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    int chapter, section, vocab;
    if (cursor_visible (structuretree)) {
	structuretree->get_cursor (p, c);
	Glib::RefPtr<VocabTreeModel>::cast_dynamic(structuretree->get_model())->get_vocab_addr 
	  (p, chapter, section, vocab);
	add_chapter (chapter);
    } else {
	add_chapter (-1);
    }
}

void MainWindow::on_structure_cut()
{    
    on_structure_copy();
    on_structure_delete (false);
};

void MainWindow::on_structure_copy()
{    
    if (! structuretree->get_selection()->count_selected_rows()) return;
    Gtk::TreePath p = *structuretree->get_selection()->get_selected_rows().begin();

    int chapter, section, vocab;
    get_current_vocab (chapter, section, vocab);
    std::vector<std::pair<int, int> > data;
    get_selected_structure (data);
    
    CopyPaste copy_paste (vocab_test);
    copy_paste.copy_structure (data);
};

void MainWindow::on_structure_paste() 
{
    int chapter, section, vocab;
    get_current_vocab (chapter, section, vocab);
    CopyPaste copy_paste (vocab_test);
    copy_paste.paste_structure (chapter, section, vocab);
    // Help the vocab treeview coping with the changes
    create_edit_structure (true);
};

void MainWindow::on_structure_delete (const bool ask)
{    
    if (! structuretree->get_selection()->count_selected_rows()) return;
    Gtk::TreePath p = *structuretree->get_selection()->get_selected_rows().begin();

    if (ask) {
	Gtk::MessageDialog whattodo (*this, _("Do you really want to delete the selected chapters/sections?"), 
	  false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_OK_CANCEL);
	if (whattodo.run() != Gtk::RESPONSE_OK) return;
    }

    int chapter, section, vocab;
    get_current_vocab (chapter, section, vocab);
    std::vector<std::pair<int, int> > data;
    get_selected_structure (data);
    VocabTest::normalize_indizes (data);
    
    for (std::vector<std::pair<int, int> >::reverse_iterator i = data.rbegin();
      i != data.rend(); ++i) {
	if (i->second != -1) 
	  vocab_test->delete_section (i->first, i->second);
	else
	  vocab_test->delete_chapter (i->first);
    }

    structuretree->set_cursor (p);    
    if (! cursor_visible (structuretree)) {
	p.prev();
	structuretree->set_cursor (p);    
    }
};

void MainWindow::on_scoreerase_clicked()
{
    Gtk::MessageDialog whattodo (*settingsdlg, _("Do you really want to erase the vocabulary score cache?"), 
      false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_OK_CANCEL);
    if (whattodo.run() != Gtk::RESPONSE_OK) return;
    VocabCache::get().clear();
    create_edit_structure (true);
};

void MainWindow::on_scoreexpire_clicked()
{
    Gtk::MessageDialog whattodo (*settingsdlg, _("Do you really want to expire all vocabularies?"), 
      false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_OK_CANCEL);
    if (whattodo.run() != Gtk::RESPONSE_OK) return;
    VocabCache::get().expire();
    create_edit_structure (true);
};

void MainWindow::on_structure_delete_path (Gtk::TreePath p, Gtk::TreeViewColumn* c)
{
    int chapter, section, vocab;
    Glib::RefPtr<VocabTreeModel>::cast_dynamic (structuretree->get_model())->get_vocab_addr 
      (p, chapter, section, vocab);
    if (section == -1)
      vocab_test->delete_chapter (chapter);
    else
      vocab_test->delete_section (chapter, section);
    structuretree->set_cursor (p);
    if (! cursor_visible (structuretree) && (p.prev() || p.up()) && p.size())
      structuretree->set_cursor (p);
}

bool MainWindow::on_search_equal (const Glib::RefPtr<Gtk::TreeModel>& model, int column, const Glib::ustring& key, const Gtk::TreeModel::iterator& iter) 
{
    try {
	Glib::RefPtr<VocabTreeModel> m = Glib::RefPtr<VocabTreeModel>::cast_dynamic (model);
	const Gtk::TreeModel::Path& p = m->get_path (iter);
	Glib::ustring search = key.casefold();
	int chapter, section, vocab;
	m->get_vocab_addr (p, chapter, section, vocab);
	if (vocab != -1) {
	    const std::vector<std::string> langs = vocab_test->get_languages();
	    for (unsigned i = 0; i < langs.size(); i++) {
		if (Glib::ustring (vocab_test->get_vocab (chapter, section, vocab, langs[i])).casefold().find (search) != Glib::ustring::npos)
		  return false;
	    }
	    return true;
	} else if (section != -1) {
	    return (Glib::ustring (vocab_test->get_section_name (chapter, section)).casefold().find (search) == Glib::ustring::npos);
	} else {
	    return (Glib::ustring (vocab_test->get_chapter_name (chapter)).casefold().find (search) == Glib::ustring::npos);
	}
    } catch (...) {};
    return true;
}

void MainWindow::add_chapter (const int chapter) 
{
    Gtk::TreePath p;
    const int chapter_new = chapter == -1 ? 0 : chapter;
    vocab_test->insert_chapter (chapter_new);
    Glib::RefPtr<VocabTreeModel>::cast_dynamic (structuretree->get_model())->get_vocab_path (
      p, chapter_new, -1, -1);
    handle_pending_events(); // gives the treeview time to update
    structuretree->set_cursor (p, *(structuretree->get_column (0)), true);
};

void MainWindow::add_section (const int chapter, const int section) 
{
    if (chapter == -1) {
        add_chapter (chapter);
    } else {
	Gtk::TreePath p;
	const int section_new = section == -1 ? 0 : section;
	vocab_test->insert_section (chapter, section_new);
	Glib::RefPtr<VocabTreeModel>::cast_dynamic (structuretree->get_model())->get_vocab_path (
	  p, chapter, section_new, -1);
	structuretree->expand_to_path (p);
	structuretree->set_cursor (p, *(structuretree->get_column (0)), true);
    }
};

void MainWindow::on_vocab_add_path (const Gtk::TreePath p, Gtk::TreeViewColumn* c) 
{
    int chapter, section, vocab;
    Glib::RefPtr<VocabTreeModel>::cast_dynamic(vocabedit->get_model())->get_vocab_addr (p, chapter, section, vocab);
    vocab_test->insert_vocab (chapter, section, vocab);
    handle_pending_events(); // gives the treeview time to update
    if (! c) {
	int new_idx = vocabedit->get_next_editable_column (-1);
	if (new_idx != -1)
	  c = vocabedit->get_column (new_idx);
	else
	  c = vocabedit->get_column (0);
    }
    vocabedit->set_cursor (p, *c, true);    
};

void MainWindow::on_vocab_delete_path (Gtk::TreePath p, Gtk::TreeViewColumn* c) 
{
    int chapter, section, vocab;
    Glib::RefPtr<VocabTreeModel>::cast_dynamic (vocabedit->get_model())->get_vocab_addr (p, chapter, section, vocab);
    vocab_test->delete_vocab (chapter, section, vocab);
};

void MainWindow::on_find()
{
    show_find_box();
    if (!findvalue->has_focus())
      findvalue->grab_focus();
    
    Glib::RefPtr<VocabTreeModel> search = VocabTreeModel::create (0, vocab_test);
    find_next_match (search, true);
}

void MainWindow::on_find_change()
{
    // This updates the timer
    show_find_box();
    
    Glib::RefPtr<VocabTreeModel> search = VocabTreeModel::create (0, vocab_test);
    find_next_match (search, false);
}

void MainWindow::find_next_match (const Glib::RefPtr<VocabTreeModel> search, const bool skip)
{
    const std::string value = findvalue->get_text();
    if (value == "") {
	set_find_state (false, ""); // No search string
	return;
    }
	
    int chapter, section, vocab;
    int level = get_current_vocab (chapter, section, vocab);
    Gtk::TreePath p, pfirst;
    if (! search->get_vocab_path (p, chapter, section, vocab)) {
	set_find_state (true, _("Phrase not found")); // No selection
	return;
    }
    pfirst = p;
    Gtk::TreeIter iter = search->get_iter (p);
    if (! iter) {
	set_find_state (false, ""); // Shouldn't happen
	return;
    }
    if (level == 3 && ! on_search_equal (search, 0, value, iter) && ! skip) {
	set_find_state (false, ""); // Stay here
	return;
    }
    
    bool turnaround = ! walk_model (search, p);
    while (p != pfirst) {
	if (p.size() == 3 && ! on_search_equal (search, 0, value, search->get_iter (p))) {
	    int chapter, section, vocab;
	    search->get_vocab_addr (p, chapter, section, vocab);
	    set_current_vocab (chapter, section, vocab);
	    set_find_state (false, turnaround ? _("Reached end of file, continued from top") : ""); // Found it
	    return;
	}
	turnaround |= ! walk_model (search, p);
    };
    set_find_state (true, _("Phrase not found"));
}

void MainWindow::set_find_state (const bool error, const std::string& text)
{
    findstate->set_text (text);
    if (error) {
        findvalue->modify_base (Gtk::STATE_NORMAL, Gdk::Color ("#FF8080"));
        findvalue->modify_text (Gtk::STATE_NORMAL, Gdk::Color ("white"));
    } else {
        findvalue->unset_base (Gtk::STATE_NORMAL);
        findvalue->unset_text (Gtk::STATE_NORMAL);
    }
}

/** Walk a tree model.
 * On a given tree model, the next node is searched.
 * \param search Gtk::TreeModel to search
 * \param p Current position (must be valid) and new position on return
 * \return Whether it was possible to find a next node without turnaround
 */

bool MainWindow::walk_model (Glib::RefPtr<Gtk::TreeModel> search, Gtk::TreePath& p)
{
    Gtk::TreeIter iter = search->get_iter (p);
    do {
	// Are there child nodes?
	if (iter && ! iter->children().empty()) {
	    p.down();
	// Is there a next node?
	} else if (iter && iter + 1) {
	    p.next();
	// Otherwise go up, next
	} else {
	    if (! p.up() || ! p.size()) {
		p = Gtk::TreePath ("0");
		return false;
	    };
	    p.next();
	}
	iter = search->get_iter (p);
    } while (! iter);
    return true;
}

bool MainWindow::show_find_box()
{
    if (tsettings.is_testmode()) return false;
    find_connection.disconnect();
    find_connection = Glib::signal_timeout().connect (sigc::mem_fun (this, &MainWindow::hide_find_box), 10 * 1000);
    findbox->show();
    return true;
}

bool MainWindow::hide_find_box()
{
    findvalue->set_text ("");
    find_connection.disconnect();
    findbox->hide();
    // This is for the timeout
    return true;
}

bool MainWindow::on_find_key_press_event (GdkEvent* event)
{
    if (event->type == GDK_KEY_PRESS) {
	if (event->key.keyval == GDK_Escape) {
	    hide_find_box();
	    vocabedit->grab_focus();
	    return true;
	} else show_find_box();
    }
    return false;
}

void MainWindow::recheck_ui() 
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    int chapter, section, vocab;

    // Access to the settings page
    if (tsettings.is_testmode()) {
        mainbook->pages()[PAGE_EDIT]->get_child()->hide();
        mainbook->pages()[PAGE_OPTIONS]->get_child()->set_sensitive (false);
    } else {
        mainbook->pages()[PAGE_EDIT]->get_child()->show();
        mainbook->pages()[PAGE_OPTIONS]->get_child()->set_sensitive (true);
    }

    // Pages
    if (mainbook->get_current_page() == PAGE_TRAINING) {
	guessready->grab_default();
    }
	
    // Edit page
    // Check whether there is sth to delete
    if (cursor_visible (vocabedit)) {
	delvocab->set_sensitive (true);
    } else {
	delvocab->set_sensitive (false);
    }
    
     // Check whether there is a current section
    if (cursor_visible (structuretree)) {
	structuretree->get_cursor (p, c);
	if (p.size() < 2) {
	    addvocab->set_sensitive (false);
	} else {
	    addvocab->set_sensitive (true);
	}
    } else {
	addvocab->set_sensitive (false);
    }

    // Is Synthesis for first language available
    if (SynthesisInterface::get().available (tsettings.get_first())) {
	playagain->set_sensitive (tsettings.is_testmode() || cursor_visible (vocabedit));
	playagain->show();
	questionlabel->hide();
    } else {
	playagain->hide();
	questionlabel->show();
    }

    if (! tsettings.is_testmode() && ! tsettings.is_resultsshown()) {
	int depth = get_current_vocab (chapter, section, vocab);
	shownext->set_sensitive (
	    depth == 3 && vocab_test->get_vocab_num (chapter, section) > vocab + 1 ||
	    depth == 2 && vocab_test->get_vocab_num (chapter, section));
	showprev->set_sensitive (
	    depth == 3 && vocab > 0 ||
	    depth == 2 && vocab_test->get_vocab_num (chapter, section));
	present_manual_vocabulary_written(); 
	present_manual_solution();
    }

    // Left panel
    starttraining->set_sensitive (
      cursor_visible (structuretree) && ! tsettings.is_testmode());
    if (tsettings.is_testmode() && tsettings.usetimer.get_value()) {
	starttraining->hide();
	pausetraining->show();
	pausetraining->set_sensitive (! settings.timerannoying.get_value());
    } else {
	starttraining->show();
	pausetraining->hide();
    }
    
    starttraining->set_sensitive (
      cursor_visible (structuretree) && ! tsettings.is_testmode());
    
    // Testing or editing
    if (tsettings.is_testmode()) {
	structuretree->set_sensitive (false);
	endtraining->set_sensitive (true);
	if (settings.timerannoying.get_value()) {
	    menu_main->hide();
	    tool_main->hide();
	}
    } else {
	structuretree->set_sensitive (true);
	endtraining->set_sensitive (false);
	menu_main->show();
	tool_main->show();
    }
}

void MainWindow::set_defaults() 
{
    const std::vector<std::string> languages = vocab_test->get_languages();
    tsettings.set_languages (languages.begin(), languages.end(), 
      settings.get_first_default(), settings.get_second_default());
    tsettings.set_mode (settings.is_audio_default(), 
			settings.is_written_default(), 
			settings.askmode.get_value() == 1);
    tsettings.presentsolution.set_value (settings.presentsolution.get_value());
    tsettings.presentsolution.status_to_ui();
    tsettings.usetimer.set_value (settings.is_use_timer_default());
    tsettings.usetimer.status_to_ui();
    main_session.reset();
}

void MainWindow::create_edit_structure (const bool force) 
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    std::vector<std::pair<Gtk::CellRenderer*, Gtk::TreeViewColumn*> > renderers;
    structuretree->get_cursor (p, c);
    Glib::RefPtr<VocabTreeModel> edit = VocabTreeModel::create (settings.scorefade.get_value(), vocab_test);
    
    Glib::RefPtr<VocabTreeModel> old = 
      Glib::RefPtr<VocabTreeModel>::cast_dynamic (vocabedit->get_model());
    if (cursor_visible (structuretree) && p.size() == 2)  {
	if (force || ! old || ! old->get_limit_branch().size() || old->get_limit_branch() != p)
	  edit->limit_to_branch (p);
	else
	  edit = old;
    } else {
	if (force || ! old || old->get_limit_level() != 3)
	  edit->set_limit_level (3); // Don't show anything
	else
	  edit = old;
    }
    if (edit == old) return;
    vocabedit->remove_all_columns();
    vocabedit->set_model (edit);
    
    // This is more or less experimental, because there are no change
    // notifications send out to inform the Gtk::TreeView
    if (settings.scorecolumns.get_value()) {
	Gtk::TreeViewColumn* const column_score = new Gtk::TreeViewColumn (_("Score"), 
	  edit->get_model_column (VocabTreeModel::COLUMN_SCORE));
	vocabedit->append_column (*Gtk::manage (column_score));
	renderers.push_back (std::make_pair (column_score->get_first_cell_renderer(), column_score));
    
	Gtk::TreeViewColumn* const column_expired = new Gtk::TreeViewColumn (_("Expires"));
	column_expired->pack_start (edit->get_model_column_bool (VocabTreeModel::COLUMN_EXPIRED), false);
	column_expired->pack_start (edit->get_model_column (VocabTreeModel::COLUMN_EXPIRE_TIME));
	vocabedit->append_column (*Gtk::manage (column_expired));
	renderers.push_back (std::make_pair (*(column_expired->get_cell_renderers().begin()), column_expired));
	renderers.push_back (std::make_pair (*(++column_expired->get_cell_renderers().begin()), column_expired));
    }
    
    const std::vector<std::string> languages = vocab_test->get_languages();
    for (unsigned int i = 0; i < languages.size(); ++i) {
	Gtk::TreeViewColumn* const column = new Gtk::TreeViewColumn (languages[i], 
	  edit->get_model_column (i + VocabTreeModel::COLUMN_TRANSLATIONS_BASE));
	column->set_resizable();
	Gtk::CellRendererText* const renderer = dynamic_cast<Gtk::CellRendererText*> (column->get_first_cell_renderer());
	debug_assert (renderer != NULL);
	renderer->property_editable() = true;
        renderer->signal_edited().connect (
	  sigc::bind<std::string>(sigc::mem_fun (*this, &MainWindow::on_vocabcell_edited), languages[i]));
	renderers.push_back (std::make_pair (renderer, column));
	vocabedit->append_column (*Gtk::manage (column));
    }
    
    Glib::RefPtr<Gtk::EntryCompletion> completion = Gtk::EntryCompletion::create();
    completion->set_model (VocabTypesModel::create (vocab_test->get_types()));
    completion->set_text_column (0);

    CellRendererCompletion* const renderer = new CellRendererCompletion (completion);
    renderer->property_editable() = true;
    renderer->signal_edited().connect (sigc::mem_fun (*this, &MainWindow::on_vocabtype_edited));
    Gtk::TreeViewColumn* const column = new Gtk::TreeViewColumn (_("Type"), *Gtk::manage (renderer));
    column->add_attribute (renderer->property_text(), 
      edit->get_model_column (VocabTreeModel::COLUMN_TYPE));
    if (settings.scorevisible.get_value()) {
	column->add_attribute (renderer->property_cell_background(), 
	  edit->get_model_column (VocabTreeModel::COLUMN_COLOR));
	renderer->property_cell_background_set() = true;
    }
    renderers.push_back (std::make_pair (renderer, column));
    vocabedit->append_column (*Gtk::manage (column));

    // Make colored if requested
    if (settings.scorevisible.get_value()) {
	std::vector<std::pair<Gtk::CellRenderer*, Gtk::TreeViewColumn*> >::const_iterator i;
	for (i = renderers.begin(); i != renderers.end(); ++i) {
	    i->second->add_attribute (i->first->property_cell_background(), 
	      edit->get_model_column (VocabTreeModel::COLUMN_COLOR));
	    // it doen't work when you set it to false: BUG?
	    i->first->property_cell_background_set() = true;
	}
    }
}
    
void MainWindow::create_structure() 
{
    Glib::RefPtr<VocabTreeModel> struc = VocabTreeModel::create (
      settings.scorefade.get_value(), vocab_test);
    struc->set_limit_level (1);
    structuretree->remove_all_columns();
    structuretree->set_model (struc);
    Gtk::TreeViewColumn* const column = new Gtk::TreeViewColumn ("", 
      struc->get_model_column (VocabTreeModel::COLUMN_TITLE));
    dynamic_cast<Gtk::CellRendererText*> (column->get_first_cell_renderer())->property_editable() = true;
    dynamic_cast<Gtk::CellRendererText*> (column->get_first_cell_renderer())->signal_edited().connect (
      sigc::mem_fun (*this, &MainWindow::on_structure_edited));
    structuretree->append_column (*Gtk::manage (column));
    
    Gtk::TreePath p ("0:0");
    structuretree->expand_to_path (p);
    structuretree->set_cursor (p);
    create_edit_structure (true);
}


void MainWindow::present_manual_vocabulary_written()
{
    int chapter, section, vocab;
    if (get_current_vocab (chapter, section, vocab) == 3) {
	Glib::ustring first = tsettings.get_first();
	Glib::ustring second = tsettings.get_second();
	present_vocabulary_written (VocabFormatter::format (vocab_test->get_vocab (
	  chapter, section, vocab, first), 0, settings.get_localized_or (first), first));
    } else {
	present_vocabulary_written ("");
    }
}

void MainWindow::present_manual_solution()
{
    int chapter, section, vocab;
    if (get_current_vocab (chapter, section, vocab) == 3) {
	Glib::ustring first = tsettings.get_first();
	Glib::ustring second = tsettings.get_second();
	present_solution_written (VocabFormatter::format (vocab_test->get_vocab (
	  chapter, section, vocab, second), 1, settings.get_localized_or (second), second));
    } else {
	set_solution_box ("", false);
    }
}

void MainWindow::present_vocabulary_written (const std::string& what) 
{
    std::string display = what;
    if (settings.is_presfont())
      display = "<span font_desc=\"" + settings.get_presfont() + "\">" + display + "</span>";
    set_question_box (display);
}

void MainWindow::present_vocabulary_audio (const unsigned wait, const std::string& what, 
  const std::string& language) 
{
    SynthesisInterface::get().play (wait, what, language);
}

void MainWindow::present_solution_written (const std::string& what, const bool audio) 
{
    set_solution_box (what, audio);
}

void MainWindow::present_solution_audio (const unsigned wait, const std::string& what, 
  const std::string& language) 
{
    SynthesisInterface::get().play (wait, what, language);
}


// UI helper

void MainWindow::set_question_box (const std::string& what) 
{
    questionlabel->set_markup (what);
    buttonquestionlabel->set_markup (what);
}

void MainWindow::set_solution_box (const std::string& what, const bool audio) 
{
    solutionbox->show();
    solutionlabel->set_markup (what);
    buttonsolutionlabel->set_markup (what);
    
    // Is Synthesis for second language available
    if (SynthesisInterface::get().available (tsettings.get_second()) && audio) {
	playsolution->set_sensitive (tsettings.is_testmode() || 
	  tsettings.is_resultsshown() || cursor_visible (vocabedit));
	playsolution->show();
	solutionlabel->hide();
    } else {
	playsolution->hide();
	solutionlabel->show();
    }
}

void MainWindow::set_question_button (const unsigned i, const std::string& s) 
{
    if (i >= pickbuttons.size()) return;
    if (s != "") {
	std::string button_title;
	if (pickbuttons.size() <= 10)
	  button_title = "_%1. %2";
	else
	  button_title = "%2";
	pickbuttons[i]->set_sensitive (true);
	dynamic_cast<Gtk::Label*> (dynamic_cast<Gtk::Button*> 
	  (pickbuttons[i])->get_child())->set_markup_with_mnemonic 
	  (strmanip::ucompose (button_title, (i + 1) % 10, s));
    } else {
	pickbuttons[i]->set_sensitive (false);
	pickbuttons[i]->set_label ("");
    }
}

/** Processes all pending events in the GTK queue */
void MainWindow::handle_pending_events() 
{
    while (Gtk::Main::events_pending())
      Gtk::Main::iteration();
}

/* this returns the currently selected chapter, section and vocab; the return value 
 * encodes the actual depth (3 = vocab, 2 = section, ...)
 */
int MainWindow::get_current_vocab (int& chapter, int& section, int& vocab) const
{
    int structure = get_treeview_vocab (structuretree, chapter, section, vocab);
    if (structure == 2) {
	int vocabe = get_treeview_vocab (vocabedit, chapter, section, vocab);
	if (vocabe == 0) {
	    vocab = -1;
	    return 2;
	} else if (vocabe == 3) {
	    return 3;
	} else {
	    g_critical ("Unregular VocabEdit selection!");
	    chapter = section = vocab = -1;
	    return -1;
	}
    } else if (structure == 1) {
	section = vocab = -1;
	return 1;
    } else if (structure == 0) {
	chapter = section = vocab = -1;
	return 0;
    } else {
	g_critical ("Unknown selection depth in StructureTree!");
	chapter = section = vocab = -1;
	return -1;
    }
}

/* Try to select the mentioned vocabulary
 */
bool MainWindow::set_current_vocab (const int chapter, const int section, const int vocab)
{
    return set_treeview_vocab (structuretree, chapter, section, -1) &&
      set_treeview_vocab (vocabedit, chapter, section, vocab);
}

int MainWindow::get_treeview_vocab (Gtk::TreeView* tv, int& chapter, int& section, int& vocab) const
{
    Gtk::TreePath p; Gtk::TreeViewColumn* c;
    if (! cursor_visible (tv)) {
	if (!structuretree->get_selection()->get_selected_rows().size())
	  return 0;
	p = *structuretree->get_selection()->get_selected_rows().begin();
    } else
      tv->get_cursor (p, c);
    Glib::RefPtr<VocabTreeModel>::cast_dynamic (tv->get_model())->
	get_vocab_addr (p, chapter, section, vocab);
    return 
      vocab   > -1 ? 3 :
      section > -1 ? 2 :
      chapter > -1 ? 1 :
	             0;
    
}

bool MainWindow::set_treeview_vocab (Gtk::TreeView* tv, const int chapter, const int section, const int vocab)
{
    Gtk::TreePath p, p2; Gtk::TreeViewColumn* c;
    if (! Glib::RefPtr<VocabTreeModel>::cast_dynamic (tv->get_model())->get_vocab_path (
      p, chapter, section, vocab)) return false;
    tv->expand_to_path (p);
    tv->set_cursor (p);
    tv->get_cursor (p2, c);
    return cursor_visible (tv) && p == p2;
}

void MainWindow::create_pick_table (const unsigned num) 
{
    while (picktable->children().size()) {
	delete picktable->children().begin()->get_widget();
    }
    const unsigned cols = 2;
    const unsigned rows = (num + cols - 1) / cols;
    picktable->resize (rows, cols);
    pickbuttons.clear();
    for (unsigned i = 0; i < num; i++) {
	Gtk::Button* button = new Gtk::Button (""); // Empty label
	pickbuttons.push_back (button);
	button->signal_clicked().connect (sigc::bind<unsigned> (sigc::mem_fun (main_session, 
	  &MainSession::handle_button_guess), i));
	picktable->attach (*Gtk::manage(button), 
	  (i / rows), 
	  (i / rows) + 1,
	  (i % rows), 
	  (i % rows) + 1);
	button->show();
    }
}

void MainWindow::add_result_list_line (const std::string what, const std::string content) 
{
    Gtk::TreeModel::Row row = *(result_store->append());
    row[result_columns.what] = what;
    row[result_columns.content] = content;
}

void MainWindow::clear_result_list() 
{
    result_store->clear();
}

void MainWindow::select_input() 
{
    if (! tsettings.is_buttons()) {
	guess->grab_focus();
	guess->select_region (0, guess->get_text().length());
    }
}

std::string MainWindow::get_guess_text() 
{
    return guess->get_text();
};

void MainWindow::get_selected_vocabs (std::vector<int>& data)
{
    data.clear();
    Gtk::TreeSelection::ListHandle_Path sel = vocabedit->get_selection()->get_selected_rows();
    int chapter, section, vocab;
    for (Gtk::TreeSelection::ListHandle_Path::const_iterator i = sel.begin(); i != sel.end(); i++) {
	Glib::RefPtr<VocabTreeModel>::cast_dynamic (vocabedit->get_model())->get_vocab_addr (*i, chapter, section, vocab);
	data.push_back (vocab);
    }
}

void MainWindow::get_selected_structure (std::vector<std::pair<int, int> >& data)
{
    data.clear();
    Gtk::TreeSelection::ListHandle_Path sel = structuretree->get_selection()->get_selected_rows();
    int chapter, section, vocab;
    for (Gtk::TreeSelection::ListHandle_Path::const_iterator i = sel.begin(); i != sel.end(); i++) {
	Glib::RefPtr<VocabTreeModel>::cast_dynamic (structuretree->get_model())->get_vocab_addr (*i, chapter, section, vocab);
	data.push_back (std::pair<int,int> (chapter, section));
    }
}
