/***************************************************************************
 *            MarkerSonogram.cc
 *
 *  Wed Jan 26 13:31:46 2005
 *  Copyright  2005  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "MarkerSonogram.h"

#include "debug.h"

#include <libgnomecanvasmm/pixbuf.h>
#include <libgnomecanvasmm/group.h>
#include <libgnomecanvasmm/line.h>
#include <libgnomecanvasmm/affinetrans.h>
#include <libgnomecanvasmm/text.h>
#include <libgnomecanvasmm/rect.h>

#include <boost/checked_delete.hpp>

#define gray50_width 2
#define gray50_height 2
static char gray50_bits[] = { 0x02, 0x01 };

MarkerSonogram::MarkerSonogram (BaseObjectType* cobject, const Glib::RefPtr<Gnome::Glade::Xml>& refGlade) :
  Glib::ObjectBase (typeid (MarkerSonogram)),
  Gnome::Canvas::Canvas (cobject),
  xml (refGlade),
  pixbuf (NULL),
  parent (Gtk::manage (new Gnome::Canvas::Group (*root()))),
  lock_x (0), lock_y (0),
  zoom_x (1), zoom_y (1),
  fit (false),
  rotation (false),
  hpath (false), hlabel (false)
{
}

void MarkerSonogram::on_size_allocate (Gtk::Allocation& allocation)
{
    if (get_width()  != allocation.get_width()  && (lock_x || fit) ||
	get_height() != allocation.get_height() && (lock_y || fit))
      recalc_display (allocation.get_width(), allocation.get_height());
    Gnome::Canvas::Canvas::on_size_allocate (allocation);
}

void MarkerSonogram::set_zoom_x (const double zoom)
{
    zoom_x = zoom;
    recalc_display_current();
}

void MarkerSonogram::set_zoom_y (const double zoom)
{
    zoom_y = zoom;
    recalc_display_current();
}

void MarkerSonogram::set_lock (const unsigned new_lock_x, const unsigned new_lock_y)
{
    lock_x = new_lock_x;
    lock_y = new_lock_y;
    set_size_request (lock_x, lock_y);
}

void MarkerSonogram::set_fit (const bool new_fit)
{
    fit = new_fit;
    recalc_display_current();
}

bool MarkerSonogram::on_event (GdkEvent* event)
{
  Gnome::Canvas::Canvas::on_event(event);

  switch (event->type) {
    case GDK_BUTTON_PRESS:
      get_bin_window()->pointer_grab (false, 
	Gdk::POINTER_MOTION_MASK | Gdk::BUTTON_RELEASE_MASK,
        Gdk::Cursor (Gdk::FLEUR),
	event->button.time);
      motion_x = event->motion.x_root;
      motion_y = event->motion.y_root;
      get_scroll_offsets (orig_x, orig_y);
      break;

    case GDK_BUTTON_RELEASE:
      get_bin_window()->pointer_ungrab (event->button.time);
      break;
  
    case GDK_MOTION_NOTIFY:
      if (event->motion.state & GDK_BUTTON1_MASK) {
	  if (!Gdk::Event::events_pending())
	  scroll_to (
	    -static_cast<int>(event->motion.x_root - motion_x) + orig_x, 
	    -static_cast<int>(event->motion.y_root - motion_y) + orig_y);
      }
      break;

    default:
      break;
  }

  return false;
}

void MarkerSonogram::set_turned (const bool new_rotation)
{
    rotation = new_rotation;
}


void MarkerSonogram::set_sonogram (Glib::RefPtr<Gdk::Pixbuf> sono)
{
    debug_log ("Loading pixmap (%1, %2)", sono->get_width(), sono->get_height());
    if (pixbuf) delete pixbuf;
    pixbuf = Gtk::manage (new Gnome::Canvas::Pixbuf (*parent, 0, 0, sono));
    pixbuf->property_width() = sono->get_width();
    pixbuf->property_width_in_pixels() = false;
    pixbuf->property_width_set() = true;
    pixbuf->property_height() = sono->get_height();
    pixbuf->property_height_in_pixels() = false;
    pixbuf->property_height_set() = true;
    pixbuf->lower_to_bottom();
    
    if (rotation) {
	real_height = sono->get_width();
	real_width = sono->get_height();
  	Gnome::Art::AffineTrans a = Gnome::Art::AffineTrans::identity();
	a.flip (false, true);
	Gnome::Art::AffineTrans b = Gnome::Art::AffineTrans::rotation (-90);
	a *= b;
	pixbuf->affine_absolute (a);
	pixbuf->property_anchor() = Gtk::ANCHOR_SE;
    } else {
	real_width = sono->get_width();
	real_height = sono->get_height();
	pixbuf->affine_absolute (get_correction());
    }
    recalc_display_current();
}

void MarkerSonogram::clear_sonogram()
{
    if (pixbuf) delete pixbuf;
}

Gnome::Art::AffineTrans MarkerSonogram::get_correction() const
{
    Gnome::Art::AffineTrans a = Gnome::Art::AffineTrans::identity();
    Gnome::Art::AffineTrans b = Gnome::Art::AffineTrans::translation (0, real_height);
    a.flip (false, true);
    a *= b;
    return a;
}

void MarkerSonogram::add_path (const Gnome::Canvas::Points& p, const std::string& color)
{
    debug_log ("Loading path with %1 points", p.size());
    if (p.size() <= 1) return;
    Gnome::Canvas::Line* l = Gtk::manage (new Gnome::Canvas::Line (*parent, p));
    if (hpath) l->hide();
    l->affine_absolute (get_correction());
//    l->affine_relative (Gnome::Art::AffineTrans::translation (0.5, 0.5));
    l->property_fill_color() = color;
    l->property_width_units() = 4.0;
    pathstuff.push_back (l);
}

void MarkerSonogram::show_labels (const bool visible)
{
    hlabel = ! visible;
    if (visible)
      std::for_each (labelstuff.begin(), labelstuff.end(), std::mem_fun (&Gnome::Canvas::Item::show));
    else
      std::for_each (labelstuff.begin(), labelstuff.end(), std::mem_fun (&Gnome::Canvas::Item::hide));
}

void MarkerSonogram::show_paths (const bool visible)
{
    hpath = ! visible;
    if (visible)
      std::for_each (pathstuff.begin(), pathstuff.end(), std::mem_fun (&Gnome::Canvas::Item::show));
    else
      std::for_each (pathstuff.begin(), pathstuff.end(), std::mem_fun (&Gnome::Canvas::Item::hide));
}

void MarkerSonogram::add_label (const int x, const int y, const std::string& text)
{
    debug_log ("Loading label at (%1,%2) with \"%3\"", x, y, text);
    if (text.size()) {
	//TODO: This text should scale
	Gnome::Canvas::Text* t = Gtk::manage (new Gnome::Canvas::Text (*parent, x, y, text));
	if (hlabel) t->hide();
	t->affine_absolute (get_correction());
//	t->affine_relative (Gnome::Art::AffineTrans::translation (0.6, 0.6));
	t->affine_relative (Gnome::Art::AffineTrans::translation (0.1, 0.1));
	t->property_fill_color() = "white";
	t->property_font() = "Sans bold 10";
	t->property_anchor() = Gtk::ANCHOR_SW;
	debug_log ("Label is %1 wide", t->property_text_width());
	labelstuff.push_back (t);

	//Gnome::Canvas::Rect* r = Gtk::manage (new Gnome::Canvas::Rect (*parent, x, y, 
	//  x + t->property_text_width(), y + t->property_text_height()));
	//r->affine_absolute (get_correction());
	//r->affine_relative (Gnome::Art::AffineTrans::translation (0.6, 0.6));
	//r->property_fill_color() = "black";
	//r->lower (1);
	//labelstuff.push_back (r);
    }
    if (x >= 0) {
	Gnome::Canvas::Line* l = Gtk::manage (new Gnome::Canvas::Line (*parent));
	if (hlabel) l->hide();
	Gnome::Canvas::Points p;
	p.push_back (Gnome::Art::Point (x, 0));
	p.push_back (Gnome::Art::Point (x, real_height));
	l->affine_absolute (get_correction());
//	l->affine_relative (Gnome::Art::AffineTrans::translation (0.5, 0));
	l->property_points() = p;
	l->property_fill_color() = "green";
	l->property_width_units() = 1;
	labelstuff.push_back (l);
    }
    if (y >= 0) {
	Gnome::Canvas::Line* l = Gtk::manage (new Gnome::Canvas::Line (*parent));
	if (hlabel) l->hide();
	Gnome::Canvas::Points p;
	p.push_back (Gnome::Art::Point (0, y)); 
	p.push_back (Gnome::Art::Point (real_width, y));
	l->affine_absolute (get_correction());
//	l->affine_relative (Gnome::Art::AffineTrans::translation (0, 0.5));
	l->property_points() = p;
	l->property_fill_color() = "green";
	l->property_width_units() = 1;
	labelstuff.push_back (l);
    }
}

void MarkerSonogram::add_fuzzy_label (const int x_min, const int y_min, const int x_max, const int y_max)
{
    debug_log ("Loading fuzzy label between (%1,%2) and (%3,%4)", x_min, y_min, x_max, y_max);
    if (x_max >= 0) {
	Gnome::Canvas::Rect* l = Gtk::manage (new Gnome::Canvas::Rect (*parent,
	  x_min, -1, x_max, real_height + 1));
	if (hlabel) l->hide();
	l->affine_absolute (get_correction());
//	l->affine_relative (Gnome::Art::AffineTrans::translation (0.5, 0));
	l->property_fill_color() = "green";    
	l->property_fill_stipple() =
          Gdk::Bitmap::create (gray50_bits, gray50_width, gray50_height);
	l->property_outline_color() = "green";
	l->property_width_pixels() = 1;
	labelstuff.push_back (l);
    }
    if (y_max >= 0) {
	Gnome::Canvas::Rect* l = Gtk::manage (new Gnome::Canvas::Rect (*parent,
	  -1, y_min, real_width + 1, y_max));
	if (hlabel) l->hide();
	l->affine_absolute (get_correction());
//	l->affine_relative (Gnome::Art::AffineTrans::translation (0, 0.5));
	l->property_fill_color() = "green";    
	l->property_fill_stipple() =
          Gdk::Bitmap::create (gray50_bits, gray50_width, gray50_height);
	l->property_outline_color() = "green";
	l->property_width_pixels() = 1;
	labelstuff.push_back (l);
    }
}

void MarkerSonogram::clear_labels()
{
    std::for_each (labelstuff.begin(), labelstuff.end(), boost::checked_deleter<Gnome::Canvas::Item>());
    labelstuff.clear();
}

void MarkerSonogram::clear_paths()
{
    std::for_each (pathstuff.begin(), pathstuff.end(), boost::checked_deleter<Gnome::Canvas::Item>());
    pathstuff.clear();
}

void MarkerSonogram::recalc_display_current()
{
    recalc_display (get_width(), get_height());
}

void MarkerSonogram::recalc_display (const unsigned width, const unsigned height)
{
    if (! pixbuf) return;	
    
    const double scale_x = (lock_x || fit) ? (width  / real_width) : zoom_x;
    const double scale_y = (lock_y || fit) ? (height / real_height) : zoom_y;

    parent->affine_absolute (Gnome::Art::AffineTrans::scaling (scale_x, scale_y));
    
//    set_scroll_region (-scale_x * real_width, -scale_y * real_height, scale_x * real_width, scale_y * real_height);
    set_scroll_region (0, 0, scale_x * real_width, scale_y * real_height);
    debug_log ("Scaling (%1, %2) to %3, %4", lock_x, lock_y, scale_x * real_width, scale_y * real_height);
}
