/***************************************************************************
 *            Singleton.cc
 *
 *  Mon Feb  7 15:12:29 2005
 *  Copyright  2005  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/** \file
 * Singleton support with ordered destruction.
 */

#include "RegisteredSingleton.h"

#include <iostream>

namespace registeredsingleton_internal
{
    
    /** Destroy the registered singletons in the right order.
     * This depends on std::multimap<unsigned> being sorted
     * in the right way.
     */
    
    SingletonRegistry::~SingletonRegistry()
    {
	// This should assure the right order of destruction
	while (classes.begin() != classes.end())
	  classes.erase (classes.begin());
    }
    
    /** Manage a singleton instance.
     * The instance is registered with the given priority and is destroyed
     * when the program finishes after all other singleton instances with lower
     * priority have been destructed.
     * \param singleton Instance pointer
     * \param priority Destruction priority
     * \return Copy of instance pointer
     */
    
    RegisteredSingletonBase* SingletonRegistry::manage (RegisteredSingletonBase* singleton, const unsigned priority)
    {
	classes.insert (std::make_pair (priority, RegisteredSingletonBasePtr (singleton)));
	return singleton;
    }
    
    /** Get the SingletonRegistry instance.
     * \return Reference to the one and only SingletonRegistry
     */
    
    SingletonRegistry& SingletonRegistry::get()
    {
	static boost::scoped_ptr<SingletonRegistry> instance (new SingletonRegistry());
	return *instance;
    }
}
