/***************************************************************************
 *            Singleton.h
 *
 *  Mon Feb  7 12:22:51 2005
 *  Copyright  2005  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _SINGLETON_H
#define _SINGLETON_H

/** \file
 * Singleton support with ordered destruction.
 */

#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>

#include <map>

/** Contains helper classes to manage the destruction sequence.
 */
 
namespace registeredsingleton_internal
{
    
    /** Base class for singletons which is noncopyable.
     */
    
    class RegisteredSingletonBase
    {
      private:
	RegisteredSingletonBase (const RegisteredSingletonBase& other); 
	RegisteredSingletonBase& operator= (const RegisteredSingletonBase& other); 
      protected:
	RegisteredSingletonBase() {};
      public:
	  
	/** Needed virtual destructor to be able to destruct the derived 
         * singletons with a base pointer.
         */
      
	virtual ~RegisteredSingletonBase() {};
    };
    
    /** Class for singleton destruction managment.
     */
    
    class SingletonRegistry
    {
      public:
	~SingletonRegistry();
	static SingletonRegistry& get();
	RegisteredSingletonBase* manage (RegisteredSingletonBase* singleton, const unsigned priority);
      private:
	typedef boost::shared_ptr<RegisteredSingletonBase> RegisteredSingletonBasePtr;
	typedef std::multimap<unsigned, RegisteredSingletonBasePtr> ClassMap;
	ClassMap classes;
    };
}

/** Base type for singletons. 
 * Derive from this template to get destruction managment for free. 
 * The template parameter \c UserClass specifies the name of the singelton 
 * class, \c level gives the time of destruction, with higher numbers being
 * destructed later.
 * \warning There is no locking done, so it is not thread-save!
 */
    
template<class UserClass, unsigned level>
class RegisteredSingleton : public registeredsingleton_internal::RegisteredSingletonBase
{
  public:
    static UserClass& get();
};

/** Return the instance of the singleton class.
 * \return Reference to the singleton instance
 */

template<class UserClass, unsigned level>
UserClass& RegisteredSingleton<UserClass, level>::get()
{
    static UserClass& instance = * static_cast<UserClass*> (
      registeredsingleton_internal::SingletonRegistry::get().manage (new UserClass(), level));
    return instance;
}

#endif /* _SINGLETON_H */
