/***************************************************************************
 *            SessionFeeder.cc
 *
 *  Fri Dec 10 12:22:04 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "SessionFeeder.h"
#include "VocabFormatter.h"

#include "support.h"

#include "debug.h"

#include <boost/bind.hpp>

#include <algorithm>

//#include <stdlib.h>

SessionFeederSection::SessionFeederSection (const unsigned chapter, const unsigned section) : 
  chapter (chapter), 
  section (section) 
{ 
}

bool SessionFeederSection::operator< (const SessionFeederSection& other) const 
{
    return 
      chapter  < other.chapter ||
      chapter == other.chapter && section  < other.section;
}

SessionFeederVocab::SessionFeederVocab (const unsigned chapter, const unsigned section, const unsigned vocab) : 
  chapter (chapter), 
  section (section),
  vocab (vocab) 
{ 
}

bool SessionFeederVocab::operator< (const SessionFeederVocab& other) const 
{
    return 
      chapter  < other.chapter ||
      chapter == other.chapter && section  < other.section ||
      chapter == other.chapter && section == other.section && vocab < other.vocab;
}

SessionFeeder::SessionFeeder (const boost::shared_ptr<VocabTest> vocab_test) :
  started (false),
  random_order (false),
  score_based (false),
  alternatives_num (0),
  first_repeat_num (1),
  wrong_repeat_num (1),
  alternatives_sorted (true),
  alternatives_samepreferred (true),
  vocab_test (vocab_test),
  vocab_cache (VocabCache::get())
{
}

SessionFeederPtr SessionFeeder::create(const boost::shared_ptr<VocabTest> vocab_test)
{
    return SessionFeederPtr (new SessionFeeder (vocab_test));
}

/** Set the languages for the feeder.
 * \param first Natural language
 * \param second Foreign language
 */

void SessionFeeder::set_languages (const std::string& first, const std::string& second)
{
    first_language = first;
    second_language = second;
}

/** Set the translations for "or" for the feeder.
 * \param first "or" in the natural language
 * \param second "or" in the foreign language
 */

void SessionFeeder::set_ors (const std::string& first, const std::string& second)
{
    first_or = first;
    second_or = second;
}

void SessionFeeder::set_mode (const bool random, const bool scorebased, const unsigned first_repeat, const unsigned wrong_repeat)
{
    random_order = random;
    score_based = scorebased;
    first_repeat_num = first_repeat;
    wrong_repeat_num = wrong_repeat;
}

void SessionFeeder::set_alternatives (const unsigned num, const bool sorted, const bool samepreferred)
{
    alternatives_num = num;
    alternatives_sorted = sorted;
    alternatives_samepreferred = samepreferred;
}

void SessionFeeder::add_chapter (const unsigned chapter)
{
    chapters.insert (chapter);
}

void SessionFeeder::add_section (const unsigned chapter, const unsigned section)
{
    sections.insert (SessionFeederSection (chapter, section));
}

bool SessionFeeder::get (SessionVocabPtr& vocab)
{
    if (! started) start();
	
    if (finals.size()) {
	SessionFeederVocab i = finals [finals.size() - 1];
	finals.pop_back();
	vocab = prepare_vocab (i.chapter, i.section, i.vocab);
	return true;
    } else {
	return false;
    }
}

unsigned SessionFeeder::size() const
{
    return finals.size();
}

bool SessionFeeder::test_score (const SessionFeederVocab& vocab) const 
{
    return vocab_cache.get_score_with_expires (vocab_test->get_id 
      (vocab.chapter, vocab.section, vocab.vocab)) != 1;
}

void SessionFeeder::start()
{
    // Lets loop through all the chapters
    for (ChaptersList::const_iterator i = chapters.begin(); i != chapters.end(); ++i) {
	append_chapter (*i);
    }

    // And then through all the left sections
    for (SectionsList::const_iterator i = sections.begin(); i != sections.end(); ++i) {
	if (chapters.find (i->chapter) != chapters.end()) continue;
	append_section (i->chapter, i->section);
    }
    
    finals.clear();
    // Because we pop_back, we need it the other way round
    if (score_based)
      std::remove_copy_if (vocabs.rbegin(), vocabs.rend(), back_inserter (finals),
        boost::bind (&SessionFeeder::test_score, this, _1));
    else
      finals.assign (vocabs.rbegin(), vocabs.rend());
    
    if (random_order) std::random_shuffle (finals.begin(), finals.end(), generic_random);

    // Not needed anymore
    chapters.clear();
    sections.clear();
    vocabs.clear();    

    started = true;
}

void SessionFeeder::append_chapter (const unsigned chapter)
{
    for (int i = vocab_test->get_section_num (chapter) - 1; i >= 0; --i)
      append_section (chapter, i);
}

void SessionFeeder::append_section (const unsigned chapter, const unsigned section)
{
    for (int i = vocab_test->get_vocab_num (chapter, section) - 1; i >= 0; --i) {
	if (vocab_test->get_vocab (chapter, section, i, first_language) != "" &&
	  vocab_test->get_vocab (chapter, section, i, second_language) != "")
          append_vocab (chapter, section, i);
    };
}

void SessionFeeder::append_vocab (const unsigned chapter, const unsigned section, const unsigned vocab) 
{
    vocabs.insert (SessionFeederVocab (chapter, section, vocab));
}

SessionVocabPtr SessionFeeder::prepare_vocab (const unsigned chapter, const unsigned section, const unsigned vocab) const
{
    // Hmm, creating the alternatives here causes the same selection to be used again
    // and again. But you should learn the vocabs fast anyway :-)
    
    Glib::ustring f = vocab_test->get_vocab (chapter, section, vocab, first_language);
    Glib::ustring s = vocab_test->get_vocab (chapter, section, vocab, second_language);
    std::vector<Glib::ustring> a;
    
    get_alternatives (chapter, section, vocab, a);
    return SessionVocabPtr (new SessionVocab (
      // This is for the vocabulary question
      VocabFormatter::format (f, 0, first_or, first_language),
      // This is a short version
      VocabFormatter::format (f, 2),
      // This is an audio formatted version
      VocabFormatter::format (f, 3),
    
      // This is for the right answer
      VocabFormatter::format (s, 1, second_or, second_language),
      // This is the right answer in raw format
      s,
      // This is for an audio formatted version
      VocabFormatter::format (s, 3),
      // This is for all buttons
      a,
      // This is for the right button
      VocabFormatter::format (s, 2),
      
      // Various other parameters
      first_repeat_num, wrong_repeat_num,
      vocab_test->get_id (chapter, section, vocab)
      ));
}

void SessionFeeder::get_alternatives (const unsigned chapter, const unsigned section, const unsigned vocab, std::vector<Glib::ustring>& alternatives) const
{
    debug_log ("Looking for alternatives to %1 %2 %3 %4", chapter, section, vocab,
      vocab_test->get_vocab (chapter, section, vocab, second_language));
    const std::string my_type = vocab_test->get_type (chapter, section, vocab);
    std::vector<Glib::ustring> sametype;
    for (unsigned i = 0; i < (unsigned) vocab_test->get_vocab_num (chapter, section); i++) {
	if (i == vocab) continue;
	Glib::ustring text;
        if ((text = vocab_test->get_vocab (chapter, section, i, second_language)) != "") {
	    if (alternatives_samepreferred && 
		my_type == vocab_test->get_type (chapter, section, i))
	      sametype.push_back (text);
	    else
              alternatives.push_back (text);
	}
    };
    std::random_shuffle (alternatives.begin(), alternatives.end(), generic_random);
    std::random_shuffle (sametype.begin(), sametype.end(), generic_random);
    if (alternatives.size() + sametype.size() > alternatives_num - 1) {
	if (sametype.size() < alternatives_num - 1) {
	    alternatives.resize (alternatives_num - 1 - sametype.size());
	} else {
	    alternatives.clear();
	    sametype.resize (alternatives_num - 1);
	}
    }
    alternatives.insert (alternatives.end(), sametype.begin(), sametype.end());
    alternatives.push_back (vocab_test->get_vocab (chapter, section, vocab, second_language));

    std::transform (alternatives.begin(), alternatives.end(), alternatives.begin(), 
      bind (&VocabFormatter::format, _1, 2));

    if (alternatives_sorted) 
      std::sort (alternatives.begin(), alternatives.end());
    else
      std::random_shuffle (alternatives.begin(), alternatives.end(), generic_random);
}
