/***************************************************************************
 *            SettingsAdaptors.h
 *
 *  Wed Jan 26 12:21:38 2005
 *  Copyright  2005  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _SETTINGSADAPTORS_H
#define _SETTINGSADAPTORS_H

#include "SettingsHelper.h"

#include <gtkmm/radiobutton.h>
#include <gtkmm/combobox.h>
#include <gtkmm/menuitem.h>
#include <gtkmm/expander.h>
#include <gtkmm/liststore.h>
#include <gtkmm/scale.h>
#include <gtkmm/entry.h>
#include <gtkmm/spinbutton.h>
#include <gtkmm/toggletoolbutton.h>

#ifndef DOXYGEN_SHOULD_SKIP_THIS

namespace Gtk {
    class ComboBox;
    class ToggleButton;
    class Expander;
    class Range;
    class TreeNodeChildren;
    class TreeIter;
    class TreeView;
    class TreeViewColumn;
}

#endif

#include <boost/lexical_cast.hpp>

class MenuMRUAdaptor : public AbstractMRUAdaptor {
  public:
    MenuMRUAdaptor (const std::string& widget_name) :
      widget_name (widget_name),
      widget (NULL)
    {
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }

    void add_entry (const unsigned pos, const std::string& value);
    void del_entry (const unsigned pos);
  private:
    const std::string widget_name;
    Gtk::MenuItem* widget;
};

class TreeViewMRUAdaptor : public AbstractMRUAdaptor {
  public:
    TreeViewMRUAdaptor (const std::string& widget_name) :
      widget_name (widget_name),
      widget (NULL)
    {
    }

    void link_widget (const Settings& settings);

    void add_entry (const unsigned pos, const std::string& value);
    void del_entry (const unsigned pos);
    
  private:  
    void signal_activated (const Gtk::TreePath& path, Gtk::TreeViewColumn* column);

    class ModelColumns : public Gtk::TreeModel::ColumnRecord
    {
      public:
	ModelColumns() { add (text); add (file); };
	Gtk::TreeModelColumn<std::string> text;
	Gtk::TreeModelColumn<std::string> file;
    };
    
    const std::string widget_name;
    Gtk::TreeView* widget;
    boost::shared_ptr<ModelColumns> columns;
    Glib::RefPtr<Gtk::ListStore> store;
};

class VisibleAction : public AbstractAction<bool>
{
  public:
    VisibleAction (const std::string& widget_name) : 
      widget_name (widget_name)
    {
    }

    virtual void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
        
    virtual void on_action (const bool& value)
    {
	debug_log ("%1: Setting widget sensitivity to %2", widget_name, value);
	widget->set_sensitive (value);
    }
    
  private:
    const std::string widget_name;
    Gtk::Widget* widget;
};

template <typename T>
class RangeAdaptor : public AbstractAdaptor<T>
{
  public:
    RangeAdaptor (const std::string& widget_name) : 
      widget_name (widget_name)
    {
	debug_log ("%1: Loading up range adaptor", widget_name);
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
    
    void link_callback() 
    {
	debug_log ("%1: Linking the UI callbacks", widget_name);
	con = widget->signal_value_changed().connect (sigc::mem_fun (*this, &RangeAdaptor<T>::on_widget_changed));
    }

    void release_callback() 
    {
	debug_log ("%1: Releasing the UI callbacks", widget_name);
	con.disconnect();
    }

    void set_ui (const T& value)
    {
	debug_log ("%1: Setting new value for UI: %2", widget_name, value);
        widget->set_value (value);
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("%1: Notifying setting of new UI value", widget_name);
        slot (static_cast<T> (widget->get_value()));
    }

    const std::string widget_name;
    Gtk::Range* widget;
    sigc::connection con;
};

class ExpanderAdaptor: public AbstractAdaptor<bool>
{
  public:
    ExpanderAdaptor (const std::string& widget_name) : 
      widget_name (widget_name)
    {
	debug_log ("%1: Loading up expander adaptor", widget_name);
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
    
    void link_callback() 
    {
	debug_log ("%1: Linking the UI callbacks", widget_name);
	con = widget->property_expanded().signal_changed().connect (sigc::mem_fun (*this, &ExpanderAdaptor::on_widget_changed));
    }

    void release_callback() 
    {
	debug_log ("%1: Releasing the UI callbacks", widget_name);
	con.disconnect();
    }

    void set_ui (const bool& value)
    {
	debug_log ("%1: Setting new value for UI: %2", widget_name, value);
        widget->set_expanded (value);
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("%1: Notifying setting of new UI value", widget_name);
        slot (widget->get_expanded());
    }
  
    const std::string widget_name;
    Gtk::Expander* widget;
    sigc::connection con;
};

template <typename T>
class SpinButtonAdaptor : public AbstractAdaptor<T>
{
  public:
    SpinButtonAdaptor (const std::string& widget_name) : 
      widget_name (widget_name)
    {
	debug_log ("%1: Loading up spin adaptor", widget_name);
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
    
    void link_callback() 
    {
	debug_log ("%1: Linking the UI callbacks", widget_name);
	con = widget->signal_changed().connect (sigc::mem_fun (*this, &SpinButtonAdaptor<T>::on_widget_changed));
    }

    void release_callback() 
    {
	debug_log ("%1: Releasing the UI callbacks", widget_name);
	con.disconnect();
    }

    void set_ui (const T& value)
    {
	debug_log ("%1: Setting new value for UI: %2", widget_name, value);
        widget->set_value (value);
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("%1: Notifying setting of new UI value", widget_name);
        slot (static_cast<T> (widget->get_value()));
    }
  
    const std::string widget_name;
    Gtk::SpinButton* widget;
    sigc::connection con;
};

template <class T>
class ToggleButtonAdaptor : public AbstractAdaptor<T>
{
  public:
    ToggleButtonAdaptor (const std::string& widget_name) : 
      widget_name (widget_name)
    {
	debug_log ("%1: Loading up togglebutton adaptor", widget_name);
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
    
    void link_callback() 
    {
	debug_log ("%1: Linking the UI callbacks", widget_name);
	con = widget->signal_toggled().connect (sigc::mem_fun (*this, &ToggleButtonAdaptor<T>::on_widget_changed));
    }

    void release_callback() 
    {
	debug_log ("%1: Releasing the UI callbacks", widget_name);
	con.disconnect();
    }

    void set_ui (const T& value)
    {
	debug_log ("%1: Setting new value for UI: %2", widget_name, value);
        widget->set_active (value);
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("%1: Notifying setting of new UI value", widget_name);
        slot (widget->get_active());
    }
  
    const std::string widget_name;
    Gtk::ToggleButton* widget;
    sigc::connection con;
};

template <class T>
class ToggleToolButtonAdaptor : public AbstractAdaptor<T>
{
  public:
    ToggleToolButtonAdaptor (const std::string& widget_name) : 
      widget_name (widget_name)
    {
	debug_log ("%1: Loading up toggletoolbutton adaptor", widget_name);
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
    
    void link_callback() 
    {
	debug_log ("%1: Linking the UI callbacks", widget_name);
	con = widget->signal_toggled().connect (sigc::mem_fun (*this, &ToggleToolButtonAdaptor<T>::on_widget_changed));
    }

    void release_callback() 
    {
	debug_log ("%1: Releasing the UI callbacks", widget_name);
	con.disconnect();
    }

    void set_ui (const T& value)
    {
	debug_log ("%1: Setting new value for UI: %2", widget_name, value);
        widget->set_active (value);
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("%1: Notifying setting of new UI value", widget_name);
        slot (widget->get_active());
    }
  
    const std::string widget_name;
    Gtk::ToggleToolButton* widget;
    sigc::connection con;
};

template <class T>
class CheckButtonAdaptor : public AbstractAdaptor<T>
{
  public:
    CheckButtonAdaptor (const std::string& widget_name) : 
      widget_name (widget_name)
    {
	debug_log ("%1: Loading up checkbutton adaptor", widget_name);
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
    
    void link_callback() 
    {
	debug_log ("%1: Linking the UI callbacks", widget_name);
	con = widget->signal_toggled().connect (sigc::mem_fun (*this, &CheckButtonAdaptor<T>::on_widget_changed));
    }

    void release_callback() 
    {
	debug_log ("%1: Releasing the UI callbacks", widget_name);
	con.disconnect();
    }

    void set_ui (const T& value)
    {
	debug_log ("%1: Setting new value for UI: %2", widget_name, value);
        widget->set_active (value);
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("%1: Notifying setting of new UI value", widget_name);
        slot (widget->get_active());
    }
  
    const std::string widget_name;
    Gtk::CheckButton* widget;
    sigc::connection con;
};

class RadioButtonAdaptor : public AbstractAdaptor<unsigned>
{
  public:
    RadioButtonAdaptor (const std::vector<std::string>& widget_names) : 
      widget_names (widget_names)
    {
	debug_log ("Loading up radiobutton adaptor");
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("Getting the XML widgets");
	for (std::vector<std::string>::iterator i = widget_names.begin();
	  i != widget_names.end(); ++i) {
	    Gtk::RadioButton* widget;
	    settings.xml->get_widget (*i, widget);
	    widgets.push_back (widget);
	}
    }
    
    void link_callback() 
    {
	debug_log ("Linking the UI callbacks");
	for (std::vector<Gtk::RadioButton*>::iterator i = widgets.begin();
	  i != widgets.end(); ++i)
	  cons.push_back ((*i)->signal_toggled().connect (sigc::mem_fun (*this, &RadioButtonAdaptor::on_widget_changed)));
    }

    void release_callback() 
    {
	debug_log ("Releasing the UI callbacks");
	std::for_each (cons.begin(), cons.end(), std::mem_fun_ref (&sigc::connection::disconnect));
	cons.clear();
    }

    void set_ui (const unsigned& value)
    {
	debug_log ("Setting new value for UI: %2", value);
        widgets[value]->set_active();
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("Notifying setting of new UI value");
        slot (std::distance (widgets.begin(), std::find_if (widgets.begin(), 
	  widgets.end(), std::mem_fun (&Gtk::RadioButton::get_active))));
    }
  
    std::vector<std::string> widget_names;
    std::vector<Gtk::RadioButton*> widgets;
    std::vector<sigc::connection> cons;
};

template <typename T>
class EntryNumericAdaptor: public AbstractAdaptor<T>
{
  public:
    EntryNumericAdaptor (const std::string& widget_name) : 
      widget_name (widget_name)
    {
	debug_log ("%1: Loading up entry adaptor", widget_name);
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
    
    void link_callback() 
    {
	debug_log ("%1: Linking the UI callbacks", widget_name);
	con = widget->signal_changed().connect (sigc::mem_fun (*this, &EntryNumericAdaptor<T>::on_widget_changed));
    }

    void release_callback() 
    {
	debug_log ("%1: Releasing the UI callbacks", widget_name);
	con.disconnect();
    }

    void set_ui (const T& value)
    {
	debug_log ("%1: Setting new value for UI: %2", widget_name, value);
	widget->set_text (boost::lexical_cast<std::string> (value));
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("%1: Notifying setting of new UI value", widget_name);
	const std::string strvalue = widget->get_text();
	try {
            slot (boost::lexical_cast<T> (strvalue));
	} catch (...) {
	};
    }
  
    const std::string widget_name;
    Gtk::Entry* widget;
    sigc::connection con;
};

template <typename T>
class ComboBoxStringAdaptor : public AbstractAdaptor<T>
{
  public:
    ComboBoxStringAdaptor (const std::string& widget_name) : 
      widget_name (widget_name)
    {
	debug_log ("%1: Loading up combobox adaptor", widget_name);
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
    
    void link_callback() 
    {
	debug_log ("%1: Linking the UI callbacks", widget_name);
	con = widget->signal_changed().connect (sigc::mem_fun (*this, &ComboBoxStringAdaptor<T>::on_widget_changed));
    }

    void release_callback() 
    {
	debug_log ("%1: Releasing the UI callbacks", widget_name);
	con.disconnect();
    }

    void set_ui (const T& value)
    {
	debug_log ("%1: Setting new value for UI: %2", widget_name, value);
	Gtk::TreeNodeChildren c = widget->get_model()->children();
	const std::string strvalue = boost::lexical_cast<std::string> (value);
	std::string temp;
	for (Gtk::TreeIter i = c.begin(); i!= c.end(); ++i) {
	    i->get_value (0, temp);
	    if (temp == strvalue) {
		widget->set_active (i);
		break;
	    }
	}
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("%1: Notifying setting of new UI value", widget_name);
	std::string strvalue;
        widget->get_active()->get_value (0, strvalue);
	T value = T();
	try {
	    value = boost::lexical_cast<T> (strvalue);
	} catch (...) {
	};
        slot (value);
    }
  
    const std::string widget_name;
    Gtk::ComboBox* widget;
    sigc::connection con;
};

template <typename T>
class ComboBoxAdaptor : public AbstractAdaptor<T>
{
  public:
    ComboBoxAdaptor (const std::string& widget_name) : 
      widget_name (widget_name)
    {
	debug_log ("%1: Loading up combobox adaptor", widget_name);
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
    
    void link_callback() 
    {
	debug_log ("%1: Linking the UI callbacks", widget_name);
	con = widget->signal_changed().connect (sigc::mem_fun (*this, &ComboBoxAdaptor<T>::on_widget_changed));
    }

    void release_callback() 
    {
	debug_log ("%1: Releasing the UI callbacks", widget_name);
	con.disconnect();
    }

    void set_ui (const T& value)
    {
	debug_log ("%1: Setting new value for UI: %2", widget_name, value);
	Gtk::TreeNodeChildren c = widget->get_model()->children();
	T temp;
	for (Gtk::TreeIter i = c.begin(); i!= c.end(); ++i) {
	    i->get_value (0, temp);
	    if (temp == value) {
		widget->set_active (i);
		break;
	    }
	}
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("%1: Notifying setting of new UI value", widget_name);
	T value;
        widget->get_active()->get_value (0, value);
        slot (value);
    }
  
    const std::string widget_name;
    Gtk::ComboBox* widget;
    sigc::connection con;
};

template <typename T>
class ComboBoxIndexAdaptor : public AbstractAdaptor<T>
{
  public:
    ComboBoxIndexAdaptor (const std::string& widget_name) : 
      widget_name (widget_name)
    {
	debug_log ("%1: Loading up combobox index adaptor", widget_name);
    }

    void link_widget (const Settings& settings) 
    {
	debug_log ("%1: Getting the XML widget", widget_name);
	settings.xml->get_widget (widget_name, widget);
    }
    
    void link_callback() 
    {
	debug_log ("%1: Linking the UI callbacks", widget_name);
	con = widget->signal_changed().connect (sigc::mem_fun (*this, &ComboBoxIndexAdaptor<T>::on_widget_changed));
    }

    void release_callback() 
    {
	debug_log ("%1: Releasing the UI callbacks", widget_name);
	con.disconnect();
    }

    void set_ui (const T& value)
    {
	debug_log ("%1: Setting new value for UI: %2", widget_name, value);
	widget->set_active (value);
    };
    
  private:
    void on_widget_changed()
    {
	debug_log ("%1: Notifying setting of new UI value", widget_name);
        slot (widget->get_active_row_number());
    }
  
    const std::string widget_name;
    Gtk::ComboBox* widget;
    sigc::connection con;
};

#endif /* _SETTINGSADAPTORS_H */
