/***************************************************************************
 *            SettingsHelper.h
 *
 *  Sat Jan 22 16:42:31 2005
 *  Copyright  2005  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _SETTINGSHELPER_H
#define _SETTINGSHELPER_H

/** \file
 * Helper classes for persistent settings managment.
 */
 
#include "GlobalXML.h"
#include "SettingsManager.h"
#include "debug.h"

#include <list>

#include <boost/bind.hpp>

class AbstractSetting;

/** Base class for an application settings class. 
 * All settings classes have to be derived from Settings to provide central
 * registration for the AbstractSetting derived classes.
 */

class Settings
{
  public:
      
    /** Register an AbstractSetting instance.
      * The instance is not deleted on destruction and has to be available
      * during the whole lifetime of the object.
      * \param setting AbstractSetting derived class instance
      */
  
    void register_setting (AbstractSetting* setting) { settings.push_back (setting); };
  protected:
      
    /** Initialize with the to be used SettingsManager.
      * \param manager SettingsManager for all settings
      */
  
    Settings (SettingsManagerPtr manager) : manager (manager), xml (GlobalXML::get().get_glade()) {};

    /** Initialize widget references.
      * This has to be called in every constructor for the class to work
      * correctly.
      */

    void init();

  public:
      
    /** SettingsManager used to load and store settings.
      */

    SettingsManagerPtr manager;
  
    /** Glade Xml instance for widget retrieval.
      */

    const Glib::RefPtr<Gnome::Glade::Xml> xml;
  
  private:
    std::list<AbstractSetting*> settings;
};
    
/** Base class for an application setting. 
 * It provides the interface used by the Settings class to 
 * initilize the widgets and to force the promotion of settings to the ui.
 */

class AbstractSetting 
{
  public:
    
    /** Initialize.
     * \param settings Settings instance responsable for storage.
     */
      
    AbstractSetting (Settings& settings) : settings (settings)
    {
	settings.register_setting (this); 
    };
    
    /** Virtual inheritance.
     */
    
    virtual ~AbstractSetting() {};
    
    /** Links ui widgets.
     */
    
    virtual void link_widgets() = 0;

    /** Promotes saved status to ui.
     */
    
    virtual void status_to_ui() = 0;
	
    /** Links ui notification signals if applicable.
     * Further ui changes should be handled.
     */
    
    virtual void link_signals() = 0;
  
  protected:    

    /** Settings instance responsable for storage.
     */
    
    Settings& settings;
};

inline void Settings::init()
{
    std::for_each (settings.begin(), settings.end(), std::mem_fun (&AbstractSetting::link_widgets));
    std::for_each (settings.begin(), settings.end(), std::mem_fun (&AbstractSetting::status_to_ui));
    std::for_each (settings.begin(), settings.end(), std::mem_fun (&AbstractSetting::link_signals));
}

/** Base class for an application setting ui adaptor. 
 * It provides the interface used by the GenericSetting class to 
 * initilize a widget and to set and retrieve data.
 * The template argument \c T is the value type for the setting.
 */

template <typename T>
class AbstractAdaptor
{
  public:
      
    /** Slot type used to notify to GenericSetting of ui changes.
     */

    typedef sigc::slot<void, T> slot_type;
  
    /** Virtual inheritance.
     */

    virtual ~AbstractAdaptor() {};
	
    /** Link a slot that is called with the new ui value when chanegs occur.
     * \param new_slot New slot to use for notification
     */

    virtual void link_slot (const slot_type& new_slot) { slot = new_slot; }
    
    /** Abstract: Link ui widget.
     * \param settings Settings class for the Glade Xml interface.
     */

    virtual void link_widget (const Settings& settings) {};
	
    /** Abstract: Link ui callbacks.
     */

    virtual void link_callback() {};

    /** Abstract: Release ui callbacks.
     */

    virtual void release_callback() {};
	
    /** Abstract: Set new value for the ui.
     * \param value New value
     */

    virtual void set_ui (const T& value) {};
    
  protected:
      
    /** Slot that is called on changes.
     */

    slot_type slot;
};

/** Base class for an application setting action. 
 * It provides the interface used by the GenericSetting class to 
 * execute a simple action on settings change. One example would be to
 * enable or disable another widget based on a checkbutton setting.
 * The template argument \c T is the value type for the setting.
 */

template <typename T>
class AbstractAction
{
  public:
      
    /** Virtual inheritance.
     */

    virtual ~AbstractAction() {};
	
    /** Abstract: Execute action.
     * \param value New settings value
     */

    virtual void on_action (const T& value) {};
	
    /** Abstract: Link ui widget.
     * \param settings Settings class for the Glade Xml interface.
     */

    virtual void link_widget (const Settings& settings) {}; 	
};

/** Recursivly generate a list.
 * This is for convenience only to be able to create std::vectors 
 * on the fly.
 * The template argument \c T is the value type for the setting.
 */

template <typename T>
class ParamList
{
  public:
    typedef T value_type;
  
    /** Initialize with a value.
     * This is just to save 2 parentheses.
     * \param value First value for the list
     */

    ParamList (const T& value) : data (1, value) {};

    /** Add another value to the end of the list.
     * \param value Value to add
     * \return The object itself
     */

    ParamList<T>& operator() (const T& value) { data.push_back (value); return *this; };

    /** Return the created list.
     * \return A vector with all the values
     */

    operator std::vector<T>&() { return data; };
    std::vector<T>& get() { return data; };

  private:
    std::vector<T> data;
};

template <typename T>
class GenericSetting : public AbstractSetting
{
  public:
    typedef sigc::signal<void> signal_type;
    typedef T value_type;

    typedef AbstractAdaptor<T> adaptor_type;
    // We need shared_ptrs here because otherwise inherited classes would get 
    // sliced down to adaptor_type when being copied by value
    typedef boost::shared_ptr<adaptor_type> adaptorptr_type;
    typedef std::vector<adaptorptr_type> adaptorlist_type;

    typedef AbstractAction<T> action_type;
    typedef boost::shared_ptr<action_type> actionptr_type;
    typedef std::vector<actionptr_type> actionlist_type;
  
    GenericSetting (Settings& settings, const std::string& group, const std::string& key,
      const std::string& hint, const std::string& description,
      const T& defvalue) : 
      AbstractSetting (settings), 
      group (group), 
      key (key),
      defvalue (defvalue)
    {
	debug_log ("%1:%2: Loading up setting", group, key);
	link_notifications();
    }

    GenericSetting (Settings& settings, const std::string& group, const std::string& key,
      const std::string& hint, const std::string& description,
      const T& defvalue, adaptorlist_type& widgets) : 
      AbstractSetting (settings), 
      group (group), 
      key (key),
      hint (hint),
      description (description),
      defvalue (defvalue), 
      widgets (widgets)
    {
	debug_log ("%1:%2: Loading up setting", group, key);
	link_notifications();
    }

    template <typename A>
    GenericSetting (Settings& settings, const std::string& group, const std::string& key,
      const std::string& hint, const std::string& description,
      const T& defvalue, A widget) : 
      AbstractSetting (settings), 
      group (group), 
      key (key),
      hint (hint),
      description (description),
      defvalue (defvalue), 
      widgets (1, adaptorptr_type (new A (widget)))
    {
	debug_log ("%1:%2: Loading up setting", group, key);
	link_notifications();
    }
    
    GenericSetting (Settings& settings, const std::string& group, const std::string& key,
      const std::string& hint, const std::string& description,
      const T& defvalue, adaptorlist_type& widgets, actionlist_type& actions) : 
      AbstractSetting (settings), 
      group (group), 
      key (key),
      hint (hint),
      description (description),
      defvalue (defvalue),
      widgets (widgets),
      actions (actions)
    {
	debug_log ("%1:%2: Loading up setting", group, key);
	link_notifications();
    }

    GenericSetting (Settings& settings, const std::string& group, const std::string& key,
      const std::string& hint, const std::string& description,
      const T& defvalue, adaptorptr_type widget, actionlist_type& actions) : 
      AbstractSetting (settings), 
      group (group), 
      key (key),
      hint (hint),
      description (description),
      defvalue (defvalue),
      widgets (1, widget),
      actions (actions)
    {
	
 	debug_log ("%1:%2: Loading up setting", group, key);
	link_notifications();
    }

    GenericSetting (Settings& settings, const std::string& group, const std::string& key,
      const std::string& hint, const std::string& description,
      const T& defvalue, adaptorlist_type& widgets, actionptr_type action) : 
      AbstractSetting (settings), 
      group (group), 
      key (key),
      hint (hint),
      description (description),
      defvalue (defvalue),
      widgets (widgets),
      actions (1, action)
    {
	debug_log ("%1:%2: Loading up setting", group, key);
	link_notifications();
    }

    template <typename A, typename B>
    GenericSetting (Settings& settings, const std::string& group, const std::string& key,
      const std::string& hint, const std::string& description,
      const T& defvalue, A widget, B action) : 
      AbstractSetting (settings), 
      group (group), 
      key (key),
      hint (hint),
      description (description),
      defvalue (defvalue), 
      widgets (1, adaptorptr_type (new A (widget))),
      actions (1, actionptr_type (new B (action)))
    {
	debug_log ("%1:%2: Loading up setting", group, key);
	link_notifications();
    }

    void set_value (const T& value)
    {
	settings.manager->set (group, key, value);
    }
    
    T get_value() const
    {
	T value;
	get_value_checked (value);
	return value;
    }

    bool get_value_checked (T& value) const
    {
	return settings.manager->get_checked (group, key, value);
    }
    
    void link_signals()
    {
	settings.manager->get_signal (group, key).connect (signal_v.make_slot());
    }
    
    signal_type get_signal() const 
    {
	return signal_v;
    }

    void link_widgets() 
    {
	debug_log ("%1:%2: Getting the XML widgets", group, key);
	std::for_each (widgets.begin(), widgets.end(), boost::bind (&adaptor_type::link_widget, _1, settings));
	std::for_each (actions.begin(), actions.end(), boost::bind (&action_type::link_widget, _1, settings));
    }
    
    void link_callbacks() 
    {
	debug_log ("%1:%2: Linking UI callbacks", group, key);
	std::for_each (widgets.begin(), widgets.end(), boost::bind (&adaptor_type::link_callback, _1));
    }

    void release_callbacks() 
    {
	debug_log ("%1:%2: Releasing UI callbacks", group, key);
	std::for_each (widgets.begin(), widgets.end(), boost::bind (&adaptor_type::release_callback, _1));
    }

  private:
    void set_uis (const T& value)
    {
	debug_log ("%1:%2: Setting new value for UI: %3", group, key, value);
	std::for_each (widgets.begin(), widgets.end(), boost::bind (&adaptor_type::set_ui, _1, value));
    }

    void notify_actions (const T& value)
    {
	debug_log ("%1:%2: Notifying actions of new UI value: %3", group, key, value);
	std::for_each (actions.begin(), actions.end(), boost::bind (&action_type::on_action, _1, value));
    }

    void link_notifications()
    {
	debug_log ("%1:%2: Linking notification signals", group, key);
	std::for_each (widgets.begin(), widgets.end(), boost::bind (&adaptor_type::link_slot, _1, sigc::mem_fun (*this, &GenericSetting<T>::on_widgets_changed)));
    }

  public:
    void status_to_ui()
    {
	debug_log ("%1:%2: Promoting saved value to UI", group, key);
	T value;
	if (! get_value_checked (value)) value = defvalue;
	set_uis (value);
	on_widgets_changed (value);
    };
    
    void on_widgets_changed (const T& value)
    {
	debug_log ("%1:%2: Reacting on UI change to %3", group, key, value);
        set_value (value);
	settings.manager->set_hint (group, key, hint);
	settings.manager->set_description (group, key, description);
        notify_actions (value);
	release_callbacks();
	set_uis (value);
	link_callbacks();
    }
  
    std::string group, key, hint, description;
    signal_type signal_v;
    T defvalue;
    
    adaptorlist_type widgets;
    actionlist_type actions;
};

class AbstractMRUAdaptor {
  public:
    typedef sigc::signal<void, std::string> signal_type;
  
    /** Virtual inheritance.
     */

    virtual ~AbstractMRUAdaptor() {};
  
    virtual void link_widget (const Settings& settings) {};
    virtual void add_entry (const unsigned pos, const std::string& value) {};
    virtual void del_entry (const unsigned pos) {};    
    virtual signal_type get_signal() const { return signal_v; };
    
  protected:
    signal_type signal_v;
};

class MRUSetting : public AbstractSetting
{
  public:
    typedef sigc::signal<void, std::string> signal_type;

    typedef AbstractMRUAdaptor adaptor_type;
    // We need shared_ptrs here because otherwise inherited classes would get 
    // sliced down to adaptor_type when being copied by value
    typedef boost::shared_ptr<adaptor_type> adaptorptr_type;
 
    template <typename A>
    MRUSetting (Settings& settings, const std::string& group, const unsigned max_num,
      A widget) : 
      AbstractSetting (settings), 
      group (group), 
      max_num (max_num), 
      widget (new A (widget))
    {
	debug_log ("%1:%2: Loading up setting", group);
    }

    void add_mru (const std::string& value);
    
    void link_widgets() 
    {
	debug_log ("%1: Getting the XML widgets", group);
	widget->link_widget (settings);
    }
    
    void link_signals()
    {
	widget->get_signal().connect (signal_v.make_slot());
    }
    
    signal_type get_signal() const 
    {
	return signal_v;
    }

    void status_to_ui();
    
  private:
    typedef std::list<std::string> valuelist_type;
    std::string group;
    valuelist_type values;
    signal_type signal_v;
    unsigned max_num;
    
    adaptorptr_type widget;
};

#endif /* _SETTINGSHELPER_H */
