/***************************************************************************
 *            SynthesisAnalysis.cc
 *
 *  Mon Mar 21 17:07:57 2005
 *  Copyright  2005  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "SynthesisAnalysis.h"

#include "debug.h"

#include <boost/lexical_cast.hpp>

#include <fstream>
#include <sstream>
#include <iomanip>

#ifndef WIN32
#include <sys/wait.h>
#endif

SynthesisEvaluationEntry::operator const std::string() const
{
    std::ostringstream s;
    s << "<tt>";
    s << std::right << std::showpos << std::setprecision (1)
      << std::fixed << std::showpoint;
    s << std::setw (10) << label;
    s << ": ";
    s << std::setw (6) << len_score << " ";
    s << std::setw (6) << energy_score;
    s << "</tt>";
    return s.str();
}	

std::ostream& operator << (std::ostream& strm, const SynthesisDisplayLabel& label)
{
    strm << label.posx << " " << label.posy << " " << label.label << "\n";
    return strm;
}

std::istream& operator >> (std::istream& strm, SynthesisDisplayLabel& label)
{
    strm >> label.posx;
    strm >> label.posy;
    getline (strm, label.label);
    while (label.label.size() && label.label[0] == ' ') label.label.erase (0, 1);
    return strm;
}


std::istream& operator >> (std::istream& s, SynthesisEvaluationEntry& e)
{
    s >> e.label;
    s >> e.ref_from;
    s >> e.ref_to;
    s >> e.user_from;
    s >> e.user_to;
    s >> e.len_score;
    s >> e.energy_score;
    return s;
}

std::istream& operator >> (std::istream& s, TwoDeData::DataPtr& data)
{
    int dim, len;
    s >> dim >> len;
    if (! s) {
	data.reset();
	return s;
    }
    data.reset (new TwoDeData::Data (boost::extents[dim][len]));
    for (int j = 0; j < len; ++j)
      for (int i = 0; i < dim; ++i)
        s >> (*data)[i][j];
    return s;    
}

SynthesisAnalysis::SynthesisAnalysis() :
  was_shutdown (true)
{
    std::vector<std::string> params;
    params.push_back ("/home/mh21/Projects/record/record");
    params.push_back ("/tmp/direct_user.wav");
    
    try {
	Glib::spawn_async (".", params, 
	  Glib::SpawnFlags (Glib::SPAWN_DO_NOT_REAP_CHILD), 
	  sigc::slot<void>(), &pid);    
	was_shutdown = false;
    } catch (...) {};
}

SynthesisAnalysis::~SynthesisAnalysis() 
{
    shutdown();
}

void SynthesisAnalysis::shutdown() 
{
    if (was_shutdown) return;
    int status;
    #ifndef WIN32
    kill (pid, SIGHUP);
    waitpid (pid, &status, 0);
    #endif
    Glib::spawn_close_pid (pid);
    was_shutdown = true;
}

void SynthesisAnalysis::analyze() 
{
    shutdown();
    Glib::spawn_command_line_sync ("/home/mh21/Projects/sonogram4/diff2 /tmp/direct 3 3");
    get_file ("/tmp/direct_reference.data", ref, ref_labels);
    get_file ("/tmp/direct_user.data", user, user_labels);
    get_errors ("/tmp/direct.error", errors);
}

void SynthesisAnalysis::play_user() const
{
    const double from = errors.size() ? errors.begin()->user_from : -1;
    const double to = errors.size() ? (errors.end() - 1)->user_to : -1;
    play ("/tmp/direct_user.wav", from, to);
}

void SynthesisAnalysis::play_ref() const
{
    const double from = errors.size() ? errors.begin()->ref_from : -1;
    const double to = errors.size() ? (errors.end() - 1)->ref_to : -1;
    play ("/tmp/direct_reference.wav", from, to);
}

void SynthesisAnalysis::play (const std::string& name, const double from, const double to) const
{
    std::vector<std::string> params;
    params.push_back ("/usr/bin/play");
    params.push_back (name);
    if (from > 0 || to > 0) {
	params.push_back ("trim");
	if (from > 0)
	  params.push_back (boost::lexical_cast<std::string> (from));
	else
	  params.push_back ("0");
	if (to > 0)
	  params.push_back (boost::lexical_cast<std::string> (to - from));    
    }
    Glib::spawn_async (".", params);
}

void SynthesisAnalysis::get_file (const std::string& name, 
  TwoDeData::DataPtr& data, SynthesisDisplayLabelData& labels) const
{
    std::ifstream s (name.c_str());
    s >> data;
    if (data) 
      TwoDeData::normalize (*data);
    labels.clear();
    std::copy (std::istream_iterator<SynthesisDisplayLabel>(s), 
      std::istream_iterator<SynthesisDisplayLabel>(), back_inserter (labels));
}

void SynthesisAnalysis::get_errors (const std::string& name, 
  SynthesisEvaluationEntryData& data) const
{
    std::ifstream s (name.c_str());
    data.clear();
    std::copy (std::istream_iterator<SynthesisEvaluationEntry>(s), 
      std::istream_iterator<SynthesisEvaluationEntry>(), back_inserter (data));
}

std::string SynthesisAnalysis::get_text() const
{
    std::ostringstream s;
    s << "<tt>   Segment  Length Energy</tt>\n";
    std::copy (errors.begin(), errors.end(), 
      std::ostream_iterator<std::string> (s, "\n"));
    return s.str();
}
