/***************************************************************************
 *            TransientSettings.cc
 *
 *  Mon Nov 29 19:32:34 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "TransientSettings.h"
#include "TempSettingsManager.h"
#include "SettingsAdaptors.h"
#include "debug.h"

#include <gtkmm/combobox.h>
#include <gtkmm/liststore.h>
#include <gtkmm/checkbutton.h>
#include <gtkmm/radiobutton.h>
#include <gtkmm/toggletoolbutton.h>

// The real class

TransientSettings::TransientSettings() :
  Settings (SettingsManagerPtr (new TempSettingsManager())),
  presentsolution (*this, "transient", "acousticsolution",
    "", "", false, ToggleButtonAdaptor<bool> ("presentsolution")),
  usetimer (*this, "transient", "use_timer",
    "", "", false, ParamList<boost::shared_ptr <AbstractAdaptor<bool> > >
    (boost::shared_ptr<ToggleButtonAdaptor<bool> > (new ToggleButtonAdaptor<bool> ("usetimer")))
    (boost::shared_ptr<ToggleToolButtonAdaptor<bool> > (new ToggleToolButtonAdaptor<bool> ("tool_usetimer"))).get()),
  language_store (Gtk::ListStore::create (language_columns))
{
    debug_log ("Loading up transient settings"); 
    
    // This is for the automatic part
    init();
    
    // and all this for the manual one
    xml->get_widget("firstlanguage", firstlanguage);
    firstlanguage->signal_changed().connect (sigc::bind<int> (sigc::mem_fun (*this, &TransientSettings::on_language_changed), 0));
    firstlanguage->set_model(language_store);
    firstlanguage->pack_start(language_columns.name);
    
    xml->get_widget ("secondlanguage", secondlanguage);
    secondlanguage->signal_changed().connect (sigc::bind<int> (sigc::mem_fun (*this, &TransientSettings::on_language_changed), 1));
    secondlanguage->set_model (language_store);
    secondlanguage->pack_start (language_columns.name);
    
    xml->get_widget ("presentwritten", presentwritten);
    presentwritten->signal_toggled().connect (sigc::mem_fun (*this, &TransientSettings::on_presentwritten_changed));
    xml->get_widget ("presentaudio", presentaudio);
    presentaudio->signal_toggled().connect (sigc::mem_fun (*this, &TransientSettings::on_presentaudio_changed));
    xml->get_widget ("askwritten", askwritten);
    askwritten->signal_toggled().connect (sigc::mem_fun (*this, &TransientSettings::on_ask_changed));
    xml->get_widget ("askpick", askpick);
    askpick->signal_toggled().connect (sigc::mem_fun (*this, &TransientSettings::on_ask_changed));
    
    // Signals
    manager->get_signal ("transient", "first").connect (signal_languages_v.make_slot());
    manager->get_signal ("transient", "second").connect (signal_languages_v.make_slot());

    manager->get_signal ("transient", "audio").connect (signal_mode_v.make_slot());
    manager->get_signal ("transient", "written").connect (signal_mode_v.make_slot());
    manager->get_signal ("transient", "buttons").connect (signal_mode_v.make_slot());
    manager->get_signal ("transient", "use_timer").connect (signal_mode_v.make_slot());
}

/** Shut down the class. Currently only a debug message is printed.
 */

TransientSettings::~TransientSettings()
{
    debug_log ("Shutting down transient settings"); 
}

// Mode selection

sigc::signal<void> TransientSettings::signal_mode()
{
    return signal_mode_v;
}

void TransientSettings::on_presentwritten_changed()
{
    set_written (presentwritten->get_active());
}

void TransientSettings::on_presentaudio_changed()
{
    set_audio (presentaudio->get_active());
}

void TransientSettings::on_ask_changed()
{
    set_buttons (askpick->get_active());
}

void TransientSettings::set_audio (const bool value, const bool silent)
{
    bool oldval = is_audio();
    if (oldval == value && ! silent) return;
    // Set the new value
    manager->set_bool ("transient", "audio", value);
    // Try to set other checkbox if its necessary
    if (! silent && ! value && ! is_written())
      set_written (true, true);
    // And make sure we display it right
    if (silent && presentaudio->get_active() != value)
      presentaudio->set_active (value);
}

void TransientSettings::set_written (const bool value, const bool silent)
{
    bool oldval = is_written();
    if (oldval == value && ! silent) return;
    // Set the new value
    manager->set_bool ("transient", "written", value);
    // Try to set other checkbox if its necessary
    if (! silent && ! value && ! is_audio())
      set_audio (true, true);
    // And make sure we display it right
    if (silent && presentwritten->get_active() != value)
      presentwritten->set_active (value);
}

void TransientSettings::set_buttons (const bool value, const bool silent)
{
    bool oldval = is_buttons();
    if (oldval == value && ! silent) return;
    // Set the new value
    manager->set_bool ("transient", "buttons", value);
    // And make sure we display it right
    if (silent && (askpick->get_active() != value)) {
        if (value)
	  askpick->set_active();
	else
	  askwritten->set_active();
    }
}

bool TransientSettings::is_audio() const
{
    return manager->get_bool ("transient", "audio");
}

bool TransientSettings::is_written() const
{
    return manager->get_bool ("transient", "written");
}

bool TransientSettings::is_buttons() const
{
    return manager->get_bool ("transient", "buttons");
}

bool TransientSettings::is_input() const
{
    return ! manager->get_bool ("transient", "buttons");
}

void TransientSettings::set_mode (const bool audio, const bool written, const bool buttons)
{
    set_audio (audio, true);
    set_written (written, true);
    set_buttons (buttons, true);
}

// Language listboxes

void TransientSettings::on_language_changed (const int which)
{
    if (language_store->children().empty()) return;
    if (!which && firstlanguage->get_active()) {
	set_first (firstlanguage->get_active()->get_value (language_columns.name));
    } else if (which && secondlanguage->get_active()) {
	set_second (secondlanguage->get_active()->get_value (language_columns.name));
    }
}

sigc::signal<void> TransientSettings::signal_languages()
{
    return signal_languages_v;
}

std::string TransientSettings::get_first() const
{
    return manager->get_string ("transient", "first");
}

std::string TransientSettings::get_second() const
{
    return manager->get_string ("transient", "second");
}

// silent means: we don't care about old values, we don't try to update the other combobox
// and we make sure we show the right entry
// TODO: this code is ugly and there should be a lot of situations where it doesn't work
// the framework should be changed altogether anyway
void TransientSettings::set_first (const std::string& value, const bool silent)
{
    std::string oldval = get_first();
    if (oldval == value && ! silent) return;
    if (! combo_find (value)) {
	if (language_store->children().size()) {
	    const std::string newvalue = language_store->children()[0].get_value (language_columns.name);
	    manager->set_string ("transient", "first", newvalue);
	    firstlanguage->set_active (0);
	}
    } else {
	// Set the new value
	manager->set_string ("transient", "first", value);
	// Try to change other combobox if its the same
	if (! silent && oldval != "" && value == get_second())
	  set_second (oldval, true);
	// And make sure we display the right entry
	if (silent && (! firstlanguage->get_active() || firstlanguage->get_active()->get_value (language_columns.name) != value))
	  combo_set (firstlanguage, value);
    }
}

void TransientSettings::set_second (const std::string& value, const bool silent)
{
    std::string oldval = get_second();
    if (oldval == value && ! silent) return;
    if (! combo_find (value)) {
	if (language_store->children().size()) {
	    std::string newvalue;
	    if (language_store->children().size() > 1)
	      newvalue = language_store->children()[1].get_value (language_columns.name);
	    else
	      newvalue = language_store->children()[0].get_value (language_columns.name);
	    manager->set_string ("transient", "second", newvalue);
	    secondlanguage->set_active (0);
	}
    } else {
	// Set the new value
	manager->set_string ("transient", "second", value);
	// Try to change other combobox if its the same
	if (! silent && oldval != "" && value == get_first())
	  set_first (oldval, true);
	// And make sure we display the right entry
	if (silent && (! secondlanguage->get_active() || secondlanguage->get_active()->get_value (language_columns.name) != value))
	  combo_set (secondlanguage, value);
    }
}

void TransientSettings::combo_set (Gtk::ComboBox* box, const std::string& value)
{
    Gtk::TreeModel::Children c = language_store->children();
    Gtk::TreeModel::const_iterator i = std::find_if (c.begin(), c.end(), 
      bind (std::equal_to<Glib::ustring>(), 
        bind (&Gtk::TreeRow::get_value<Glib::ustring>, _1, language_columns.name), 
        value));
    if (i != c.end()) box->set_active (i);
}

bool TransientSettings::combo_find (const std::string& value)
{
    Gtk::TreeModel::Children c = language_store->children();
    Gtk::TreeModel::const_iterator i = std::find_if (c.begin(), c.end(), 
      bind (std::equal_to<Glib::ustring>(), 
        bind (&Gtk::TreeRow::get_value<Glib::ustring>, _1, language_columns.name), 
        value));
    return i != c.end();
}

void TransientSettings::append_language (const std::string& lang) 
{
    language_store->append()->set_value (language_columns.name, Glib::ustring (lang));
}

// Session management

sigc::signal<void> TransientSettings::signal_testmode()
{
    return manager->get_signal ("transient", "testmode");
}

void TransientSettings::set_testmode (const bool value)
{
    manager->set_bool ("transient", "testmode", value);
}

bool TransientSettings::is_testmode() const
{
    return manager->get_bool ("transient", "testmode");
}

sigc::signal<void> TransientSettings::signal_resultsshown()
{
    return manager->get_signal ("transient", "resultsshown");
}

void TransientSettings::set_resultsshown (const bool value)
{
    manager->set_bool ("transient", "resultsshown", value);
}

bool TransientSettings::is_resultsshown() const
{
    return manager->get_bool ("transient", "resultsshown");
}
