/***************************************************************************
 *            TwoDeData.cc
 *
 *  Fri Jan 28 14:44:13 2005
 *  Copyright  2005  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "TwoDeData.h"

#include "debug.h"

#include "functions.h"

#include <boost/bind.hpp>

TwoDeData::DataRef TwoDeData::ref (Data& data)
{
    return DataRef (data);
}

TwoDeData::ConstDataRef TwoDeData::ref (const Data& data)
{
    return ConstDataRef (data);
}

TwoDeData::FlatRef TwoDeData::col (Data& data, const unsigned num)
{
    TwoDeData::FlatRef result = data[boost::indices[Data::index_range()][num]];
    result.reindex (data.index_bases()[0]);
    return result;
}

TwoDeData::FlatRef TwoDeData::row (Data& data, const unsigned num)
{
    FlatRef result = data[boost::indices[num][Data::index_range()]];
    result.reindex (data.index_bases()[1]);
    return result;
}

TwoDeData::ConstFlatRef TwoDeData::col (const Data& data, const unsigned num)
{
    ConstFlatRef result = data[boost::indices[Data::index_range()][num]];
    result.reindex (data.index_bases()[0]);
    return result;
}

TwoDeData::ConstFlatRef TwoDeData::row (const Data& data, const unsigned num)
{
    TwoDeData::ConstFlatRef result = data[boost::indices[num][Data::index_range()]];
    result.reindex (data.index_bases()[1]);
    return result;
}

TwoDeData::FlatRef TwoDeData::flatten (Data& data)
{
    DataRef a = ref (data);
    return flatten (a);
}

TwoDeData::ConstFlatRef TwoDeData::flatten (const Data& data)
{
    ConstDataRef a = ref (data);
    return flatten (a);
}

TwoDeData::FlatRef TwoDeData::flatten (DataRef& data)
{
    unsigned maxi = data.shape()[0];
    unsigned maxj = data.shape()[1];

    DataRef dataref (data);
    boost::array<DataRef::index,2> dims = {{1, maxi * maxj}};  
    dataref.reshape (dims);
    dataref.reindex (0);
    return dataref[boost::indices[0][DataRef::index_range()]];
}

TwoDeData::ConstFlatRef TwoDeData::flatten (const ConstDataRef& data)
{
    unsigned maxi = data.shape()[0];
    unsigned maxj = data.shape()[1];

    ConstDataRef dataref (data);
    boost::array<ConstDataRef::index,2> dims = {{1, maxi * maxj}};  
    dataref.reshape (dims);
    dataref.reindex (0);
    return dataref[boost::indices[0][ConstDataRef::index_range()]];
}

TwoDeData::DataPtr TwoDeData::wrap (const ConstFlatRef& data)
{
    DataPtr result (new Data (boost::extents[1][Data::extent_range (
      data.index_bases()[0], data.index_bases()[0] + data.shape()[0])]));
    debug_log ("wraping flat array of [%1,%2)", data.index_bases()[0], 
      data.index_bases()[0] + data.shape()[0]);
    (*result)[0] = data;
    return result;
}

TwoDeData::DataPtr TwoDeData::diff (const Data& data1, const Data& data2)
{  
    unsigned maxi = data1.shape()[1];
    unsigned maxj = data2.shape()[1];
    debug_log ("Diff with %1, %2", maxi, maxj);
    DataPtr B (new Data (boost::extents[maxi][maxj]));

    // We can't handle data that doesn't start with 0
    for (unsigned i = 0; i < maxi; ++i) {
	const ConstFlatRef sp1 = col (data1, i);
	for (unsigned j = 0; j < maxj; ++j) {
	    const ConstFlatRef sp2 = col (data2, j);
	    (*B)[i][j] = sqrt (std::inner_product (sp1.begin(), sp1.end(), sp2.begin(), 0.0, 
	      std::plus<double>(), 
	      boost::bind (std::multiplies<double>(), 
	        boost::bind (std::minus<double>(), _1, _2), 
	        boost::bind (std::minus<double>(), _1, _2))));
	}
    }
    return B;
}

void TwoDeData::gamma (Data& data, const double gamma) 
{
    debug_log ("Gamma with %1", gamma);
    FlatRef B = flatten (data);
    std::transform (B.begin(), B.end(), B.begin(), func::Gamma<double> (gamma));
}

void TwoDeData::gamma (const Data& data1, Data& data2, const double gamma) 
{
    debug_log ("Gamma with %1", gamma);
    ConstFlatRef b1 = flatten (data1);
    FlatRef b2 = flatten (data2);
    std::transform (b1.begin(), b1.end(), b2.begin(), func::Gamma<double> (gamma));
}

void TwoDeData::normalize (Data& data) 
{
    debug_log ("Normalize");
    if (! data.num_elements()) return;
    FlatRef B = flatten (data);
    double max = *std::max_element (B.begin(), B.end());
    std::transform (B.begin(), B.end(), B.begin(), boost::bind (std::divides<double>(), _1, max));
}

void TwoDeData::normalize_neg (Data& data) 
{
    debug_log ("Normalize");
    if (! data.num_elements()) return;
    FlatRef B = flatten (data);
    const double max1 = std::abs (*std::min_element (B.begin(), B.end()));
    const double max2 =           *std::max_element (B.begin(), B.end());
    const double max = std::max (max1, max2);
    std::transform (B.begin(), B.end(), B.begin(), boost::bind (std::divides<double>(), _1, max));
}

void TwoDeData::normalize_log (Data& data, const double scale) 
{
    debug_log ("Normalize");
    if (! data.num_elements()) return;
    FlatRef B = flatten (data);
    const double max = *std::max_element (B.begin(), B.end()) / scale - 1;
    std::transform (B.begin(), B.end(), B.begin(), 
      boost::bind (std::minus<double>(), 
        boost::bind (std::divides<double>(), _1, scale),
        max));
}

void TwoDeData::normalize_log (const Data& data1, Data& data2, const double scale) 
{
    debug_log ("Normalize");
    if (! data1.num_elements()) return;
    ConstFlatRef B = flatten (data1);
    FlatRef B2 = flatten (data2);
    const double max = *std::max_element (B.begin(), B.end()) / scale - 1;
    std::transform (B.begin(), B.end(), B2.begin(), 
      boost::bind (std::minus<double>(), 
        boost::bind (std::divides<double>(), _1, scale),
        max));
}

void TwoDeData::normalize_log (Data& data, const double scale, const double max) 
{
    debug_log ("Normalize");
    if (! data.num_elements()) return;
    FlatRef B = flatten (data);
    std::transform (B.begin(), B.end(), B.begin(), 
      boost::bind (std::minus<double>(), 
        boost::bind (std::divides<double>(), _1, scale),
        max / scale - 1));
}

void TwoDeData::normalize_log (const Data& data1, Data& data2, const double scale, const double max) 
{
    debug_log ("Normalize");
    if (! data1.num_elements()) return;
    ConstFlatRef B = flatten (data1);
    FlatRef B2 = flatten (data2);
    std::transform (B.begin(), B.end(), B2.begin(), 
      boost::bind (std::minus<double>(), 
        boost::bind (std::divides<double>(), _1, scale),
        max / scale - 1));
}

void TwoDeData::normalize (const Data& data1, Data& data2) 
{
    debug_log ("Normalize");
    if (! data1.num_elements()) return;
    ConstFlatRef b1 = flatten (data1);
    FlatRef b2 = flatten (data2);
    double max = *std::max_element (b1.begin(), b1.end());
    std::transform (b1.begin(), b1.end(), b2.begin(), boost::bind (std::divides<double>(), _1, max));
}

void TwoDeData::normalize_neg (const Data& data1, Data& data2) 
{
    debug_log ("Normalize");
    if (! data1.num_elements()) return;
    ConstFlatRef b1 = flatten (data1);
    FlatRef b2 = flatten (data2);
    const double max1 = std::abs (*std::min_element (b1.begin(), b1.end()));
    const double max2 =           *std::max_element (b1.begin(), b1.end());
    const double max = std::max (max1, max2);
    std::transform (b1.begin(), b1.end(), b2.begin(), boost::bind (std::divides<double>(), _1, max));
}

void TwoDeData::clear (Data& data) 
{
    FlatRef B = flatten (data);
    std::fill (B.begin(), B.end(), 0);
}
