/***************************************************************************
 *            VocabCache.cc
 *
 *  Sat Feb  5 13:56:10 2005
 *  Copyright  2005  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "VocabCache.h"
#include "ProgramPaths.h"
#include "functions.h"
#include "debug.h"
#include "stringhelper.h"

#include <glibmm/fileutils.h>
#include <glibmm/miscutils.h>
#include <glibmm/markup.h>

#include <ctime>
#include <fstream>
#include <map>

/** \file
 * Caches vocabulary scores.
 */

#ifndef DOCYGEN_SHOULD_SKIP_THIS

namespace
{
    class VocabCacheImpl
    {
      public:
	class VocabData
	{
	  public:
	    VocabData() :
	      score (1), 
	      expires (time (NULL))
	    {
	    };
	    
	    void expire_now()
	    {
		expire (0);
	    }
	    	    
	    void expire (const unsigned minutes)
	    {
		set_expire (time (NULL) + 60 * minutes);
	    }
	    
	    void set_expire (const time_t& new_expires)
	    {
		expires = new_expires;
	    }

	    time_t get_expire() const
	    {
		return expires;
	    }
	    
	    bool expired() const
	    {
		return expires <= time (NULL);
	    }
	    
	    double score;
	    
	  private:
	    time_t expires;
	};
    
	typedef std::map<unsigned, VocabData> VocabMap;
	  
	VocabCacheImpl (const std::string& filename);
	void load();
	void save() const;
	void update (const unsigned id, const bool dir);
	void expire();
	static void expire_vocab (VocabMap::value_type& v);
      
	VocabMap vocabs;
	std::string filename;
	func::QuadraticDecay<double> decay;
	unsigned days;
    };
    
    class VocabCacheParser : public Glib::Markup::Parser
    {
      public:
	VocabCacheParser (VocabCacheImpl::VocabMap& data);
    
      protected:
	virtual void on_start_element (Glib::Markup::ParseContext& context,
	  const Glib::ustring& element_name, const AttributeMap& attributes);
	virtual void on_end_element (Glib::Markup::ParseContext& context,
	  const Glib::ustring& element_name) {};
	virtual void on_text (Glib::Markup::ParseContext& context, 
	  const Glib::ustring& text);
    
      private:
	VocabCacheImpl::VocabMap& data;
	unsigned last_id;
        time_t now;
    };
    
    std::ostream& operator << (std::ostream& strm, const VocabCacheImpl::VocabMap::value_type& i);
    std::ostream& operator << (std::ostream& strm, const VocabCacheImpl& d);
    
    VocabCacheImpl::VocabCacheImpl (const std::string& filename) :
      filename (filename),
      decay (1, 5),
      days (28)
    {
    }
    
    void VocabCacheImpl::load()
    {    
	debug_log ("Reading vocabulary cache from %1", filename);
	try {
	    VocabCacheParser parser (vocabs);
	    Glib::Markup::ParseContext context (parser);
	    context.parse (Glib::file_get_contents (filename));
	    context.end_parse();
	} catch (...) {};
    }
    
    void VocabCacheImpl::save() const
    {
	debug_log ("Writing vocabulary cache to %1", filename);
	std::ofstream file (filename.c_str());
	if (!file) return;
	file << *this;
    }
    
    void VocabCacheImpl::update (const unsigned id, const bool dir)
    {
	VocabData& vocab = vocabs[id];
	// We calculate new if it was:
	//   expired and wrong or right
	//   not expired and wrong
	if (vocab.expired() || ! dir)
	  vocab.score = decay (vocab.score, dir);
	// We expire now if it was wrong
	if (! dir)
	  vocab.expire_now();
	// We expire later if it was right and expired
	else if (vocab.expired())
	  vocab.expire (static_cast<unsigned> (days * 60 * 24 * (1 - vocab.score)));
    }
    
    void VocabCacheImpl::expire()
    {
	std::for_each (vocabs.begin(), vocabs.end(), &VocabCacheImpl::expire_vocab);
    }
    
    void VocabCacheImpl::expire_vocab (VocabMap::value_type& v)
    {
	v.second.expire_now();
    }

    std::ostream& operator << (std::ostream& strm, const VocabCacheImpl::VocabMap::value_type& i)
    {
	strm << "    <score id=\"" << i.first << "\" expires=\"" <<
	  i.second.get_expire() << "\">" << i.second.score << "</score>\n";
	return strm;
    }
    
    std::ostream& operator << (std::ostream& strm, const VocabCacheImpl& d) { 
	strm << "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n";
	strm << "<!DOCTYPE cache SYSTEM \"cache.dtd\">\n";
	strm << "<cache>\n";
	std::copy (d.vocabs.begin(), d.vocabs.end(), 
	  std::ostream_iterator<VocabCacheImpl::VocabMap::value_type> (strm));
	strm << "</cache>\n";
	return strm; 
    };
    
    VocabCacheParser::VocabCacheParser (VocabCacheImpl::VocabMap& data) :
      data (data),
      last_id (0),
      now (time (NULL))
    {
    }
    
    void VocabCacheParser::on_start_element (Glib::Markup::ParseContext& context,
      const Glib::ustring& element_name, const AttributeMap& attributes)
    {
	if (element_name == "score") {
	    AttributeMap::const_iterator id = attributes.find ("id");
	    if (id != attributes.end())
	      last_id = from_string (id->second, 0);
	    else
	      last_id = 0;
	    AttributeMap::const_iterator expires = attributes.find ("expires");
	    if (expires != attributes.end())
	      data[last_id].set_expire (from_string (expires->second, now));
	}
    }
    
    void VocabCacheParser::on_text (Glib::Markup::ParseContext& context, 
      const Glib::ustring& text)
    {
	Glib::ustring trimmed = trim_whitespace (text);
	if (!trimmed.empty()) 
	  data[last_id].score = from_string (text, 0.0);
    }    
}

#endif

/** Initialize and load the saved cache.
 * The vocabulary cache is loaded from the file <tt>.voctraincache</tt> in the 
 * users home directory.
 */

VocabCache::VocabCache() :
  impl (new VocabCacheImpl(Glib::build_filename (ProgramPaths::home(), ".vocabtraincache")))
{
    
    impl->load();
}

/** Destruct and flush the cache data to disk.
 */

VocabCache::~VocabCache()
{
    impl->save();
}

/** Set the number of answers needed for the score calculation.
 * The used equation is documented at func::QuadraticDecay. The
 * parameter \c k is currently fixed at 5.
 * \param new_repeats New number \c N
 */

void VocabCache::set_repeats (const unsigned new_repeats)
{
    impl->decay.set (new_repeats, 5);
}

/** Set the maximum period until a vocabulary expires.
 * Expired vocabularies get a virtual score of 1 until they are 
 * answered correct again.
 * \param days Period in days
 */

void VocabCache::set_expires (const unsigned days)
{
    impl->days = days;
}

/** Clear the cache.
 */

void VocabCache::clear()
{
    impl->vocabs.clear();
}

/** Expire all vocabularies.
 */

void VocabCache::expire()
{
    impl->expire();
}

/** Record a wrong answer.
 * The score is adjusted automatically.
 * \param id %Vocabulary id
 */

void VocabCache::wrong (const unsigned id)
{
    impl->update (id, false);
}

/** Record a right answer.
 * The score and the expire date is adjusted automatically.
 * \param id %Vocabulary id
 */

void VocabCache::right (const unsigned id)
{
    impl->update (id, true);
}

/** Get the current score value.
 * \param id %Vocabulary id
 * \return Score value or 1 if not found
 */

double VocabCache::get_score (const unsigned id) const
{
    VocabCacheImpl::VocabMap::const_iterator i;
    if ((i = impl->vocabs.find (id)) != impl->vocabs.end())
      return i->second.score;
    else
      return VocabCacheImpl::VocabData().score;
}

/** Get the current score value.
 * Different from get_score(), an expired vocabulary returns
 * a score of 1 here.
 * \param id %Vocabulary id
 * \return Score value or 1 if not found or expired
 */

double VocabCache::get_score_with_expires (const unsigned id) const
{
    VocabCacheImpl::VocabMap::const_iterator i;
    if ((i = impl->vocabs.find (id)) != impl->vocabs.end())
      return i->second.expired() ? 1 : i->second.score;
    else
      return VocabCacheImpl::VocabData().score;
}

/** Get the date and time when a vocabulary expires.
 * \param id %Vocabulary id
 * \return Expire date and time in seconds since 1970 or the current
 * time if not found
 */

time_t VocabCache::get_expires (const unsigned id) const
{
    VocabCacheImpl::VocabMap::const_iterator i;
    if ((i = impl->vocabs.find (id)) != impl->vocabs.end())
      return i->second.get_expire();
    else
      return time (NULL);
}
