/***************************************************************************
 *            VocabChecker.cc - Check whether a guess is contained in the 
 *                              saved translation
 *
 *  Sat Nov 20 19:11:00 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "VocabChecker.h"

#include "debug.h"

#include <iostream>

/** 
 * Checks a given user input against the possible translations. Several special tags are recognized:
 * - { starts a grammar group. Can be used for grammatical information.
 * - } ends a grammar group.
 * - ( starts a comment group. Can be used for additional words.
 * - ) ends a comment group.
 * - ; separates different translations.
 * 
 * Whitespace is removed.
 */
bool VocabChecker::check (const Glib::ustring& test, const Glib::ustring& base) 
{
    const std::string base_string = base.casefold();
    const separator base_sep (" ", ";(){}");
    const tokenizer base_tokens (base_string, base_sep);
    VocabChecker vocab_checker = std::for_each (base_tokens.begin(), base_tokens.end(), 
      VocabChecker (test));
    return vocab_checker.get_result();
}
    
VocabChecker::VocabChecker (const Glib::ustring& test) :
  test_str (test.casefold()),
  test_sep (" "),
  test_tokens (test_str, test_sep),
  test_len (std::distance (test_tokens.begin(), test_tokens.end())),
  base_comment (false),
  base_grammar (false),
  result (false)
{
    debug_log ("Init VocabChecker with %1", test);
}

bool VocabChecker::operator() (const std::string& base_token) 
{
    if (result) return true;
	
    if (base_token == "{") {
	base_grammar = true;
    } else if (base_token == "}") {
	base_grammar = false;
    } else if (base_token == "(") {
	base_comment = true;
    } else if (base_token == ")") {
	base_comment = false;
    } else if (base_token == ";") {
	if (does_match()) {
	    result = true;
	    return true;
	}
	base_part.clear();
	base_part_comment.clear();
	base_comment = false;
	base_grammar = false;
    } else if (! base_grammar) {
	base_part.push_back (base_token);
	base_part_comment.push_back (base_comment);
    };
    return false;
}

bool VocabChecker::get_result() const {
    debug_log ("Result %1", result);
    return result || does_match();
}

bool VocabChecker::does_match() const {
    tokenizer::const_iterator test = test_tokens.begin();
    std::list<std::string>::const_iterator base = base_part.begin();
    std::list<bool>::const_iterator comment = base_part_comment.begin();
    // This will fail with sth. like "(test) test" <=> user: "test",
    // because it already matches the first pattern
    // TODO: Use boost regexp
    for (; base != base_part.end(); ++base, ++comment) {
	if (*comment) {
	    if (test != test_tokens.end() && *test == *base) 
	      ++test;
	} else {
	    if (test == test_tokens.end() || *test != *base) 
	      return false;
	    else
	      ++test;
	}
    }
    return test == test_tokens.end();
}
