/***************************************************************************
 *            VocabFormatter.cc - Format a vocabulary entry in various ways
 *
 *  Sat Nov 20 19:11:00 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "debug.h"

#include "VocabFormatter.h"
#include "StressBaseInterface.h"

#include <glibmm/markup.h>

#include <sstream>

std::string VocabFormatter::format (const std::string& desc, const int type)
{
    return format (desc, type, ";", "");
}
    
/** 
 * Formats a given string for presentation using pango markup, depending on a requested size.
 * There are 3 sizes available:
 * - size 0:
 *   - { ... } groups become smaller and italized.
 *   - ( ... ) are not bold.
 *   - ; becomes ... \n or \n ...
 *   - Whitespace is removed, "," and "." are recognized.
 * - size 1: like size 0, except
 *   - ; becomes space or space
 * - size 2: very small
 *   - only matching words remain
 * - size 3: like size 2, without any formatting (no escaping done)
 */
std::string VocabFormatter::format (const std::string& desc, const int type, const std::string& ortext,
  const std::string& language)
{
    const separator sep (" ", ",.;(){}");
    const tokenizer tokens (desc, sep);
    std::string result;
    std::ostringstream res;
    std::transform (tokens.begin(), tokens.end(), std::ostream_iterator<std::string> (res), 
      VocabFormatter (type, ortext, language));
    return res.str();
}

VocabFormatter::VocabFormatter (const int type, const std::string& ortext, const std::string& language) :
    ortext (ortext),
    language (language),
    type (type),
    space (false),
    gram (false),
    unused (false),
    orspace (ortext != ";" && ortext != ",")
{
}

std::string VocabFormatter::operator() (const std::string& token) 
{
    if (token == "{") {
	gram = true;
	return "";
    } else if (token == "}") {
	gram = false;
	return "";
    } else if (token == "(") {
	unused = true;
	if (type < 2) {
	    if (space) {
		space = false;
		return " (";
	    } else return "(";
	} else return "";
    } else if (token == ")") {
	unused = false;
	if (type < 2) {
	    space = true;
	    return ")";
	} else return "";
    } else if (token == ";") {
	std::string result;
	switch (type) {
	    case 0: result += "\n"; break;
	    default: if (space && orspace) result += " "; break;
	}
	if (type <= 1) result += "<i>";
	if (type < 3)
          result += Glib::Markup::escape_text (ortext);
	else
	  result += ortext;
	if (type <= 1) result += "</i>";
	switch (type) {
	    case 0: result += "\n"; space = false; break;
	    default: space = true; break;
	}
	return result;
    } else  {
	if (type < 2 || (! gram && ! unused)) {
	    std::string result;
	    if (token != "," && token != "." && space) result += " ";
	    if (type == 0 && ! gram) result += "<big>";
	    if (! (unused || gram) && (type != 3)) result += "<b>";
	    if (gram) result += "<i><small>";
	    if (type < 3) {
		int stresspos = StressBaseInterface::get().get_stress_pos (token, language);
		if (stresspos != -1) {
		    debug_log ("Stress position of %1 is %2", token, stresspos);
                    Glib::ustring token_utf8 = token;
		    result += Glib::Markup::escape_text (token_utf8.substr (0, stresspos - 1));
		    result += "<u>";
		    result += Glib::Markup::escape_text (token_utf8.substr (stresspos - 1, 1));
		    result += "</u>";
		    result += Glib::Markup::escape_text (token_utf8.substr (stresspos));
		} else {
		    result += Glib::Markup::escape_text (token);
		}
	    } else {
	        result += token;
	    }
	    if (gram) result += "</small></i>";
	    if (! (unused || gram) && (type != 3)) result += "</b>";
	    if (type == 0 && ! gram) result += "</big>";
	    space = true;
	    return result;
	} else return "";
    };
}
