/***************************************************************************
 *            vocabparser.cc
 *
 *  Tue Oct 12 17:32:22 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "VocabParser.h"

#include "stringhelper.h"

template <typename K, typename V, typename C> V VocabParser::get_map 
  (const std::map<K, V, C>& map, const K& key)
{
    typename std::map<K, V, C>::const_iterator i;
    if ((i = map.find (key)) != map.end())
      return i->second;
    else
      return V();
}

VocabParser::VocabParser (DataChapters& chapters, VocabTypesPtr types, 
  VocabTest::LanguageSet& languages, VocabTest::PropertyMap& properties) :
  chapters(chapters),
  types (types),
  languages (languages),
  properties (properties)
{
}

VocabParser::~VocabParser()
{
}

void VocabParser::on_start_element (Glib::Markup::ParseContext& context,
  const Glib::ustring& element_name, const AttributeMap& attributes)
{
    if (element_name == "course") {
	chapters.push_back (DataChapterPtr (
	  new DataChapter (get_map (attributes, Glib::ustring ("title")), 
	  from_string (get_map (attributes, Glib::ustring ("difficulty")), 1))));
	cur_chapter = *(chapters.end() - 1);
    } else if (element_name == "lesson") {
	cur_chapter->sections.push_back (DataSectionPtr (
	  new DataSection (get_map (attributes, Glib::ustring ("title")))));
	cur_section = *(cur_chapter->sections.end() - 1);
    } else if (element_name == "vocabulary") {
	cur_section->vocabs.push_back (VocabularyPtr (
	  new Vocabulary (types, from_string (get_map (attributes, Glib::ustring 
	  ("id")), 0))));
	cur_vocab = *(cur_section->vocabs.end() - 1);
    } else if (element_name == "translation") {
	last_lang = get_map (attributes, Glib::ustring ("lang"));
    } else if (element_name == "property") {
	last_property = get_map (attributes, Glib::ustring ("title"));
    }
    last_node = element_name;
}

void VocabParser::on_text (Glib::Markup::ParseContext& context, 
  const Glib::ustring& text)
{
    Glib::ustring trimmed = trim_whitespace (text);
    if (!trimmed.empty()) {
	if (last_node == "translation") {
	    if (cur_vocab)
	      cur_vocab->set_translation (last_lang, trimmed);
	    languages.insert (last_lang);
	} else if (last_node == "property") {
	    properties.insert (std::make_pair (last_property, trimmed));
	} else {
            if (cur_vocab)
	      cur_vocab->set_type (trimmed);
	}
    }
}

std::ostream& operator << (std::ostream& strm, const VocabularyPtr& d) {
    strm << "      <vocabulary id=\"" << d->id << "\">\n";
    transform (d->translations.begin(), d->translations.end(), std::ostream_iterator<std::string> (strm), 
      Vocabulary::print_trans<Vocabulary::LanguageMap::value_type>);
    strm << "        <type>" << Glib::Markup::escape_text (d->get_type()) << "</type>\n";
    strm << "      </vocabulary>\n";
    return strm; 
}

std::ostream& operator << (std::ostream& strm, const DataSectionPtr& d) {
    strm << "    <lesson title=\"" << Glib::Markup::escape_text (d->get_title()) << "\">\n";
    copy (d->vocabs.begin(), d->vocabs.end(), std::ostream_iterator<const VocabularyPtr> (strm));
    strm << "    </lesson>\n";
    return strm; 
}

std::ostream& operator << (std::ostream& strm, const DataChapterPtr& d) {
    strm << "  <course difficulty=\"" << d->difficulty << 
      "\" title=\"" << Glib::Markup::escape_text (d->get_title()) << "\">\n";
    copy (d->sections.begin(), d->sections.end(), std::ostream_iterator<const DataSectionPtr> (strm));
    strm << "  </course>\n";
    return strm; 
}

std::ostream& operator << (std::ostream& strm, const DataChapters& d) { 
    copy (d.begin(), d.end(), std::ostream_iterator<const DataChapterPtr> (strm));
    return strm; 
};

std::string wrap_prop (const VocabTest::PropertyMap::value_type& i)
{
    return 
      "    <property title=\"" + i.first + "\">\n" + 
      "       " + i.second + "\n" + 
      "    </property>\n";
}

std::string wrap_lang (const VocabTest::LanguageSet::value_type& i)
{
    return 
      "    <language name=\"" + i + "\" />\n";
}

std::ostream& operator << (std::ostream& strm, const VocabTest& d) { 
    strm << "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n";
    strm << "<!DOCTYPE training SYSTEM \"vocab.dtd\">\n";
    strm << "<?xml-stylesheet type=\"text/xsl\" href=\"vocab.xsl\" ?>\n";
    strm << "<training>\n";
    strm << "  <properties>\n";
    transform (d.properties.begin(), d.properties.end(), 
      std::ostream_iterator<std::string> (strm), 
      &wrap_prop);
    strm << "  </properties>\n";
    strm << "  <languages>\n";
    transform (d.languages.begin(), d.languages.end(), 
      std::ostream_iterator<std::string> (strm), 
      &wrap_lang);
    strm << "  </languages>\n";
    strm << d.chapters;
    strm << "</training>\n";
    return strm; 
};
