/***************************************************************************
 *            VocabSession.cc
 *
 *  Fri Nov 26 13:39:40 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "VocabSession.h"
#include "VocabFormatter.h"

#include "support.h"

#include "debug.h"

#include <glib.h>

#include <boost/bind.hpp>
#include <boost/utility.hpp>

#include <stdlib.h>

VocabSession::VocabSession (const SessionFeederPtr& feeder) :
  feeder (feeder)
{
}

void VocabSession::set_mode (const bool random)
{
    random_order = random;
}

void VocabSession::set_interval (const unsigned num)
{
    interval = num;
}

void VocabSession::set_feed_continuous (const bool feed)
{
    feed_continuous = feed;
}

// gives only an estimation
unsigned int VocabSession::get_vocabs_left() const
{
   return vocabs.size() + sooner.size() + defered.size() + feeder->size();

}

bool VocabSession::start()
{
    vocab_old = SessionVocabPtr();
    return feed();
}

bool VocabSession::advance()
{
    debug_log ("Advancing to next vocabulary with vocabs %1 sooner %2 defered %3", 
      vocabs.size(), sooner.size(), defered.size());
    
    // there are none left, try to feed new ones
    if (! vocabs.size()) {
	return feed();
    }
    
    history.insert (vocabs [0]);
    vocab_old = vocabs [0];
    vocabs.erase (vocabs.begin());
	 
    if (! vocabs.size()) {
	debug_log ("We are at the end");
	if (feed_continuous && feed()) {
	} else if (sooner.size()) {
	    debug_log ("Copying sooner");
	    if (random_order) std::random_shuffle (sooner.begin(), sooner.end(), generic_random);
	    vocabs.insert (vocabs.end(), sooner.begin(), sooner.end());
	    sooner.clear();
	} else if (defered.size()) {
	    debug_log ("Copying defered");
	    if (random_order) std::random_shuffle (defered.begin(), defered.end(), generic_random);
	    vocabs.insert (vocabs.end(), defered.begin(), defered.end());
	    defered.clear();
	}

    }
    
    return vocabs.size() > 0;
}

bool VocabSession::feed()
{
    SessionVocabPtr vocab;
    // If we don't have timer mode, we feed one by one to make it 
    // easier to switch to timer mode; it will be instantly activated
    // after the sooner and defered bufers are empty
    for (unsigned i = vocabs.size(); i < (feed_continuous ? 1 : interval); ++i) {
	if (! feeder->get (vocab))
	  return i > 0;
	vocabs.push_back (vocab);
    }
    return true;
}

SessionVocabPtr VocabSession::get() const
{
    if (vocabs.size())
      return vocabs[0];
    else
      return SessionVocabPtr();
}

SessionVocabPtr VocabSession::get_old() const
{
    return vocab_old;
}

void VocabSession::defer (const bool force)
{
    // not answered vocabularies are forced and go unchanged
    // to be asked again before all others
    if (force)
      defer_vocab (true);
    else if (--(get()->repeat_num))
      defer_vocab (false);
}

void VocabSession::defer_vocab (const bool ask_sooner)
{
    if (ask_sooner)
      sooner.push_back (vocabs[0]);
    else
      defered.push_back (vocabs[0]);
}

void VocabSession::prepare_stats()
{
    stats_right = 0;
    stats_wrong = 0;
    stats_unknown = 0;
    
    std::for_each (history.begin(), history.end(), 
      bind (&VocabSession::update_stats, this, _1));
}

void VocabSession::update_stats (const SessionVocabPtr vocab)
{
    stats_right += vocab->answered_right;
    stats_wrong += vocab->answered_wrong;
    stats_unknown += vocab->answered_unknown;
}

unsigned VocabSession::get_stats_vocabs() const
{
    return history.size();
}

unsigned VocabSession::get_stats_right() const
{
    return stats_right;
}

unsigned VocabSession::get_stats_wrong() const
{
    return stats_wrong;
}

unsigned VocabSession::get_stats_unknown() const
{
    return stats_unknown;
}

unsigned VocabSession::get_stats_total() const
{
    return stats_right + stats_wrong + stats_unknown;
}
