/***************************************************************************
 *            VocabTest.cc
 *
 *  Tue Oct 19 14:09:50 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "VocabTest.h"
#include "VocabParser.h"
#include "exception.h"

#include "langsupport.h"
#include "debug.h"

#include <glibmm/fileutils.h>
#include <glibmm/convert.h>

#include <iostream>
#include <fstream>
#include <algorithm>
#include <numeric>

VocabTest::VocabTest() : 
  types (new VocabTypes)
{ 
    reset(); 
}

VocabTest::~VocabTest()
{ 
}

void VocabTest::read_xml (const std::string& docname)
{
    VocabParser parser (chapters, types, languages, properties);
    Glib::Markup::ParseContext context (parser);
      
    try {
	const std::string contents = Glib::file_get_contents (docname);
	// Only if the reading was successful a reset is done
	// Will still mess up on xml syntax errors TODO
	reset();
	context.parse (contents);
	context.end_parse();
	
	filename = docname;
	was_changed (false);
    } catch(const Glib::Error& ex) {
	throw ApplicationError (ex.what());
    }
}

void VocabTest::read_xml_data (const std::string& data)
{
    reset();
    
    VocabParser parser (chapters, types, languages, properties);
    Glib::Markup::ParseContext context (parser);
      
    try {
	context.parse (data);
	context.end_parse();
	was_changed (false);
    } catch (const Glib::Error& ex) {
	throw ApplicationError (ex.what());
    }
}

void VocabTest::simple()
{
    reset();

    insert_chapter (0);
    set_chapter_name (0, "Unnamed chapter");

    insert_section (0, 0);
    set_section_name (0, 0, "Unnamed section");

    was_changed (false);
}

void VocabTest::write_xml() {
    write_xml (filename);
}

void VocabTest::write_xml (const std::string& docname)
{
    try {
	std::ofstream file (Glib::filename_from_utf8 (docname).c_str());
	if (!file) throw ApplicationError (_("Unknown error."));
	write_xml (file);
	filename = docname;
	was_changed (false);
    } catch (const Glib::Error& ex) {
	throw ApplicationError (ex.what());
    }
}

void VocabTest::write_xml (std::ostream& stream)
{
    stream << *this;
}

const std::string& VocabTest::get_filename() const {
    return filename;
}

const bool VocabTest::get_modified() const {
    return modified;
}

void VocabTest::was_changed (const bool changed) {
    modified = changed;
    signal_on_modified_v (changed);
}

/** Reset all fields to empty.
 */

void VocabTest::reset()
{
    chapters.clear();
    languages.clear();
    properties.clear();

    filename.clear();
    was_changed (false);
}

const std::vector<std::string> VocabTest::get_languages() const
{ 
    return std::vector<std::string>(languages.begin(), languages.end());
}

void VocabTest::add_language (const std::string& language)
{ 
    languages.insert (language);
}

const VocabTypesPtr VocabTest::get_types() const
{ 
    return types;
}

bool compare_pair (const std::pair<int, int>& left, const std::pair<int, int>& right) 
{
    if (left.first < right.first)
      return true;
    else if (left.first > right.first)
      return false;
    else 
      return left.second < right.second;
}

/** Normalize a list of chapter and section numbers.
 * The list is sorted and redundant entries (e.g. sections in also selected
 * chapters) are removed.
 */
void VocabTest::normalize_indizes (std::vector<std::pair<int, int> >& data) 
{
    std::sort (data.begin(), data.end(), &compare_pair);
    int last_chapter = -1;
    for (std::vector<std::pair<int, int> >::iterator i = data.begin(); i != data.end();) {
	if (i->second == -1) {
	  last_chapter = i->first;
	  ++i;
	} else if (i->first == last_chapter) {
	  i = data.erase (i);
	} else {
	  ++i;
	}
    }
}


// Internal stuff

const std::string VocabTest::get_vocab (int chapter, int section, int index, const std::string& language) const
{
    try {
	const VocabularyPtr vocab = chapters.at (chapter)->sections.at (section)->vocabs.at (index);
	return vocab->get_translation (language);
    } catch (...) {};
    return "";
}

void VocabTest::set_vocab (int chapter, int section, int index, const std::string& language, const std::string& newtrans) 
{
    try {
	VocabularyPtr vocab = chapters.at (chapter)->sections.at (section)->vocabs.at (index);
	if (vocab->get_translation (language) != newtrans) {
	    vocab->set_translation (language, newtrans);
	    vocab->new_id();
	    signal_on_changed_v (chapter, section, index);
	    was_changed (true);
	}
    } catch (...) {};
}

unsigned VocabTest::get_id (int chapter, int section, int index) const
{
    try {
	VocabularyPtr vocab = chapters.at (chapter)->sections.at (section)->vocabs.at (index);
	return vocab->get_id();
    } catch (...) {};
    return 0;
}

void VocabTest::set_id (int chapter, int section, int index, unsigned id)
{
    try {
	VocabularyPtr vocab = chapters.at (chapter)->sections.at (section)->vocabs.at (index);
	vocab->set_id (id);
    } catch (...) {};
}

void VocabTest::insert_vocab (int chapter, int section, int index) 
{
    try {
	DataSectionPtr sec = chapters.at (chapter)->sections.at (section);
	sec->vocabs.insert (sec->vocabs.begin() + index, VocabularyPtr(new Vocabulary (types)));
	notify_insert_vocab (chapter, section, index);
    } catch (...) {};
    was_changed (true);
}

void VocabTest::copy_vocab (int source_chapter, int source_section, int source_vocab, int dest_chapter, int dest_section, int dest_vocab)
{
    try {
	DataSectionPtr sec = chapters.at (dest_chapter)->sections.at (dest_section);
	const VocabularyPtr vocab = chapters.at (source_chapter)->sections.at (source_section)->vocabs.at (source_vocab);
	if (dest_vocab == -1) dest_vocab = 0;
	sec->vocabs.insert (sec->vocabs.begin() + dest_vocab, VocabularyPtr (new Vocabulary (*vocab)));
	notify_insert_vocab (dest_chapter, dest_section, dest_vocab);
    } catch (...) {};
    was_changed (true);
}

void VocabTest::copy_vocabs (int source_chapter, int source_section, std::vector<int> source_vocabs, int dest_chapter, int dest_section, int dest_vocab)
{
    // We don't copy to ourselves
    if (source_chapter == dest_chapter && source_section == dest_section) return;
    try {
	DataSectionPtr sec = chapters.at (dest_chapter)->sections.at (dest_section);
	const DataSectionPtr source = chapters.at (dest_chapter)->sections.at (dest_section);
	if (dest_vocab == -1) dest_vocab = 0;
	// This should be nicer
	for (int i = source_vocabs.size(); i; i--) {
	    const VocabularyPtr vocab = source->vocabs.at (source_vocabs[i - 1]);
	    sec->vocabs.insert (sec->vocabs.begin() + dest_vocab, VocabularyPtr (new Vocabulary (*vocab)));
	    notify_insert_vocab (dest_chapter, dest_section, dest_vocab);
	}
    } catch (...) {};
    was_changed (true);
}

void VocabTest::insert_section (int chapter, int section) 
{
    try {
	DataChapterPtr chap = chapters.at (chapter);
	chap->sections.insert (chap->sections.begin() + section, DataSectionPtr(new DataSection));
	notify_insert_section (chapter, section);
    } catch (...) {}
    was_changed (true);
}

void VocabTest::notify_insert_vocab (const int chapter, const int section, const int vocab)  {
    if (get_vocab_num (chapter, section) == 1)
      signal_on_has_child_toggled_v (chapter, section);
    signal_on_inserted_v (chapter, section, vocab);
}

// BUG: doesn't notify for child_toggled if its a copied section; works anyway
void VocabTest::notify_insert_section (const int chapter, const int section)  {
    if (get_section_num (chapter) == 1)
      signal_on_has_child_toggled_v (chapter, -1);
    signal_on_inserted_v (chapter, section, -1);
    int vocab_num = get_vocab_num (chapter, section);
    for (int i = 0; i < vocab_num; i++) 
      notify_insert_vocab (chapter, section, i);
}

void VocabTest::notify_insert_chapter (const int chapter)  {
    signal_on_inserted_v (chapter, -1, -1);
    int section_num = get_section_num (chapter);
    for (int i = 0; i < section_num; i++) 
      notify_insert_section (chapter, i);
}

void VocabTest::copy_section (int source_chapter, int source_section, int dest_chapter, int dest_section)
{
    try {
	DataChapterPtr chap = chapters.at (dest_chapter);
	const DataSectionPtr sec = chapters.at (source_chapter)->sections.at (source_section);
	if (dest_section == -1) dest_section = 0;
	chap->sections.insert (chap->sections.begin() + dest_section, DataSectionPtr (new DataSection (*sec)));
    } catch (...) {};
    notify_insert_section (dest_chapter, dest_section);
    was_changed (true);
}

void VocabTest::insert_chapter (int chapter) 
{
    try {
	chapters.insert (chapters.begin() + chapter, DataChapterPtr (new DataChapter));
	notify_insert_chapter (chapter);
    } catch (...) {}
    was_changed (true);
}

void VocabTest::copy_chapter (int source_chapter, int dest_chapter)
{
    try {
	const DataChapterPtr chap = chapters.at (source_chapter);
	if (dest_chapter == -1) dest_chapter = 0;
	chapters.insert (chapters.begin() + dest_chapter, DataChapterPtr (new DataChapter (*chap)));
	notify_insert_chapter (dest_chapter);
    } catch (...) {};
    was_changed (true);
}

void VocabTest::delete_vocab (int chapter, int section, int index) 
{
    try {
	DataSectionPtr sec = chapters.at (chapter)->sections.at (section);
	if (sec->vocabs.size() <= (unsigned) index) return;
	sec->vocabs.erase (sec->vocabs.begin() + index);
	if (! sec->vocabs.size()) signal_on_has_child_toggled_v (chapter, section);
	signal_on_deleted_v (chapter, section, index);
    } catch (...) {}
    was_changed (true);
}

void VocabTest::delete_vocabs (int chapter, int section, std::vector<int> vocabs)
{
    try {
	DataSectionPtr sec = chapters.at (chapter)->sections.at (section);
	std::sort (vocabs.begin(), vocabs.end());
	// This should be nicer
	for (int i = vocabs.size(); i; i--) {
	    sec->vocabs.erase (sec->vocabs.begin() + vocabs[i - 1]);
	    signal_on_deleted_v (chapter, section, vocabs[i - 1]);
	}
	if (! sec->vocabs.size()) signal_on_has_child_toggled_v (chapter, section);
    } catch (...) {};
    was_changed (true);
}

void VocabTest::delete_section (int chapter, int section) 
{
    try {
	DataChapterPtr chap = chapters.at (chapter);
	chap->sections.erase (chap->sections.begin() + section);
	if (! chap->sections.size()) signal_on_has_child_toggled_v (chapter, -1);
	signal_on_deleted_v (chapter, section, -1);
    } catch (...) {}
    was_changed (true);
}

void VocabTest::delete_chapter (int chapter) 
{
    chapters.erase (chapters.begin() + chapter);
    signal_on_deleted_v (chapter, -1, -1);
    was_changed (true);
}

const std::string VocabTest::get_type (int chapter, int section, int index) const
{
    try {
	return chapters.at (chapter)->sections.at (section)->vocabs.at (index)->get_type();
    } catch (...) {}
    return "";
}

void VocabTest::set_type (int chapter, int section, int index, const std::string& newtype) 
{
    try {
	const VocabularyPtr vocab = chapters.at (chapter)->sections.at (section)->vocabs.at (index);
	if (vocab->get_type() != newtype) {
	    vocab->set_type (newtype);
	    signal_on_changed_v (chapter, section, index);
	    was_changed (true);
	}
    } catch (...) {}
}

int VocabTest::get_vocab_num (int chapter, int section) const
{
    try {
	return chapters.at (chapter)->sections.at (section)->vocabs.size();
    } catch (...) {}
    return -1;
}

int VocabTest::get_section_num (int chapter) const
{
    try {
	return chapters.at (chapter)->sections.size();
    } catch (...) {}
    return -1;
}

int VocabTest::get_chapter_num() const
{
    return chapters.size();
}

void VocabTest::set_section_name (int chapter, int section, const std::string& newname) 
{
    try {
	const DataSectionPtr sec = chapters.at (chapter)->sections.at (section);
	if (sec->get_title() != newname) {
	    sec->set_title (newname);
	    signal_on_changed_v (chapter, section, -1);
	    was_changed (true);
	}
    } catch (...) {}
}

void VocabTest::set_chapter_name (int chapter, const std::string& newname)
{
    try {
	const DataChapterPtr chap = chapters.at (chapter);
	if (chap->get_title() != newname) {
	    chap->set_title (newname);
	    signal_on_changed_v (chapter, -1, -1);
	    was_changed (true);
	}
    } catch (...) {}
}

const std::string VocabTest::get_section_name (int chapter, int section) const
{
    try {
	return chapters.at (chapter)->sections.at (section)->get_title();
    } catch (...) {}
    return "";
}

const std::string VocabTest::get_chapter_name (int chapter) const
{
    try {
	return chapters.at (chapter)->get_title();
    } catch (...) {}
    return "";
}

/** Get a global file property value.
 * These are saved in the XML file.
 * \param key Identification key for the property
 * \return Property value
 */

std::string VocabTest::get_property (const std::string& key) const
{
    PropertyMap::const_iterator i;
    if ((i = properties.find (key)) != properties.end())
      return i->second;
    else
      return "";
}

/** Set a global file property value.
 * These are saved in the XML file.
 * \param key Identification key for the property
 * \param value Property value
 */

void VocabTest::set_property (const std::string& key, const std::string& value)
{
    if (value == get_property (key)) return;
    if (value.size())
      properties[key] = value;
    else
      properties.erase (key);
    was_changed (true);
}

VocabTest::type_signal_changed VocabTest::signal_on_changed()
{
  return signal_on_changed_v;
}

VocabTest::type_signal_inserted VocabTest::signal_on_inserted()
{
  return signal_on_inserted_v;
}

VocabTest::type_signal_deleted VocabTest::signal_on_deleted()
{
  return signal_on_deleted_v;
}

VocabTest::type_signal_has_child_toggled VocabTest::signal_on_has_child_toggled()
{
  return signal_on_has_child_toggled_v;
}

VocabTest::type_signal_modified VocabTest::signal_on_modified()
{
  return signal_on_modified_v;
}
