/***************************************************************************
 *            VocabTreeModel.cc
 *
 *  Wed Oct 20 13:49:54 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "debug.h"
#include "functions.h"
#include "langsupport.h"
#include "stringhelper.h"

#include "VocabTreeModel.h"
#include "VocabTest.h"
#include "VocabCache.h"

#include <gtk/gtktreednd.h>

#include <boost/bind.hpp>

#include <iomanip>

boost::array<double, 3> red_array =   {{0,   1,   1}};
boost::array<double, 3> green_array = {{1,   1,   0}};

func::PieceWiseLinear<double, double, 3> red_color (red_array);
func::PieceWiseLinear<double, double, 3> green_color (green_array);

VocabTreeModel::VocabTreeModel (const double fade, boost::shared_ptr<VocabTest> vocab_test) :
  Glib::ObjectBase (typeid (VocabTreeModel)), 
  Glib::Object(),
  vocab_test (vocab_test), 
  vocab_cache (VocabCache::get()), 
  depth (2), 
  stamp (1),
  fade (fade)
{
    // Register all columns
    columns_record.add (type_column);
    columns_record.add (title_column);
    columns_record.add (colorname_column);
    columns_record.add (score_column);
    columns_record.add (expired_column);
    columns_record.add (expire_time_column);
    translation_columns.resize (vocab_test->get_languages().size());
    std::for_each (translation_columns.begin(), translation_columns.end(), 
      bind (&VocabTreeModel::ColumnRecord::add, &columns_record, _1));
    
    // only set iterator, path should be empty anyway
    get_root_iterator (root_iterator);
    
    vocab_test->signal_on_changed().connect (sigc::mem_fun (*this,
      &VocabTreeModel::on_test_changed));
    vocab_test->signal_on_inserted().connect (sigc::mem_fun (*this,
      &VocabTreeModel::on_test_inserted));
    vocab_test->signal_on_deleted().connect (sigc::mem_fun (*this,
      &VocabTreeModel::on_test_deleted));
    vocab_test->signal_on_has_child_toggled().connect (sigc::mem_fun (*this,
      &VocabTreeModel::on_test_has_child_toggled));
}

VocabTreeModel::~VocabTreeModel()
{
}

Glib::RefPtr<VocabTreeModel> VocabTreeModel::create (const double fade, boost::shared_ptr<VocabTest> vocab_test)
{
    return Glib::RefPtr<VocabTreeModel> (new VocabTreeModel (fade, vocab_test));
}

Gtk::TreeModelFlags VocabTreeModel::get_flags_vfunc() const
{
    return Gtk::TreeModelFlags (0);
}

int VocabTreeModel::get_n_columns_vfunc() const
{
    return COLUMN_TRANSLATIONS_BASE + translation_columns.size();
}

GType VocabTreeModel::get_column_type_vfunc (int index) const
{
    switch (index) {
      case COLUMN_TYPE:
	return type_column.type();
      case COLUMN_TITLE:
	return title_column.type();
      case COLUMN_COLOR:
	return colorname_column.type();
      case COLUMN_SCORE:
	return score_column.type();
      case COLUMN_EXPIRED:
	return expired_column.type();
      case COLUMN_EXPIRE_TIME:
	return expire_time_column.type();
      default:
	if (index >= 0 && 
	    index - COLUMN_TRANSLATIONS_BASE < (int) translation_columns.size())
	  return translation_columns[index - COLUMN_TRANSLATIONS_BASE].type();
	else
	  return 0;
    }
}

void VocabTreeModel::get_value_vfunc (const TreeModel::iterator& iter, int column, Glib::ValueBase& value) const
{
    if (iter_is_valid (iter)) {
        if (column >= 0 && 
	    column - COLUMN_TRANSLATIONS_BASE < (int) translation_columns.size()) {
            VocabColumn::ValueType valuet;
            BoolColumn::ValueType valuet2;

            valuet.init (VocabColumn::ValueType::value_type());  
            valuet2.init (BoolColumn::ValueType::value_type());  

            int chapter = GPOINTER_TO_INT (iter.gobj()->user_data);
	    int section = GPOINTER_TO_INT (iter.gobj()->user_data2);
	    int vocab   = GPOINTER_TO_INT (iter.gobj()->user_data3);
	    
	    Glib::ustring result;
	    const unsigned day_length = 60 * 60 * 24;
	    
	    if (section == -1) {
		if (column == COLUMN_TITLE)
		  result = vocab_test->get_chapter_name (chapter);
	    } else if (vocab == -1) {
		if (column == COLUMN_TITLE)
		  result = vocab_test->get_section_name (chapter, section);
	    } else {
		std::ostringstream s;
		double score;
		time_t expire;
		time_t now;
		switch (column) {
		  case COLUMN_TYPE:
		    result = vocab_test->get_type (chapter, section, vocab);
		    break;
		  case COLUMN_COLOR:
		    score = vocab_cache.get_score (
		      vocab_test->get_id (chapter, section, vocab));
		    result = func::x_color (
		      fade + red_color (score) * (1.0 - fade), 
		      fade + green_color (score) * (1.0 - fade),
		      fade);
		    break;
		  case COLUMN_SCORE:
		    score = vocab_cache.get_score (
		      vocab_test->get_id (chapter, section, vocab));
		    s << std::fixed << std::setprecision (2) << score;
		    result = s.str();
		    break;
		  case COLUMN_EXPIRED:
		    // We short-circuit it here because the type is different
		    valuet2.set (vocab_cache.get_score_with_expires (
		      vocab_test->get_id (chapter, section, vocab)) == 1);
		    value.init (BoolColumn::ValueType::value_type());
		    value = valuet2;
		    return;
		    break;
		  case COLUMN_EXPIRE_TIME:
		    expire = vocab_cache.get_expires (
		      vocab_test->get_id (chapter, section, vocab));
		    now = time (NULL);
		    if (expire <= now)
		      result = _("expired");
		    else if (expire / day_length == now / day_length)
		      result = format_time ("%X", expire);
		    else if (expire / day_length == now / day_length + 1)
		      result = _("tomorrow");
		    else 
		      result = strmanip::ucompose (ngettext ("in %1 day", "in %1 days", 
		        expire / day_length - now / day_length), 
		        expire / day_length - now / day_length);
		    break;
		  default:
		    result = vocab_test->get_vocab (chapter, section, vocab, 
		      vocab_test->get_languages()[column - COLUMN_TRANSLATIONS_BASE]);
		    break;
	        }
	    }
		
	    valuet.set (result);
	    value.init (VocabColumn::ValueType::value_type());
	    value = valuet;
	}
    }
}

bool VocabTreeModel::iter_next_vfunc (const iterator& iter, iterator& iter_next) const
{ 
    iter_next = iterator();
    if (!iter_is_valid (iter)) return false;

    int chapter = GPOINTER_TO_INT (iter.gobj()->user_data);
    int section = GPOINTER_TO_INT (iter.gobj()->user_data2);
    int vocab   = GPOINTER_TO_INT (iter.gobj()->user_data3);
	
    if (chapter == -1) return false;
    
    if (section == -1) {
	if (vocab_test->get_chapter_num() <= chapter + 1) return false;
	iter_next = iter;
	iter_next.gobj()->user_data = GINT_TO_POINTER (GPOINTER_TO_INT (iter.gobj()->user_data) + 1);
	return true;
    }
    
    if (vocab == -1) {
	if (vocab_test->get_section_num(chapter) <= section + 1) return false;
	iter_next = iter;
	iter_next.gobj()->user_data2 = GINT_TO_POINTER (GPOINTER_TO_INT (iter.gobj()->user_data2) + 1);
	return true;
    }
    
    if (vocab_test->get_vocab_num(chapter, section) <= vocab + 1) return false;
    iter_next = iter;
    iter_next.gobj()->user_data3 = GINT_TO_POINTER (GPOINTER_TO_INT (iter.gobj()->user_data3) + 1);
    return true;
}

bool VocabTreeModel::iter_children_vfunc (const iterator& parent, iterator& iter) const
{
    return iter_nth_child_vfunc (parent, 0, iter);
}

bool VocabTreeModel::iter_has_child_vfunc (const iterator& iter) const
{
    return (iter_n_children_vfunc (iter) > 0);
}

int VocabTreeModel::iter_n_children_vfunc (const iterator& iter) const
{
    if(!iter_is_valid (iter)) return 0;

    int chapter = GPOINTER_TO_INT (iter.gobj()->user_data);
    int section = GPOINTER_TO_INT (iter.gobj()->user_data2);
    int vocab   = GPOINTER_TO_INT (iter.gobj()->user_data3);

    if ((vocab   != -1) ||
	(section != -1 && depth <= 1) ||
	(chapter != -1 && depth == 0))
      return 0;
	
    if (section != -1) return vocab_test->get_vocab_num (chapter, section);

    if (chapter != -1) return vocab_test->get_section_num (chapter);

    return vocab_test->get_chapter_num();
}

int VocabTreeModel::iter_n_root_children_vfunc() const
{
    if(!iter_is_valid (root_iterator)) return 0;
	
    int chapter = GPOINTER_TO_INT (root_iterator.gobj()->user_data);
    int section = GPOINTER_TO_INT (root_iterator.gobj()->user_data2);
    int vocab   = GPOINTER_TO_INT (root_iterator.gobj()->user_data3);

    if (vocab   != -1) return 0;
    if (section != -1) return vocab_test->get_vocab_num(chapter, section);
    if (chapter != -1) return vocab_test->get_section_num(chapter);
    return vocab_test->get_chapter_num();
}

bool VocabTreeModel::iter_nth_child_vfunc (const iterator& parent, int n, iterator& iter) const
{
    iter = iterator();
    if(!iter_is_valid (parent)) return false;

    int chapter = GPOINTER_TO_INT (parent.gobj()->user_data);
    int section = GPOINTER_TO_INT (parent.gobj()->user_data2);
    int vocab   = GPOINTER_TO_INT (parent.gobj()->user_data3);

    if ((vocab   != -1) ||
	(section != -1 && depth <= 1) ||
	(chapter != -1 && depth == 0))
      return false;
	
    if (section == -1) {
	if (vocab_test->get_section_num(chapter) <= n) return false;
	iter = parent;
        iter.gobj()->user_data2 = GINT_TO_POINTER (n);
	return true; 
    }

    if (vocab_test->get_vocab_num(chapter, section) <= n) return false;
    iter = parent;
    iter.gobj()->user_data3 = GINT_TO_POINTER (n);
    return true; 
}

bool VocabTreeModel::iter_nth_root_child_vfunc (int n, iterator& iter) const
{
    return iter_nth_child_vfunc (root_iterator, n, iter); 
}
  

bool VocabTreeModel::iter_parent_vfunc (const iterator& child, iterator& iter) const
{
    iterator temp;
    
    iter = iterator();
    if (!iter_is_valid (child)) return false;

    int chapter = GPOINTER_TO_INT (child.gobj()->user_data);
    int section = GPOINTER_TO_INT (child.gobj()->user_data2);
    int vocab   = GPOINTER_TO_INT (child.gobj()->user_data3);

    if (chapter == -1) return false;

    if (section == -1) return false; // there is no way to get the internal root node here

    if (vocab == -1) {
	temp = child;
	temp.gobj()->user_data2 = GINT_TO_POINTER (-1);
	if (iters_are_equal (temp, root_iterator)) return false;
	iter = temp;
	return true;
    }

    temp = child;
    temp.gobj()->user_data3 = GINT_TO_POINTER (-1);
    if (iters_are_equal(temp, root_iterator)) return false;
    iter = temp;
    return true;
}

Gtk::TreeModel::Path VocabTreeModel::get_path_vfunc (const iterator& iter) const
{
    if(!iter_is_valid (iter)) return Path();

    int chapter  = GPOINTER_TO_INT (iter.gobj()->user_data);
    int section  = GPOINTER_TO_INT (iter.gobj()->user_data2);
    int vocab    = GPOINTER_TO_INT (iter.gobj()->user_data3);

    int rchapter = GPOINTER_TO_INT (root_iterator.gobj()->user_data);
    int rsection = GPOINTER_TO_INT (root_iterator.gobj()->user_data2);
    int rvocab   = GPOINTER_TO_INT (root_iterator.gobj()->user_data3);

    Path p;
    if (chapter  == -1) return p;
    if (rchapter == -1) p.push_back (chapter);
    if (section  == -1) return p;
    if (rsection == -1) p.push_back (section);
    if (vocab    == -1) return p;
    if (rvocab   == -1) p.push_back (vocab);
    return p;    
}

bool VocabTreeModel::get_iter_vfunc (const Path& path, iterator& iter) const
{
    iter = iterator();
    
    // Prepend our root path
    Path p (root_path);
    p.append (path.begin(), path.end());
    
    int sz = (int) p.size();
    if (sz < 1 || sz > 1 + depth) return false;
    
    int chapter =                p[0];
    int section = sz <= 1 ? -1 : p[1];
    int vocab   = sz <= 2 ? -1 : p[2];
    
    if (                 vocab_test->get_chapter_num()                  <= chapter) return false;
    if (section != -1 && vocab_test->get_section_num (chapter)           <= section) return false;
    if (vocab   != -1 && vocab_test->get_vocab_num   (chapter, section)  <= vocab)   return false;
    
    iter.set_stamp (stamp);
    iter.gobj()->user_data  = GINT_TO_POINTER (chapter);
    iter.gobj()->user_data2 = GINT_TO_POINTER (section);
    iter.gobj()->user_data3 = GINT_TO_POINTER (vocab);
    return true; 
}

void VocabTreeModel::get_vocab_addr (const Path& path, int &chapter, int &section, int &vocab) const
{
    // Prepend our root path
    Path p (root_path);
    p.append (path.begin(), path.end());

    unsigned sz = p.size();

    chapter = sz == 0 ? -1 : p[0];
    section = sz <= 1 ? -1 : p[1];
    vocab   = sz <= 2 ? -1 : p[2];
}

/** Return a Gtk::TreePath relative to the root of the VocabTreeModel.
 * \param path New Gtk::TreePath
 * \param chapter Chapter of the vocabulary
 * \param section Section of the vocabulary
 * \param vocab Vocab number of the vocabulary
 * \return Whether the vocabulary was found below the model root and \c path is valid
 */

bool VocabTreeModel::get_vocab_path (Path& path, const int chapter, const int section, const int vocab) const
{
    path.clear();

    unsigned sz = root_path.size();
    if ((sz >= 1 && chapter != -1 && chapter != root_path[0]) ||
	(sz >= 2 && section != -1 && section != root_path[1]) ||
	(sz == 3 && vocab   != -1 && vocab   != root_path[2]))
      return false;
   
    if (sz < 1 && chapter != -1) path.push_back (chapter);
    if (sz < 2 && section != -1) path.push_back (section);
    if (sz < 3 && vocab   != -1) path.push_back (vocab);

    if (sz + depth + 1 < path.size() || ! path.size()) return false;

    return true;
}

/** Return an absolute Gtk::TreePath. It is not checked whether it can 
 * be represented in the model.
 * \param path New Gtk::TreePath
 * \param chapter Chapter of the vocabulary
 * \param section Section of the vocabulary
 * \param vocab Vocab number of the vocabulary
 */

void VocabTreeModel::get_root_vocab_path (Path& path, const int chapter, const int section, const int vocab) const
{
    if (chapter != -1) path.push_back (chapter);
    if (section != -1) path.push_back (section);
    if (vocab   != -1) path.push_back (vocab);
}

VocabTreeModel::VocabColumn& VocabTreeModel::get_model_column (const int column)
{
    switch (column) {
      case COLUMN_TYPE:
	return type_column;
      case COLUMN_TITLE:
	return title_column;
      case COLUMN_COLOR:
	return colorname_column;
      case COLUMN_SCORE:
	return score_column;
      case COLUMN_EXPIRE_TIME:
	return expire_time_column;
      default:
	return translation_columns[column - COLUMN_TRANSLATIONS_BASE];
    }
}

VocabTreeModel::BoolColumn& VocabTreeModel::get_model_column_bool (const int column)
{
    switch (column) {
      case COLUMN_EXPIRED:
	return expired_column;
      // Don't do that
      default:
	return expired_column;

    }
}

bool VocabTreeModel::iter_is_valid (const iterator& iter) const
{
    return stamp == iter.get_stamp();
}

// Doesn't care about model or stamp
bool VocabTreeModel::iters_are_equal (const iterator& iter1, const iterator& iter2) const
{
    return (iter1.gobj()->user_data  == iter2.gobj()->user_data) &&
           (iter1.gobj()->user_data2 == iter2.gobj()->user_data2) &&
           (iter1.gobj()->user_data2 == iter2.gobj()->user_data3); 
}

void VocabTreeModel::limit_to_branch (const iterator iter)
{
    // go back to show all
    get_root_iterator (root_iterator);
    // and limit
    root_path = get_path_vfunc (iter);
    root_iterator = iter;
}

void VocabTreeModel::limit_to_branch (const Path path)
{
    // go back to show all
    root_path.clear();
    // and limit
    if (path.size()) 
      get_iter_vfunc (path, root_iterator);
    else
      get_root_iterator (root_iterator);
    root_path = path;
}

// This is only a pseudo-valid iterator!
void VocabTreeModel::get_root_iterator (iterator& iter) const
{
    iter = iterator();
    
    iter.set_stamp (stamp);
    iter.gobj()->user_data  = GINT_TO_POINTER (-1);
    iter.gobj()->user_data2 = GINT_TO_POINTER (-1);
    iter.gobj()->user_data3 = GINT_TO_POINTER (-1);
}

void VocabTreeModel::on_test_changed (const int chapter, const int section, const int index) 
{
    Path p;
    // send signal only when the entry is also displayed; no iterators will be invalidated
    if (get_vocab_path (p, chapter, section, index)) {
	row_changed (p, get_iter (p));
    }
}

void VocabTreeModel::on_test_inserted (const int chapter, const int section, const int index) 
{
    Path p;
    if (get_vocab_path (p, chapter, section, index)) {
	// invalidate iterators and send appropriate signal if it is displayed
	stamp++;
	row_inserted (p, get_iter (p));
    } else if (root_path.size()) {
	get_root_vocab_path (p, chapter, section, index);
	unsigned lnew = p.size();
	if (root_path.size() <= lnew &&
	    std::equal (p.begin(), p.begin() + lnew - 1, root_path.begin()) &&
	    p[lnew - 1] <= root_path[lnew - 1]) {
	    root_path[lnew - 1]++;
	    // invalidate iterators and send fake change signal
	    stamp++;
	    row_changed (Path ("0"), get_iter (Path ("0")));
	}
    }
}

void VocabTreeModel::on_test_deleted (const int chapter, const int section, const int index) 
{
    Path p;
    if (get_vocab_path (p, chapter, section, index)) {
	// invalidate iterators and send deleted signal when this entry was displayed
	stamp++;
	row_deleted (p);
    } else if (root_path.size()) {
	get_root_vocab_path (p, chapter, section, index);
	unsigned lnew = p.size();
	if (root_path.size() <= lnew &&
	    std::equal (p.begin(), p.begin() + lnew - 1, root_path.begin()) &&
	    // We don't care for the case that we are delete ourselves; 
	    // there is nothing to display anyway and whoever deleted us
	    // should provide a new root path
	    p[lnew - 1] < root_path[lnew - 1]) {
	    root_path[lnew - 1]--;
	    // invalidate iterators and send fake change signal
	    stamp++;
	    row_changed (Path ("0"), get_iter (Path ("0")));
	}
    }
}

void VocabTreeModel::on_test_has_child_toggled (const int chapter, const int section) 
{
    Path p;
    // send signal only when the entry and its children are displayed; no iterators 
    // will be invalidated (there should be an insert or delete signal coming anyway)
    if (get_vocab_path (p, chapter, section, -1) && depth >= (int) p.size()) {
	row_has_child_toggled (p, get_iter (p));
    }
}

bool VocabTreeModel::row_draggable_vfunc (const Path& path) const 
{
    return true;
}

bool VocabTreeModel::row_drop_possible_vfunc (const Path& dest, const Gtk::SelectionData& selection_data) const 
{
    debug_log ("Trying to drop on %1", dest.to_string());
    if (selection_data.get_target() == "GTK_TREE_MODEL_ROW") {
	Glib::RefPtr<Gtk::TreeModel> model; Path source;
	Glib::RefPtr<VocabTreeModel> our_model;
	Path::get_from_selection_data (selection_data, model, source);
	try {
	    our_model = Glib::RefPtr<VocabTreeModel>::cast_dynamic (model);
	} catch (...) {
	}
	if (our_model) {
	    int source_chapter, source_section, source_vocab;
	    our_model->get_vocab_addr (source, source_chapter, source_section, source_vocab);
	    int dest_chapter, dest_section, dest_vocab;
	    get_vocab_addr (dest, dest_chapter, dest_section, dest_vocab);
	    return (source_vocab   != -1 && dest_section   != -1 ||
		    source_vocab   == -1 && source_section != -1 && dest_vocab == -1 ||
		    source_section == -1 && dest_section   == -1);
	} else {
	    return false;
	}
    } else return Gtk::TreeDragDest::row_drop_possible_vfunc (dest, selection_data);
}

bool VocabTreeModel::drag_data_delete_vfunc (const Path& path)
{
    int chapter, section, vocab;
    get_vocab_addr (path, chapter, section, vocab);
    if (vocab != -1)
      vocab_test->delete_vocab (chapter, section, vocab);
    else if (section != -1)
      vocab_test->delete_section (chapter, section);
    else 
      vocab_test->delete_chapter (chapter);
    return true;
}

bool VocabTreeModel::drag_data_received_vfunc (const Path& dest, const Gtk::SelectionData& selection_data) 
{
    debug_log ("Dropping on %1", dest.to_string());
    if (selection_data.get_target() == "GTK_TREE_MODEL_ROW") {
	Glib::RefPtr<Gtk::TreeModel> model; Path source;
	Glib::RefPtr<VocabTreeModel> our_model;
	Path::get_from_selection_data (selection_data, model, source);
	try {
	    our_model = Glib::RefPtr<VocabTreeModel>::cast_dynamic (model);
	} catch (...) {
	}
	if (our_model) {
	    int source_chapter, source_section, source_vocab;
	    our_model->get_vocab_addr (source, source_chapter, source_section, source_vocab);
	    int dest_chapter, dest_section, dest_vocab;
	    get_vocab_addr (dest, dest_chapter, dest_section, dest_vocab);
	    if (source_vocab != -1) {
		// Correct address to make it more intuitive and to hinder vocabularies from disappearing
		if (dest_vocab == -1 && dest_section > 0) dest_section--;
		vocab_test->copy_vocab (source_chapter, source_section, source_vocab, dest_chapter, dest_section, dest_vocab);
	    } else if (source_section != -1) {
		// Correct address to make it more intuitive and to hinder sections from disappearing
		if (dest_section == -1 && dest_chapter > 0) dest_chapter--;
	        vocab_test->copy_section (source_chapter, source_section, dest_chapter, dest_section);
	    } else
	      vocab_test->copy_chapter (source_chapter, dest_chapter);
	    return true;
	} else {
	    return false;
	}
    } else return Gtk::TreeDragDest::drag_data_received_vfunc (dest, selection_data);
}
