/***************************************************************************
 *            VocabTypesModel.cc
 *
 *  Thu Dec  9 16:03:45 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "debug.h"

#include "VocabTypesModel.h"

#include <gtkmm/treepath.h>

#include <boost/bind.hpp>

#include <iostream>

VocabTypesModel::VocabTypesModel (const VocabTypesPtr types) :
  Glib::ObjectBase (typeid (VocabTypesModel)), 
  types (types),
  stamp (1)
{
    columns_record.add (text_column);
    columns_record.add (num_column);
    
    types->signal_on_changed().connect (sigc::mem_fun (*this,
      &VocabTypesModel::on_types_changed));
    types->signal_on_inserted().connect (sigc::mem_fun (*this,
      &VocabTypesModel::on_types_inserted));
    types->signal_on_deleted().connect (sigc::mem_fun (*this,
      &VocabTypesModel::on_types_deleted));
}

VocabTypesModel::~VocabTypesModel()
{
}

Glib::RefPtr<VocabTypesModel> VocabTypesModel::create (const VocabTypesPtr types)
{
    return Glib::RefPtr<VocabTypesModel> (new VocabTypesModel (types));
}

Gtk::TreeModelFlags VocabTypesModel::get_flags_vfunc() const
{
    return Gtk::TreeModelFlags (0);
}

int VocabTypesModel::get_n_columns_vfunc() const
{
    debug_log ("Returning that we have 2 Columns");
    return 2;
}

GType VocabTypesModel::get_column_type_vfunc (int index) const
{
    switch (index) {
      case 0: 
	debug_log ("Returning that type of column 0 is %1", text_column.type());
	return text_column.type();
      case 1:
	debug_log ("Returning that type of column 1 is %1", num_column.type());
        return num_column.type();
      default:
	debug_log ("Returning that we have no idea about the type of column %1", index);
        return 0;
    }
}

VocabTypesModel::TextColumn& VocabTypesModel::get_column_text()
{
    return text_column;
}

VocabTypesModel::NumColumn& VocabTypesModel::get_column_num()
{
    return num_column;
}

void VocabTypesModel::get_value_vfunc (const TreeModel::iterator& iter, int column, Glib::ValueBase& value) const
{
    if (iter_is_valid (iter)) {
        if (column < get_n_columns_vfunc()) {
            int idx = GPOINTER_TO_INT (iter.gobj()->user_data);
	    if (column == 0) {
		TextColumn::ValueType valuet;
		valuet.init (TextColumn::ValueType::value_type());  
		valuet.set (types->get_name (idx));
		value.init (TextColumn::ValueType::value_type());
		value = valuet;
	    } else {
		NumColumn::ValueType valuet;
		valuet.init (NumColumn::ValueType::value_type());  
		valuet.set (types->get_num (idx));
		value.init (TextColumn::ValueType::value_type());
		value = valuet;
	    }
	}
    }
}

bool VocabTypesModel::iter_next_vfunc (const iterator& iter, iterator& iter_next) const
{ 
    iter_next = iterator();
    if (! iter_is_valid (iter)) {
	debug_log ("Returning because the given iterator is invalid");
	return false;
    }

    unsigned idx = GPOINTER_TO_INT (iter.gobj()->user_data);
	
    if (idx + 1 >= types->size()) {
	debug_log ("Returning because the new iterator would be invalid");
	return false;
    }
	
    iter_next = iter;
    iter_next.gobj()->user_data = GINT_TO_POINTER (idx + 1);
    debug_log ("Returning new iterator");
    return true;
}

bool VocabTypesModel::iter_children_vfunc (const iterator& parent, iterator& iter) const
{
    iter = iterator();
    return false;
}

bool VocabTypesModel::iter_has_child_vfunc (const iterator& iter) const
{
    return false;
}

int VocabTypesModel::iter_n_children_vfunc (const iterator& iter) const
{    
    return 0;
}

int VocabTypesModel::iter_n_root_children_vfunc() const
{
    debug_log ("Returning that we have %1 entries", types->size());
    return types->size();
}

bool VocabTypesModel::iter_nth_child_vfunc (const iterator& parent, int n, iterator& iter) const
{
    iter = iterator();
    return false;
}

bool VocabTypesModel::iter_nth_root_child_vfunc (int n, iterator& iter) const
{
    iter = iterator();
    if ((unsigned) n >= types->size()) {
	debug_log ("Returning that we no root child %1", n);
	return false;
    }
    
    iter.gobj()->user_data  = GINT_TO_POINTER (n);
    iter.set_stamp (stamp);
    debug_log ("Returning iterator for root child %1", n);
    return true; 
}
  

bool VocabTypesModel::iter_parent_vfunc (const iterator& child, iterator& iter) const
{
    iter = iterator();
    return false;
}

Gtk::TreeModel::Path VocabTypesModel::get_path_vfunc (const iterator& iter) const
{
    if(! iter_is_valid (iter)) return Path();

    Path p;
    p.push_back (GPOINTER_TO_INT (iter.gobj()->user_data));
    return p;    
}

bool VocabTypesModel::get_iter_vfunc (const Path& path, iterator& iter) const
{
    iter = iterator();
    
    if (path.size() != 1) {
	debug_log ("Returning because the path has the wrong length %1", path.size());
	return false;
    }
    
    iter.gobj()->user_data  = GINT_TO_POINTER (path[0]);
    iter.set_stamp (stamp);
    debug_log ("Returning a new iterator for %1", path.to_string());
    return true; 
}

bool VocabTypesModel::iter_is_valid (const iterator& iter) const
{
    if (stamp != iter.get_stamp()) {
	debug_log ("Returning that the given stamp %1 (should be %2) is invalid", iter.get_stamp(), stamp);
	return false;
    } 
    if (types->size() <= (unsigned) GPOINTER_TO_INT (iter.gobj()->user_data)) {
	debug_log ("Returning that the given position %1 is invalid", GPOINTER_TO_INT (iter.gobj()->user_data));
	return false;
    };
    return true;
}

void VocabTypesModel::on_types_changed (const unsigned idx) 
{
    debug_log ("Sth. was changed at %1", idx);
    Path p;
    p.push_back (idx);
    row_changed (p, get_iter (p));
}

void VocabTypesModel::on_types_inserted (const unsigned idx) 
{
    debug_log ("Sth. was inserted at %1", idx);
    Path p;
    p.push_back (idx);
    stamp++;
    row_inserted (p, get_iter (p));
}

void VocabTypesModel::on_types_deleted (const unsigned idx)
{
    debug_log ("Sth. was deleted at %1", idx);
    Path p;
    p.push_back (idx);
    stamp++;
    row_deleted (p);
}
