/***************************************************************************
 *            data.cc
 *
 *  Tue Oct 12 18:13:48 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "data.h"
#include "support.h"

#include <boost/bind.hpp>

#include <iostream>

Vocabulary::Vocabulary(VocabTypesPtr types) :
  id (generic_random()),
  types (types)
{
    init();
}

Vocabulary::Vocabulary (VocabTypesPtr types, const unsigned id) :
  id (id ? id : generic_random()),
  types (types)
{
    init();
}

Vocabulary::~Vocabulary()
{
    types->del_usage (type);
}

void Vocabulary::init()
{
    types->add_usage (type);
}

Vocabulary::Vocabulary (const Vocabulary& other)
{
    copy (other);
}
    
const Vocabulary& Vocabulary::operator= (const Vocabulary& other)
{
    if (this == &other) return *this;
    copy (other);
    return *this;
}

void Vocabulary::copy (const Vocabulary& other)
{
    id = other.id;
    translations = other.translations;
    if (types) types->del_usage (type);
    types = other.types;
    type = other.type;
    types->add_usage (type);
}

/** Get the translation for a specific language.
 * \param which Language for the translation
 * \return Translation
 */

std::string Vocabulary::get_translation (const std::string& which) const {
    LanguageMap::const_iterator i;
    if ((i = translations.find (which)) != translations.end())
      return i->second;
    else
      return "";
}

/** Set the translation for a specific language.
 * \param which Language for the translation
 * \param what The translation itself
 */

void Vocabulary::set_translation (const std::string& which, const std::string& what) {
    translations[which] = what;
}

/** Set a new id value.
 * Should be called after the vocabulary has been modified.
 */

void Vocabulary::new_id() {
    id = generic_random();
}

DataSection::DataSection (const DataSection& other)
{
    copy (other);
}
    
const DataSection& DataSection::operator= (const DataSection& other)
{
    if (this == &other) return *this;
    copy (other);
    return *this;
}

void DataSection::copy (const DataSection& other)
{
    title = other.title;
    std::for_each (other.vocabs.begin(), other.vocabs.end(), 
      bind (&DataSection::copy_vocab, this, _1));
}

void DataSection::copy_vocab (const VocabularyPtr& other)
{
    vocabs.push_back (VocabularyPtr (new Vocabulary (*other)));
}

DataChapter::DataChapter (const DataChapter& other)
{
    copy (other);
}

const DataChapter& DataChapter::operator= (const DataChapter& other)
{
    if (this == &other) return *this;
    copy (other);
    return *this;
}

void DataChapter::copy (const DataChapter& other)
{
    title = other.title;
    difficulty = other.difficulty;
    std::for_each (other.sections.begin(), other.sections.end(), 
      bind (&DataChapter::copy_section, this, _1));
}

void DataChapter::copy_section (const DataSectionPtr& other)
{
    sections.push_back (DataSectionPtr (new DataSection (*other)));
}
