/***************************************************************************
 *            debug.cc
 *
 *  Sat Nov  6 16:04:36 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/*  NOTES: based on a similar interface in rhythmbox:
 *    Copyright (C) 2002 Jorn Baayen
 *    Copyright (C) 2003 Colin Walters <walters@verbum.org>
 */

/** \file 
 * Debug functions for logging and debugger support.
 * \todo TODO: This needs to be sorted a little bit with 
 * g_log_set_default_handler() from 2.6, and do we need debug_critical
 * without debug system enabled?
 */

#ifdef DEBUG

#include "RegisteredSingleton.h"
#include "ProgramPaths.h"
#include "debug.h"
#include "langsupport.h"

#include <glibmm/arrayhandle.h>

#include <boost/bind.hpp>
#include <boost/scoped_ptr.hpp>

#include <algorithm>
#include <iostream>

#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <time.h>

namespace debug_internal {

    // Registrar - Declaration
        
    /** Small singleton class to manage the list of available debug modules.
     */
    
    class Registrar : public RegisteredSingleton<Registrar, 99> 
    {
      public:
	void add (const char* base);

	/** List of all known debug modules. */
        std::set<std::string> modules;
      
	/** List of all enabled debug modules. */
        std::set<std::string> enabled;
    };

    /** Add file to list of known modules.
     * \param base File name
     */
    
    void Registrar::add (const char* base)
    {
	modules.insert (base);
    }

    // RegisterFile
    
    /** Register a file with the debug interface. 
     * See Registrar for details. 
     * \param base Normally __BASE_FILE__ with gcc to register the current input file
     */
    
    RegisterFile::RegisterFile (const char* base)
    {
	Registrar::get().add (base);
    }
}

using namespace debug_internal;

namespace 
{    
    // C log stuff

    gboolean debugging = FALSE;

    void log_handler (const char *domain,
		      GLogLevelFlags level,
		      const char *message,
		      gpointer data);

    // The glib options stuff
    
    gboolean debug_enable = FALSE;
    gboolean debug_list = FALSE;
    gboolean debug_all = FALSE;
    gchar** debug_domains = NULL;
    
    GOptionEntry debug_entries[] = 
    {
	{ "debug", 'd', 0, G_OPTION_ARG_NONE, &debug_enable, N_("Enable debugging"), NULL },
	{ "debug-list", 0, 0, G_OPTION_ARG_NONE, &debug_list, N_("Display available debug domains"), NULL },
	{ "debug-domain", 0, 0, G_OPTION_ARG_STRING_ARRAY, &debug_domains, N_("Enable a specific debugging domain"), N_("MODULE") },
	{ "debug-all", 0, 0, G_OPTION_ARG_NONE, &debug_all, N_("Enable all debugging messages"), NULL },
	{ NULL }
    };

    /** Raise a SIGINT signal to get the attention of the debugger.
     * When not running under the debugger, we don't want to stop,
     * so we ignore the signal for just the moment that we raise it.
     */

    void debug_log_stop_in_debugger()
    {
	void (* saved_handler) (int);

	saved_handler = signal (SIGINT, SIG_IGN);
	raise (SIGINT);
	signal (SIGINT, saved_handler);
    }
    
    /** Stop in the debugger after running the default log handler.
     * This makes certain kinds of messages stop in the debugger
     * without making them fatal (you can continue).
     * \param domain Glib debug domain
     * \param level Glib log level
     * \param message Glib debug message
     * \param data Glib additional data
     */

    void log_handler (const char *domain, GLogLevelFlags level, const char *message, gpointer data)
    {
	g_log_default_handler (domain, level, message, data);
	if ((level & (G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING)) != 0)
	{
	    debug_log_stop_in_debugger();
	}
    }
}

/** Setup the debug command line parameters.
 * If DEBUG is not defined, this does nothing. Otherwise, the available parameters are:
\verbatim
  --help-debug              Show debug options

Debug Options:
  -d, --debug               Enable debugging
  --debug-list              Display available debug domains
  --debug-domain=MODULE     Enable a specific debugging domain
  --debug-all               Enable all debugging messages
\endverbatim
 */

void
debug_log_setup_cmdline (GOptionContext* context)
{
    GOptionGroup* debug_group = g_option_group_new ("debug", _("Debug Options:"), _("Show debug options"), NULL, NULL);
    g_option_group_set_translation_domain (debug_group, ProgramPaths::domain().c_str());
    g_option_group_add_entries (debug_group, debug_entries);
    g_option_context_add_group (context, debug_group);
}

/** Setup the debug system using the parameters provided on the command line.
 * You need to call debug_log_setup_cmdline() and g_option_context_parse()
 * first.
 */

void
debug_log_init_cmdline() 
{
    debug_log_init (debug_enable || debug_domains || debug_all);
    
    if (debug_domains)
      debug_log_show (debug_domains);
    if (debug_all)
      debug_log_show_all();
    
    if (debug_list) {
	std::cout << _("Available debug domains:\n");
	debug_log_available();
	exit (1);
    }

    if (debug_enable || debug_domains || debug_all) {
	std::cout << _("Activated debug domains:\n");
	debug_log_list();
    }
}

/** Setup the debug system. 
 * The Glib debug handler is changed to log_handler() to raise a SIGINT to get
 * the attention of the debugger.
 * \param debug Whether or not to enable the debug system
 */

void
debug_log_init (bool debug)
{
	Registrar& r = Registrar::get();
	r.enabled.insert (__FILE__);

	/* This is a workaround for the fact that there is no way to 
	 * make this useful debugging feature happen for ALL domains.
	 */
	static const char * const standard_log_domains[] = {
		"",
		"Gtk",
		"GLib",
		"GLib-GObject",
//		"glibmm"
	};

	debugging = debug;

	if (debugging)
		for (guint i = 0; i < G_N_ELEMENTS (standard_log_domains); i++)
			g_log_set_handler (standard_log_domains[i], G_LOG_LEVEL_MASK, log_handler, NULL);

	debug_log ("Debugging enabled");
}

/** Print out all enabled debug domains.
 * The output goes to std::cout.
 */

void debug_log_list()
{
    Registrar& r = Registrar::get();
    std::transform (r.enabled.begin(), r.enabled.end(), 
      std::ostream_iterator<std::string> (std::cout, "\n"),
      boost::bind (std::plus<std::string>(), "  ", _1));
}

/** Print out all available debug domains.
 * The output goes to std::cout.
 */

void debug_log_available()
{
    Registrar& r = Registrar::get();
    std::transform (r.modules.begin(), r.modules.end(), 
      std::ostream_iterator<std::string> (std::cout, "\n"),
      boost::bind (std::plus<std::string>(), "  ", _1));
}

/** Enables debug domains.
 * Intersects the lists of available domains with the list in \c domains and
 * enables them.
 * \param domains List of domains to activate
 */

void debug_log_show (const char * const * domains)
{
    Registrar& r = Registrar::get();
    Glib::ArrayHandle<std::string> ah (domains);
    std::set<std::string> s (ah.begin(), ah.end());
    std::set_intersection (r.modules.begin(), r.modules.end(), 
      s.begin(), s.end(), std::inserter (r.enabled, r.enabled.begin()));
}

/** Enables all debug domains.
 */

void debug_log_show_all()
{
    Registrar& r = Registrar::get();
    r.enabled.insert (r.modules.begin(), r.modules.end());
}

namespace debug_internal {

    /** Our own funky debugging function.
     * Should only be used when something is not going wrong, if something *is*
     * wrong use g_warning. The message is only printed if the debug domain in
     * \c file was activated with debug_log_show_all() or debug_log_show().
     * \param func Name of the current function
     * \param file Current file name
     * \param message Debug message
     */

    void debug_log_real (const char *func, const char *file, const std::string& message) 
    {
	char *str_time;
	time_t the_time;

	Registrar& r = Registrar::get();
	if (! debugging || r.enabled.find (file) == r.enabled.end()) return;

	time (&the_time);
	str_time = g_new0 (char, 255);
	strftime (str_time, 254, "%H:%M:%S", localtime (&the_time));

	g_printerr ("[%p] [%s] %s (%s): %s\n", g_thread_self(),
		    func, file, str_time, message.c_str());
	
	g_free (str_time);
    }

    void debug_critical_real (const char *func, const char *file, const std::string& message) 
    {
	char *str_time;
	time_t the_time;

	Registrar& r = Registrar::get();
	if (! debugging || r.enabled.find (file) == r.enabled.end()) return;

	time (&the_time);
	str_time = g_new0 (char, 255);
	strftime (str_time, 254, "%H:%M:%S", localtime (&the_time));

	g_critical ("[%p] [%s] %s (%s): %s\n", g_thread_self(),
	  func, file, str_time, message.c_str());
	
	g_free (str_time);
    }

    void debug_assert_real (const char *func, const char *file, const int line,
	const bool cond, const char* message)
    {
	char *str_time;
	time_t the_time;

	if (cond) return;

	time (&the_time);
	str_time = g_new0 (char, 255);
	strftime (str_time, 254, "%H:%M:%S", localtime (&the_time));

	g_error ("[%p] [%s] %s:%d (%s): Assertion failed: %s\n", g_thread_self (),
	  func, file, line, str_time, message);

	g_free (str_time);
    }
    
    
}

#endif
