/***************************************************************************
 *            debug.h
 *
 *  Sat Nov  6 16:08:36 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _DEBUG_H
#define _DEBUG_H

/** \file 
 * Debug functions for logging and debugger support.
 */

#if defined(DEBUG) 

    #include <ucompose.hpp>
    #include <glib.h>

    #include <string>
    #include <set>

    #if __GNUC__ >= 3
	/** Log a debug message using a syntax similar to printf.
	 * \code
	 * debug_log ("Timeout was changed to %1:", timeout);
	 * \endcode
	 * will output something like
\verbatim
[...] [Setting::changed()] Settings.cc (12:00:12): Timeout was changed to 30
\endverbatim
	 */
	#define debug_log(...) debug_internal::debug_log_real (__PRETTY_FUNCTION__, __BASE_FILE__, strmanip::ucompose (__VA_ARGS__))
	#define debug_critical(...) debug_internal::debug_critical_real (__PRETTY_FUNCTION__, __BASE_FILE__, strmanip::ucompose (__VA_ARGS__))
	#define debug_assert(cond) debug_internal::debug_assert_real (__PRETTY_FUNCTION__, __FILE__, __LINE__, cond, # cond)
    #elif defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
	#define debug_log(...) debug_internal::debug_log_real (__func__, __BASE_FILE__, strmanip::ucompose (__VA_ARGS__))
	#define debug_critical(...) debug_internal::debug_critical_real (__func__, __BASE_FILE__, strmanip::ucompose (__VA_ARGS__))
	#define debug_assert(cond) debug_internal::debug_assert_real (__func__, __FILE__,  __LINE__, cond, # cond)
    #else
	#define debug_log(...) (static_cast<void>(0)) 
	#define debug_critical(...) (static_cast<void>(0)) 
	#define debug_assert(...) (static_cast<void>(0))
    #endif

void debug_log_setup_cmdline (GOptionContext* context);
void debug_log_init_cmdline();
void debug_log_init (bool debug);
void debug_log_show (const char * const * domains);
void debug_log_show_all();
void debug_log_list();
void debug_log_available();

/** This contains internal debug classes and functions
 */

namespace debug_internal {

    void debug_log_real             (const char *func,
				     const char *file,
				     const std::string& message);

    void debug_critical_real        (const char *func,
				     const char *file,
				     const std::string& message);

    void debug_assert_real 	    (const char *func,
				     const char *file,
                                     const int line,
                                     const bool cond,
                                     const char* message);
    
    /** Register the current file with the debug interface.
     * An instance of this class is included in an anonymous namespace in each debug.h
     * inclusion to track the available modules for debugging.
     */
    class RegisterFile 
    {
      public:
	RegisterFile (const char* base);
    }; 
    
}

namespace {
    debug_internal::RegisterFile file (__BASE_FILE__);
}

#else

    #define debug_log(...) (static_cast<void>(0))
    #define debug_assert(...) (static_cast<void>(0))
    #define debug_critical(...) (static_cast<void>(0))
    #define debug_log_setup_cmdline(context) (static_cast<void>(0))
    #define debug_log_init_cmdline() (static_cast<void>(0))
    
#endif

#endif
