/***************************************************************************
 *            functions.h
 *
 *  Thu Jan 20 12:33:49 2005
 *  Copyright  2005  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _FUNCTIONS_H
#define _FUNCTIONS_H

#include <boost/static_assert.hpp>
#include <boost/array.hpp>

#include <cmath>

/** \file
 * Defines some commonly used function classes.
 */
 
/** This containes some commonly used function classes.
 */
 
namespace func 
{
    
    /** Calculate the natural logarithm.
     * The template parameter \c T specifies the value type of the class.
     */
     
    template <typename T>
    class Ln
    {
      public:
	/** Return the result.
	 * \param value Value to use
	 * \return ln (x)
	 */
      
	T operator() (T value) 
	{
	    return std::log (value);
	}
    };
    
    /** Calculate logarithm with a base of 2 (rounded down).
     * The template parameter \c T specifies the value type of the class.
     */
     
    template <typename T>
    class Log2
    {
      public:
	/** Return the result.
	 * The value is shifted left until it becomes zero,
	 * the number of steps is counted and returned minus 1.
	 * \warning This may give not the expected behaviour if
	 * it is used with negative values!
	 * \param value Value to use
	 * \return Number of bits
	 */
      
	T operator() (T value) 
	{
	    T help = 0;
	    while (value != 0) {
		value >>= 1;
		help++;
	    }
	    return help - 1;
	}
    };
    
    /** Transform a frequency value to mel scale. The function used here is specified by 
     * the equation \f$m = 1127.01048 \log_\e(1+f/700)\f$. 
     * The template parameter \c T specifies the value type of the class.
     */
     
    template <typename T>
    class Mel
    {
      public:
	/** Return the mel equivalent.
	 * \param value Frequency value to use
	 * \return Mel value
	 */
      
	T operator() (T value) 
	{
	    return 1127.01048 * std::log (1 + value / 700);
	}
    };

    /** Transform a frequency value to mel scale. The function used here is specified by 
     * the equation \f$m = 1127.01048 \ln(1+f/700)\f$. 
     * The template parameter \c T specifies the value type for the function.
     */
    
    template <typename T>
    T mel (T value) 
    {
	    return 1127.01048 * std::log (1 + value / 700);
    }

    std::string x_color (const double r, const double g, const double b);
    
    /** Calculate a value on a piecewiece linear function with equidistant points.
     * The template parameter \c T specifies the value type of the class, \c V
     * specifies the argument type and \c N gives the number of fixed points for 
     * the curve.
     */
     
    template <typename T, typename V, int N> 
    class PieceWiseLinear 
    {
      public:
	typedef T           value_type;            
	typedef V           index_type;            
	typedef int         size_type;      
      
	BOOST_STATIC_ASSERT (N >= 2);
      
	/** Initialize the function with the function values given in \c list.
	 * They are assumed to be equidistantly distributed in <tt>[min,max]</tt>, 
	 * with the first point at <tt>(min, v1)</tt> and the last point at <tt>(max, vn)</tt>.
	 * \param list Container with function values. Has to fullfill
	 * Collection requirements.
	 * \param min Minimum argument expected
	 * \param max Maximum argument expected
	 */
     
	template <typename A> 
	PieceWiseLinear (const A& list, const V min = 0, const V max = 1) :
          min (min),
          max (max) 
	{
	    std::copy (list.begin(), list.end(), pieces.begin());
	}
	
	/** Calculate a value.
	 * The argument range is clipped by 0 and 1, returning the corresponding 
	 * values for the function. Between the points, values are interpolated
	 * with linear functions.
	 * \param value Argument for the function
	 * \return Calculated result	 
	 */
     
	T operator() (const V value)
	{
	    if (value < min) 
	      return pieces[0];
	    if (value >= max) 
	      return pieces[N - 1];
	    size_type pos = static_cast<size_type> ((value - min) / (max - min) * (N - 1));
	    
	    // These shouldn't be necessary
	    if (pos < 0) 
	      return pieces[0];
	    if (pos >= N - 1) 
	      return pieces[N - 1];
	    
	    V delta = (value - min) / (max - min) * (N - 1) - pos;
	    return static_cast<value_type> (pieces[pos + 1] * delta + pieces[pos] * (1 - delta));
	}
	
      private:
	boost::array<T, N> pieces;
        const V min, max;
      
    };
    
    /** Calculate a value on a %Gamma function.
     * The %Gamma curve used here is specified by the equation
     * \f$f_G(x)=x^{1/\gamma}\f$. The template parameter \c T denotes
     * the value and argument type of the function.
     */
     
    template<typename T> 
    class Gamma
    {
      public:
	/** Initialize the function with a certain gamma value.
	 * \param gamma %Gamma value
	 */
      
	Gamma (const T& gamma) : gamma (gamma) 
        {
	};

	/** Calculate a value.
	 * \param value Argument for the function
	 * \return Calculated result	 
	 */
     
	T operator() (const T& value)
	{
	    return pow (value, 1 / gamma);
	}
	
      private:
	T gamma;
    };
    
    /** Calculate the energy for a number of values.
     * The formula used here is specified by the equation
     * \f$E=\sqrt{\sum_i{(x_i-\overline{x})^2}}\f$. The template parameter \c 
     * T denotes the value and argument type of the function.
     */
     
    template<typename T> 
    class Energy 
    {
      public:
	Energy() : sum (0), sq_sum (0), num (0) 
        {
	}
    
        /** Process a value.
         * This adds a value to the internal state.
	 * \param value Function value to add
	 */
     
	void operator() (const T& value) 
	{
	    sq_sum += value * value;
	    sum += value;
	    num++;
	}
    
        /** Return the current energy.
         * Returns the energy for all of the currently seen values. 
	 * More values can be added.
	 * \return Energy
	 */
     
	T energy() const 
	{ 
	    T result;
	    result = sqrt (num * sq_sum - sum * sum) / num; 
	    // Check for NaN
	    if (result != result) 
	      return 0;
	    else
	      return result;
	}
    
      private:
	T sum, sq_sum;
	unsigned num;
    };
    
    /** Calculate a value on a decay function based on a boolean parameter.
     * Given wether a condition is true or false, returns an altered value
     * further decayed or increased again. The formula used here for a given 
     * decay value \c c and boolean value \c b is specified by the equation 
     * \f$x_{n+1}=x_n * (1 - \frac{1}{c}) + b \frac{1}{c}\f$. 
     * The template parameter \c T denotes the value and argument type of the function.
     */
     
    template<typename T> 
    class Decay
    {
      public:
	/** Initialize the function with a decay value of 2.
	 */
      
	Decay() :
	  decay (2)
	{
	}
	
	/** Initialize the function with a certain decay value.
	 * \param decay %Decay value
	 */
      
	Decay (const unsigned decay) :
	  decay (static_cast<T> (decay))
	{
	}
    
	/** Set the decay value.
	 * \param new_decay %Decay value
	 */
      
	void set_decay (const unsigned new_decay)
	{
	    decay = static_cast<T> (new_decay);
	}
    
        /** Calculate a new value.
         * Returns an increased or decreased value.
	 * \param value Current value
	 * \param dist Whether the returned value should be decreased (\c true) or
	 * increased (\c false)
	 * \return New value
	 */
     
	T operator() (const T value, const bool dist)
	{
	    if (dist)
	      return value * (1 - 1 / decay);
	    else
	      return std::max (1.0, value * (1 - 1 / decay) + 1 / decay);
	}
	
      private:
	T decay;
    };

    /** Calculate a value on a quadratic decay function based on a boolean parameter.
     * Given wether a condition is true or false, returns an altered value
     * further decayed or increased again. The increment is done on a linear
     * function with the steepest slope of the quadratic function.
     * The template parameter \c T denotes the value and argument type of the function.
     */
    
    template<typename T> 
    class QuadraticDecay
    {
      public:

        /** Initialize and set the decay parameters. 
         * After \c n right answers, the score will be zero. The score will
         * start to degrade with a slope k times lower than at the end.
         * (This means \f$y(0) = 1, y(N) = 0, y'(0) = d, y'(N) = k d\f$)
	 * \param n Number of right answers needed
	 * \param k Steepness difference
	 */
	
	QuadraticDecay (const T n, const T k) :
	  n (n),
	  k (k)
	{
	}
    
	/** Set the decay parameters.
	 * \param new_n Number of right answers needed
	 * \param new_k Steepness difference
	 */
      
	void set (const double new_n, const double new_k)
	{
	    n = new_n;
	    k = new_k;
	}
    
	double operator() (const double x)
	{
	    return calc (x);
	}
	
        /** Calculate a new value.
         * Returns an increased or decreased value.
	 * \param value Current value
	 * \param dist Whether the returned value should be decreased (\c true) or
	 * increased (\c false)
	 * \return New value
	 */
	
     	T operator() (const T value, const bool dist)
	{
	    if (dist) {
		const T x = calc_inv (value) + 1;
		const T y = calc (x);
		return (y >= 0 ? y : 0); 
	    } else {
		const T p = 2 * k / (1 + k);
		const T y = value + p / n;
		return (y <= 1 ? y : 1);
	    }
	}
	
        /** Calculate a value on the quadratic function.
	 * \param x X value
	 * \return Y value
	 */
	
	T calc (const T x)
	{
	    const T xx = (x / n);
	    const T a = (1 - k) / (1 + k);
	    const T b = -2 / (1 + k);
	    const T c = 1;
	
	    const T y = a * xx * xx + b * xx + c;
	    return y;
	}
    
        /** Calculate the inverse value on the quadratic function.
	 * \param y Y value
	 * \return X value
	 */
	
	T calc_inv (const T y)
	{
	    if (k == 1)
	      return n * (1 - y);

	    const T p = 1 / (1 - k);
	    const T q = (1 + k) / (1 - k);
	    
	    const T xx = p + std::sqrt (p * p + q * (y - 1));
	    const T x = xx * n;
	    return x;
	}
    
      private:
	T n, k;
    };
 
}

/** Change the C locale temporarily during the lifetime of the instance.
 */
 
class TempLocale
{
  public:
    TempLocale (const std::string& new_locale);
    ~TempLocale();
	     
  private:
    const std::string old_locale;
};		 

#endif /* _FUNCTIONS_H */
