/***************************************************************************
 *            stringhelper.cc
 *
 *  Sat Nov 20 18:18:28 2004
 *  Copyright  2004  Michael Hofmann
 *  mh21@gmx.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/** \file
 * Various string functions.
 */

#include "stringhelper.h"

#include <ctime>
#include <locale>

/** Trim leading and trailing whitespace of a Glib::ustring.
 * \param text String to trim
 * \return Trimmed string
 */
 
Glib::ustring trim_whitespace (const Glib::ustring& text) {
    
  // Trim from first non-white character
  Glib::ustring::const_iterator pbegin = std::find_if (text.begin(), text.end(),
    std::not1 (std::ptr_fun (Glib::Unicode::isspace)));
  
  // To last non-white character
  Glib::ustring::const_iterator pend = std::find_if (text.rbegin(), 
    Glib::ustring::const_reverse_iterator (pbegin),
    std::not1 (std::ptr_fun (Glib::Unicode::isspace))).base();

  return Glib::ustring (pbegin, pend);
}

/** Return whether a Glib::ustring consists solely of whitespace.
 * \param text String to test
 * \return Whether the string was only whitespace
 */

bool is_only_whitespace (const Glib::ustring& text) {
  Glib::ustring::const_iterator pbegin (text.begin());
  Glib::ustring::const_iterator pend   (text.end());

  // Return true if there is no non-white character
  return std::find_if (text.begin(), text.end(), 
    std::not1 (std::ptr_fun (Glib::Unicode::isspace))) == text.end();
}

/** Format a time value (obtained from std::time() etc.) to a string.
 * \param format Format string for std::strftime:
 *   - \c %%a  abbreviated weekday name
 *   - \c %%A  full weekday name
 *   - \c %%b  abbreviated month name
 *   - \c %%B  full month name
 *   - \c %%c  locale's date and time representation
 *   - \c %%d  month as a number (01-31)
 *   - \c %%H  hour in 24-hour format (00-23)
 *   - \c %%I  hour in 12-hour format (01-12)
 *   - \c %%j  day of the year (001-366)
 *   - \c %%m  month as a number (01-12)
 *   - \c %%M  minute (00-59)
 *   - \c %%p  AM or PM (locale-specific)
 *   - \c %%S  second (00-59)
 *   - \c %%U  week of the year starting Sunday (0-53)
 *   - \c %%w  weekday as a number (0=Sunday to 6=Saturday)
 *   - \c %%W  week of the year starting Monday (0-53)
 *   - \c %%x  locale's date representation
 *   - \c %%X  locale's time representation
 *   - \c %%y  two-digit year (00-99)
 *   - \c %%Y  four-digit year (1900+)
 *   - \c %%z  time zone name
 *   - \c %%%  The percent symbol (%%)
 * \param time Time value to be formated
 * \return Formatted string
 */
 
std::string format_time (const std::string& format, const time_t& time)
{
    char buf[128];
    if (std::strftime (buf, sizeof (buf), format.c_str(), localtime (&time)))
      return std::string (buf);
    else
      return std::string();
}
